#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
"""SBAS Toolbox"""

from statics import *
from classes import *
import da
import arcpy
import json
import os
import imports
import get
import load
import architecture

__author__ = 'k5opjeas'

sbas_dir = os.path.dirname(os.path.realpath(__file__))
arcpy.env.parallelProcessingFactor = '75%'

UNITS = ['Cubic Yards/Year', 'Cubic Meters/Year', 'Cubic Yards', 'Cubic Meters']


class Toolbox(object):
    """Toolbox for SBAS"""

    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = 'SBAS Toolbox'
        self.alias = ''
        self.description = 'A toolbox to perform SBAS-related functions in Arc.'

        # List of tool classes associated with this toolbox
        self.tools = [
            ConvertGraphics, CreateScenario, ImportGenCade, SetUpEnvironment, UpdateLittoralScenario,
            LoadAlternative, UpdateFluxScenario, ChangeSymbology, AddFeaturesToMicroAlternative,
            AddFeaturesToMacroAlternative, UploadToEnterprise, EditScenario, ReverseFluxlineDirection,
            ImportSBAS, ExportToZip, ImportFromZip, DeleteAlternative
        ]


class BaseSBAS(object):
    """Base class for all SBAS tools."""

    def __init__(self):
        """Tool initialization"""
        super(BaseSBAS, self).__init__()
        self.label = ''
        self.description = ''
        self.canRunInBackground = False
        self.workspace = arcpy.env.workspace
        self.addOutputsToMap = arcpy.env.addOutputsToMap

        # Logging
        script_name = 'SBAS'
        self._logger = logging.getLogger(__name__)
        self._initlogging(r'{0}\logs\{1}-{2}.log'.format(sbas_dir, script_name,
                                                         str(datetime.date.today())))
        sys.excepthook = self._myexcepthook
        self.sb = SBAS()

    def getParameterInfo(self):
        """Create parameter definitions"""
        return None

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def execute(self, parameters, messages):
        """The code that is run when the user runs the tool."""
        self.sb.log_params(parameters)
        # arcpy.env.addOutputsToMap = True
        ArcObjects().activate_data_frame('SBAS')

    def __enter__(self):
        return self

    def __exit__(self):
        arcpy.env.addOutputsToMap = self.addOutputsToMap
        self._logger.removeHandler(logging.StreamHandler(sys.stdout))

    def _initlogging(self, file_path):
        """Start logging to specified file_path"""

        d = os.path.dirname(file_path)
        if not os.path.isdir(d):
            os.makedirs(d)

        logging.basicConfig(
            filename=file_path
            , format='%(asctime)s - %(levelname)s - %(message)s'
            , level=logging.INFO
            , filemode='a'
        )

        # add a channel for standard out
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        self._logger.addHandler(ch)

    def _myexcepthook(self, extype, exval, trace):
        """Exception hook to capture and log all exception stack traces"""
        self._logger.error('Logging an exception', exc_info=(extype, exval, trace))
        arcpy.AddMessage(traceback.format_exception(extype, exval, trace))
        traceback.print_exception(extype, exval, trace)

    def isLicensed(self):
        """Get whether tool is licensed to execute."""
        return True

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return


class ChangeSymbology(BaseSBAS):
    """Tool to alter the symbology of an alternative based on a chosen field."""

    def __init__(self):
        """Tool initialization"""
        super(ChangeSymbology, self).__init__()
        self.label = '05a - Change Symbology'
        self.description = """Will alter the symbology of an alternative based 
            on a chosen field."""
        self.canRunInBackground = False
        self.fields = {
            'Residual Volume': 'residualVolume'
            , 'Delta Volume': 'deltaVolume'
            , 'Placement Volume': 'placementVolume'
            , 'Removal Volume': 'removalVolume'
            , 'Confidence': 'valConfidence'
        }

    def getParameterInfo(self):
        """Parameter definitions"""
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative to reclassify'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_loaded_scenarios()
        val = self.sb.get_current_scen_name

        if val:
            scenario.value = val

        field = arcpy.Parameter(
            displayName='Field under which to classify'
            , name='field'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        field.filter.type = 'ValueList'
        field.filter.list = self.fields.keys()

        units = arcpy.Parameter(
            displayName='Units'
            , name='units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = 'ValueList'
        units.filter.list = UNITS

        out_layer = arcpy.Parameter(
            displayName='New Layer'
            , name='out_layer'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned = arcpy.Parameter(
            displayName='Unassigned'
            , name='unassigned'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned_fluxes = arcpy.Parameter(
            displayName='Unassigned Fluxes'
            , name='unassigned_fluxes'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        show_unassigned = arcpy.Parameter(
            displayName='Show Unassigned Features'
            , name='show_unassigned'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        show_unassigned.value = False

        fluxes = arcpy.Parameter(
            displayName='Fluxes'
            , name='fluxes'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        show_units = arcpy.Parameter(
            displayName='Show Units on Labels'
            , name='show_units'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        show_units.value = True

        return [scenario, field, units, out_layer, unassigned, unassigned_fluxes, show_unassigned, fluxes, show_units]

    def updateParameters(self, parameters):
        """Updates parameters"""
        # skip updating parameters on subsequent updates
        # for p in parameters:
        #     if p.altered:
        #         return

        parameters[0].filter.list = get.get_loaded_scenarios()
        val = self.sb.get_current_scen_name

        if val:
            parameters[0].value = val

        parameters[2].filter.type = 'ValueList'
        parameters[2].filter.list = UNITS
        parameters[2].value = configuration.get_current_units()

    def execute(self, parameters, messages):
        """Executes the tool"""
        self.sb.log('Executing ChangeSymbology()')
        BaseSBAS.execute(self, parameters, messages)

        scenario = parameters[0].value
        field = parameters[1].value
        units = parameters[2].value
        show_unassigned = parameters[6].value
        show_units = parameters[8].value

        self.sb.log(self.fields[field])
        self.sb.log('\tUpdating current units')
        configuration.set_current_units(units)
        lyr = self.sb.apply_symbology(scenario, units, self.fields[field], show_units)
        flux_lyr = self.sb.apply_symbology('{} Fluxes'.format(scenario), units, 'qValue')
        arcpy.SetParameter(3, lyr)
        arcpy.SetParameter(7, flux_lyr)

        if show_unassigned:
            # Load unassigned cells and fluxes. Apply symbology to unassigned cells.
            scenario_id = get.get_scenario_id_by_name(scenario)
            load.load_unassigned_cells(scenario_id)
            unassigned_fluxes = load.load_unassigned_fluxes(scenario_id)
            unassigned = self.sb.apply_symbology('Unassigned Cells', units, self.fields[field])
            arcpy.SetParameter(4, unassigned)
            arcpy.SetParameter(5, unassigned_fluxes)


class UpdateResidualVolumes(BaseSBAS):
    """Tool to handle updating residual volumes after new features are added."""

    def __init__(self):
        """Tool initialization"""
        super(UpdateResidualVolumes, self).__init__()
        self.label = 'Update Residual Volumes'
        self.description = """Updates residual volumes for new cells that intersect existing fluxes."""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""
        existing_fluxes = arcpy.Parameter(
            displayName='Existing Fluxes (semicolon-delimited)'
            , name='existing_fluxes'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        flux_class = arcpy.Parameter(
            displayName='Flux Class'
            , name='flux_class'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        new_cell_dictionary = arcpy.Parameter(
            displayName='New Cell Dictionary (json)'
            , name='new_cell_dictionary'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        scen_name = arcpy.Parameter(
            displayName='Alternative Name'
            , name='scen_name'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        return [existing_fluxes, flux_class, new_cell_dictionary, scen_name]

    def execute(self, parameters, messages):
        """Runs the tool."""
        BaseSBAS.execute(self, parameters, messages)
        existing_fluxes, flux_class, new_cells, scen_name = \
            parameters[0].value, parameters[1].value, parameters[2].value, parameters[3].value

        existing_fluxes = json.loads(existing_fluxes)
        new_cells = json.loads(new_cells)
        self.sb.calc_new_cell_volumes(existing_fluxes, flux_class, new_cells, scen_name)
        arcpy.RefreshActiveView()


class ConvertGraphics(BaseSBAS):
    """Tool to convert graphics objects into SBAS features."""

    def __init__(self):
        """Tool initialization"""
        super(ConvertGraphics, self).__init__()
        self.label = '02a - Convert Graphics to New Features'
        self.description = """Will convert selected graphics to SBAS features,
            then remove the graphics."""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""

        feature_name = arcpy.Parameter(
            displayName='Feature Name'
            , name='feature_name'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
        )

        scenario = arcpy.Parameter(
            displayName='Add Features to Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = []

        micro = arcpy.Parameter(
            displayName='Base Micro Alternative'
            , name='micro'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
            , enabled=False
        )

        convert = arcpy.Parameter(
            displayName='Convert Graphics'
            , name='convert'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        convert.filter.type = 'ValueList'
        convert.filter.list = ['Convert All Graphics', 'Convert Only Selected Graphics']
        convert.value = convert.filter.list[0]

        clear = arcpy.Parameter(
            displayName='Clear All Graphics'
            , name='clear_all'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
        )
        clear.filter.type = 'ValueList'
        clear.filter.list = ['Clear All Graphics', 'Clear Only Selected Graphics',
                             'Do Not Clear Graphics']
        clear.value = clear.filter.list[0]

        out_layer = arcpy.Parameter(
            displayName='New Layer'
            , name='out_layer'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        return [feature_name, scenario, micro, convert, clear, out_layer]

    def updateParameters(self, parameters):
        """
        Updates parameters.
        :param parameters: list of arcpy parameters
        """
        if not parameters[1].altered:
            parameters[1].filter.type = 'ValueList'
            l = get.get_loaded_scenarios()
            l.append('(None)')
            parameters[1].filter.list = l
            val = self.sb.get_current_scen_name

            if val:
                parameters[1].value = val

        if parameters[1].value not in [None, '(None)'] and not parameters[2].altered:
            budget_type = get.get_budget_type(get.get_scenario_id_by_name(parameters[1].value))
            if budget_type.strip().lower() == 'macro':
                parameters[2].filter.type = 'ValueList'
                parameters[2].filter.list = get.get_loaded_scenarios(budget_type='micro')
                parameters[2].enabled = True
            else:
                parameters[2].enabled = False

        if parameters[1].value == '(None)':
            mxd = arcpy.mapping.MapDocument('CURRENT')
            if len(arcpy.mapping.ListLayers(mxd, 'Unassigned*')) == 0:
                raise ValueError('Unassigned layers required if you are not adding features to an alternative. '
                                 + 'Run Change Symbology or Load Alternative tool to load unassigned layers.')

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: arcpy parameters from the tool
        :param messages: any messages generated by the tool
        """
        self.sb.log('Executing ConvertGraphics()')
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        feat_name = parameters[0].value
        if feat_name == 'None':
            feat_name = None

        scen_name = parameters[1].value
        convert = parameters[3].value
        clear = parameters[4].value

        scen_id = get.get_scenario_id_by_name(scen_name)
        budget_type = get.get_budget_type(scen_id)

        macro = '{}'.format(budget_type).lower().strip() == 'macro'
        micro_budget = parameters[2].value

        if macro and micro_budget is None:
            raise ValueError('Base Micro Alternative required for Macro alternatives.')

        convert_all = False
        if convert == 'Convert All Graphics':
            convert_all = True

        if scen_name == '(None)':
            scen_name = None

        scen, new_flux, existing_flux_list, new_cells, macro_scen_id = self.sb.graphic_conversion_worker(
            scen_name, feat_name, convert_all, clear, micro_budget
        )
        self.sb.log('Conversion worker done')
        self.sb.log('\t{}\n\t{}\n\t{}\n\t{}\n\t{}'.format(scen, new_flux, existing_flux_list, new_cells, macro_scen_id))

        if new_cells is not None or new_flux:
            flux_class = '{} Fluxes'.format(scen_name)

            # arcpy does not handle list or dictionary parameters, because apparently
            # it's designed by people who never use python...
            self.sb.log('Getting JSON objects')
            existing_fluxes = json.dumps(existing_flux_list)
            new_cells_json = json.dumps(new_cells)

            self.sb.log('reload_sbas_layers("{0}")'.format(macro_scen_id))
            load.reload_sbas_layers(scen_name, 'Unassigned Cells', scen, macro_id=macro_scen_id)

            if len(new_cells) > 0:
                # Ensure the cell volumes are recalculated if they intersect existing fluxes
                up_volumes = UpdateResidualVolumes()
                p = up_volumes.getParameterInfo()
                p[0].value = existing_fluxes
                p[1].value = flux_class
                p[2].value = new_cells_json
                p[3].value = scen_name
                up_volumes.execute(p, messages)

            if new_flux:
                load.reload_sbas_layers(flux_class, 'Unassigned Fluxes', scen)

            arcpy.RefreshActiveView()


class UpdateFluxScenario(BaseSBAS):
    """A tool to update flux values for alternatives."""

    def __init__(self):
        """Tool initialization"""
        super(UpdateFluxScenario, self).__init__()
        self.label = '04b - Update Flux Alternative Values'
        self.description = 'Will adjust values for a flux alternative.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Scenario to update'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'

        cur_scen_list = get.get_scenario_names()
        mxd = arcpy.mapping.MapDocument('CURRENT')
        layers = arcpy.mapping.ListLayers(mxd)

        scenario.filter.list = [lyr.name for lyr in layers if lyr.name in cur_scen_list]

        feat_name = arcpy.Parameter(
            displayName='Feature Name'
            , name='feat_name'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        q_val = arcpy.Parameter(
            displayName='Q value (decimal)'
            , name='q_val'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
            , enabled=False
        )

        confidence = arcpy.Parameter(
            displayName='Confidence'
            , name='confidence'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
            , enabled=False
        )
        confidence.filter.type = 'ValueList'
        confidence.filter.list = ['High', 'Medium', 'Low', 'No Data']

        units = arcpy.Parameter(
            displayName='Units'
            , name='units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = 'ValueList'
        units.filter.list = UNITS

        unassigned_fluxes = arcpy.Parameter(
            displayName='Unassigned Fluxes'
            , name='unassigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        assigned_fluxes = arcpy.Parameter(
            displayName='Assigned Fluxes'
            , name='assigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        assigned_cells = arcpy.Parameter(
            displayName='Assigned Cells'
            , name='assigned_cells'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        feat_desc = arcpy.Parameter(
            displayName='Feature Description'
            , name='feat_desc'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
        )

        return [scenario, feat_name, q_val, confidence, units, unassigned_fluxes,
                assigned_fluxes, assigned_cells, feat_desc]

    def updateParameters(self, parameters):
        """Updates parameters"""

        # Bail if we're executing after form load
        for p in parameters:
            if p.altered:
                return

        self.sb.log('Updating parameters for UpdateFluxScenario()')

        # if we know the scenario a user has loaded, select it
        cur_scen_list = get.get_scenario_names()
        mxd = arcpy.mapping.MapDocument('CURRENT')
        layers = arcpy.mapping.ListLayers(mxd)

        parameters[0].filter.list = [lyr.name for lyr in layers if lyr.name in cur_scen_list]

        scen_name = self.sb.get_current_scen_name

        if scen_name:
            parameters[0].value = scen_name

        scen_name = parameters[0].value
        flux_class = scen_name + ' Fluxes'

        for p in parameters:
            p.enabled = True

        dir_id = get.get_selected_feature_id(
            flux_class, 'Unassigned Fluxes', 'sedimentTransportDirectionIDPK')

        # Get the flux feature name
        ft_name = get.get_flux_feature_name(dir_id)
        parameters[1].value = ft_name

        self.sb.log('\tSelected fluxline id: {0}'.format(dir_id))
        scen = get.get_scenario_by_name(scen_name, dir_id=dir_id)
        self.sb.log('\tFluxes found: {0}'.format(len(scen['flux'])))

        if len(scen['flux']) == 0:
            parameters[2].value = 0
            parameters[3].value = 'No Data'
            return

        parameters[4].value = configuration.get_current_units()
        units = parameters[4].value

        # Display flux scenario's qValue
        parameters[2].value = scen['flux'][0]['qValue']
        if 'Meters' in units:
            parameters[2].value = to_cubic_meters(parameters[2].value)

        if not parameters[3].value:
            confidence = scen['flux'][0]['valConfidence']
            if confidence not in ['High', 'Medium', 'Low', 'No Data']:
                confidence = 'No Data'

            parameters[3].value = confidence

        arcpy.env.workspace = ws = configuration.get_workspace()
        table = '{0}\\SedimentTransportDirection'.format(ws)
        where = build_where_clause(table, 'sedimentTransportDirectionIDPK', dir_id)
        parameters[8].value = da.get_single_value_from_table(table, 'sdsFeatureDescription', where)

    def execute(self, parameters, messages):
        """Runs the tool."""
        self.sb.log('Executing UpdateFluxScenario()')
        BaseSBAS.execute(self, parameters, messages)

        scen_name = parameters[0].value
        feat_name = parameters[1].value
        q_val = parameters[2].value
        confidence = parameters[3].value
        units = parameters[4].value
        feat_desc = parameters[8].value

        configuration.set_current_units(units)

        if 'Meters' in units:
            q_val = to_cubic_yards(q_val)

        flux_class = '{} Fluxes'.format(scen_name)
        new, scen, cells = self.sb.update_flux_values(
            scen_name=scen_name
            , flux_class=flux_class
            , q_val=q_val
            , confidence=confidence
            , sdsfeaturename=feat_name
            , sdsfeaturedesc=feat_desc
        )

        arcpy.SetParameter(7, cells)

        if new:
            load.reload_sbas_layers(flux_class, 'Unassigned Fluxes', scen)
            arcpy.RefreshActiveView()


class UpdateLittoralScenario(BaseSBAS):
    """A tool to update placement and removal volumes for alternatives."""

    def __init__(self):
        """Tool initialization"""
        super(UpdateLittoralScenario, self).__init__()
        self.label = '04a - Update Littoral Alternative Values'
        self.description = 'Will adjust values for a littoral alternative.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Scenario to update'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'

        cur_scen_list = get.get_scenario_names(budget_type='micro')
        mxd = arcpy.mapping.MapDocument('CURRENT')
        layers = arcpy.mapping.ListLayers(mxd)

        scenario.filter.list = [lyr.name for lyr in layers if lyr.name in cur_scen_list]

        delta = arcpy.Parameter(
            displayName='Delta volume'
            , name='delta'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
            , enabled=False
        )

        placement = arcpy.Parameter(
            displayName='Placement value'
            , name='placement'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
            , enabled=False
        )

        removal = arcpy.Parameter(
            displayName='Removal value'
            , name='removal'
            , datatype='GPDouble'
            , parameterType='Required'
            , direction='Input'
            , enabled=False
        )

        confidence = arcpy.Parameter(
            displayName='Confidence'
            , name='confidence'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
            , enabled=False
        )
        confidence.filter.type = 'ValueList'
        confidence.filter.list = ['High', 'Medium', 'Low', 'No Data']

        unassigned = arcpy.Parameter(
            displayName='Unassigned Cells'
            , name='unassigned'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        assigned = arcpy.Parameter(
            displayName='Assigned Cells'
            , name='assigned'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        units = arcpy.Parameter(
            displayName='Units'
            , name='units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = 'ValueList'
        units.filter.list = UNITS

        feat_name = arcpy.Parameter(
            displayName='Feature Name'
            , name='feat_name'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        assigned_fluxes = arcpy.Parameter(
            displayName='Assigned Fluxes'
            , name='assigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned_fluxes = arcpy.Parameter(
            displayName='Unassigned Fluxes'
            , name='unassigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        feat_desc = arcpy.Parameter(
            displayName='Feature Description'
            , name='feat_desc'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
        )

        flux_values = arcpy.Parameter(
            displayName='Intersecting Flux Values (For Display Only)'
            , name='flux_values'
            , datatype='GPValueTable'
            , parameterType='Optional'
            , direction='Input'
        )
        flux_values.columns = [
            ['String', 'Flux Line (read only)']
            , ['String', 'Type (read only)']
            , ['Double', 'Q Value (read only)']
        ]

        return [scenario, feat_name, delta, placement, removal, confidence, unassigned,
                assigned, units, assigned_fluxes, unassigned_fluxes, feat_desc, flux_values]

    def updateParameters(self, parameters):
        """Updates parameters"""
        arcpy.env.workspace = ws = configuration.get_workspace()

        # Bail if we're executing after form load
        if parameters[0].value is not None:
            return

        self.sb.log('Updating parameters for UpdateLittoralScenario()')

        # if we know the scenario a user has loaded, select it
        cur_scen_list = get.get_scenario_names(budget_type='micro')
        mxd = arcpy.mapping.MapDocument('CURRENT')
        layers = arcpy.mapping.ListLayers(mxd)

        parameters[0].filter.list = [lyr.name for lyr in layers if lyr.name in cur_scen_list]

        scen_name = self.sb.get_current_scen_name

        if scen_name:
            parameters[0].value = scen_name

        scen_name = parameters[0].value

        units = parameters[8].value
        if not units:
            units = configuration.get_current_units()
        parameters[8].value = units

        confidence = parameters[5].value
        if confidence is None:
            confidence = 'No Data'
        parameters[5].value = confidence

        for p in parameters:
            p.enabled = True

        # get selected littoral cell
        lit_id = get.get_selected_feature_id(scen_name, 'Unassigned Cells', 'littoralCellIDPK')

        self.sb.log('\tSelected littoral cell id: {0}'.format(lit_id))
        scen = get.get_scenario_by_name(scen_name, lit_id)
        self.sb.log('\tCells found: {0}'.format(len(scen['cell'])))

        cell_name = parameters[1].value
        if not parameters[1].altered:
            # Get the littoral cell name from the database
            cell_name = get.get_cell_feature_name(lit_id)
        parameters[1].value = cell_name

        scen_id = scen['scenario']['scenarioIDPK']
        if len(scen['cell']) == 0:
            # Create cell scenario with 0 as default values
            self.sb.log('\tCreating cell under {0}'.format(scen_id))
            statics.new_littoral_scenario(scen_id, lit_id, confidence)
            scen = get.get_scenario_by_name(scen_name, lit_id)

        # keep track of whether the user is editing values
        # or whether we need to load values from disk.
        for i in (2, 3, 4):
            if parameters[i].value not in (None, ''):
                return

        parameters[8].value = configuration.get_current_units()
        units = parameters[8].value

        for k, v in scen['cell'][0].iteritems():
            self.sb.log('\tLoading field {0}\tVal {1}'.format(k, v))

        # Display cell scenario's values
        keys = ('deltaVolume', 'placementVolume', 'removalVolume')
        for i, k in enumerate(keys):
            val = scen['cell'][0][k]
            if 'Meters' in units:
                # Store as cubic yards in the database. Use a display expression to display as cubic meters
                # on the map.
                val = to_cubic_meters(val)

            parameters[i + 2].value = val

        if not parameters[5].value:
            confidence = scen['cell'][0]['valConfidence']
            if confidence not in ['High', 'Medium', 'Low', 'No Data']:
                confidence = 'No Data'

            parameters[5].value = confidence

        # get description
        table = '{0}\\SedimentManagementCell'.format(ws)
        where = build_where_clause(table, 'littoralCellIDPK', lit_id)
        parameters[11].value = da.get_single_value_from_table(table, 'sdsFeatureDescription', where)

        # Load flux values for display
        flux_values = []
        flux_data = get.get_source_and_sink_fluxes(lit_id, scen_id)
        for src in flux_data['source']:
            val = src[2]
            if 'Meters' in units:
                val = to_cubic_meters(val)
            flux_values.append([src[1], 'Source', val])
        for snk in flux_data['sink']:
            val = snk[2]
            if 'Meters' in units:
                val = to_cubic_meters(val)
            flux_values.append([snk[1], 'Sink', val])

        parameters[12].values = flux_values

    def execute(self, parameters, messages):
        """Runs the tool."""
        self.sb.log('Executing UpdateLittoralScenario()')
        BaseSBAS.execute(self, parameters, messages)
        scen_name = parameters[0].value
        units = parameters[8].value
        sdsfeaturename = parameters[1].value
        sdsfeaturedesc = parameters[11].value

        if not parameters[5].value:
            parameters[5].value = 'No Data'

        # Get values.
        vals = {
            'deltaVolume': parameters[2].value
            , 'placementVolume': parameters[3].value
            , 'removalVolume': parameters[4].value
            , 'valConfidence': parameters[5].value
        }

        # save volumes in cubic yards
        if 'Meters' in units:
            for k, v in vals.iteritems():
                if k.endswith('Volume'):
                    vals[k] = to_cubic_yards(v)

        configuration.set_current_units(units)

        # First check the alternative FC for matches, if and only if one record is selected
        scen, new = self.sb.update_cell_values(
            scen_name=scen_name,
            vals=vals,
            sdsfeaturename=sdsfeaturename,
            sdsfeaturedesc=sdsfeaturedesc
        )

        # if the selected littoral is not currently in the alternative,
        # we will need to re-load the alternative so it will be displayed.
        if new:
            # re-load the alternative
            self.sb.log('\tCell is not in current alternative. Adding cell.')
            result = statics.load_scenario(scen_name)
            arcpy.SetParameter(6, result[0])
            arcpy.SetParameter(7, result[1])
            arcpy.SetParameter(9, result[3])
            arcpy.SetParameter(10, result[2])

        # Ensure the cells are properly classified based on recent changes
        mxd = arcpy.mapping.MapDocument('CURRENT')
        df = arcpy.mapping.ListDataFrames(mxd, 'SBAS')[0]
        l = arcpy.mapping.ListLayers(mxd, scen_name, df)
        lyr = l[0]

        self.sb.log('\tReclassifying layer {0}'.format(lyr.name))

        # Determine the symbology field. If an alternative has not
        # been loaded or symobology changed since arc was loaded,
        # we will have to do some guesswork to determine the symbology
        # field, due to arc not having a consistent way to identify
        # the field. (valueField is only available for gradient symbology,
        # not unique values.)
        search_val = self.sb.get_search_val(lyr)

        # Change symbology as needed
        ch = ChangeSymbology()
        p = ch.getParameterInfo()
        val_field = None

        for key, val in ch.fields.iteritems():
            if val == search_val:
                val_field = key
                break

        # scenario, field, units, out_layer, unassigned
        unassigned = None
        show_unassigned = len(arcpy.mapping.ListLayers(mxd, 'Unassigned Cells')) > 0
        if show_unassigned:
            unassigned = arcpy.mapping.Layer('Unassigned Cells')

        p[0].value = scen_name
        p[1].value = val_field
        p[2].value = units
        p[3].value = lyr
        p[4].value = unassigned
        p[5].value = show_unassigned

        ch.execute(p, messages)


class CreateScenario(BaseSBAS):
    """A tool to create an SBAS alternative"""

    def __init__(self):
        """Tool initialization"""
        super(CreateScenario, self).__init__()
        self.label = '01a - Create SBAS Alternative'
        self.description = 'Will initialize a new SBAS alternative.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""

        out_table = arcpy.Parameter(
            displayName='New Table View'
            , name='out_table'
            , datatype='GPTableView'
            , parameterType='Derived'
            , direction='Output'
        )

        out_table2 = arcpy.Parameter(
            displayName='New Table View 2'
            , name='out_table2'
            , datatype='GPTableView'
            , parameterType='Derived'
            , direction='Output'
        )

        out_table3 = arcpy.Parameter(
            displayName='New Table View 3'
            , name='out_table3'
            , datatype='GPTableView'
            , parameterType='Derived'
            , direction='Output'
        )

        budget_type = arcpy.Parameter(
            displayName='Budget Type'
            , name='budget_type'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        budget_type.filter.type = 'ValueList'
        budget_type.filter.list = ['Micro', 'Macro']
        budget_type.value = 'Micro'

        scen_name = arcpy.Parameter(
            displayName='Alternative Name'
            , name='scen_name'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        # Parameter is required for metadata purposes
        scen_desc = arcpy.Parameter(
            displayName='Alternative Description'
            , name='scen_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        # metadata again
        area_desc = arcpy.Parameter(
            displayName='Geographic Area Name or Description'
            , name='area_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        author = arcpy.Parameter(
            displayName='Author'
            , name='scenarioAuthor'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        author.value = architecture.get_full_name_of_current_user()

        date_from = arcpy.Parameter(
            displayName='Date From'
            , name='date_from'
            , datatype='GPDate'
            , parameterType='Required'
            , direction='Input'
        )

        date_thru = arcpy.Parameter(
            displayName='Date Through'
            , name='date_thru'
            , datatype='GPDate'
            , parameterType='Required'
            , direction='Input'
        )

        scen_id = arcpy.Parameter(
            displayName='Scenario ID [not user-editable]'
            , name='scen_id'
            , datatype='GPString'
            , parameterType='Derived'
            , direction='Output'
        )
        scen_id.enabled = False

        return [out_table, out_table2, out_table3, budget_type, scen_name,
                scen_desc, author, area_desc, date_from, date_thru, scen_id]

    def updateParameters(self, parameters):
        """
        Updates parameters and provides input validation.
        :param parameters: parameter array
        :raise ValueError: 'Through Date must be after From Date.'
        """
        if parameters[8].altered and parameters[9].altered:
            if parameters[8].value > parameters[9].value:
                raise ValueError('Through Date must be after From Date.')

    def execute(self, parameters, messages):
        """Executes the tool"""
        BaseSBAS.execute(self, parameters, messages)

        budget_type = parameters[3]
        scen_name = parameters[4]
        scen_desc = parameters[5]
        author = parameters[6]
        area_desc = parameters[7]
        df = parameters[8]
        dt = parameters[9]
        scen_id = parameters[10]

        date_from, date_thru = None, None

        if df.value:
            date_from = datetime.datetime.date(df.value)
        if dt.value:
            date_thru = datetime.datetime.date(dt.value)

        statics.new_scenario(budget_type.value, scen_name.value, scen_desc.value, author.value,
                             date_from, date_thru, area_desc.value, scen_id.value)


class EditScenario(BaseSBAS):
    """A tool to edit values in an SBAS alternative"""

    def __init__(self):
        """Tool initialization"""
        super(EditScenario, self).__init__()
        self.label = '01e - Edit SBAS Alternative'
        self.description = 'Will edit values for an existing alternative'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""

        edit_alt = arcpy.Parameter(
            displayName='Edit Alternative'
            , name='edit_alt'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        scen_name = arcpy.Parameter(
            displayName='Alternative Name'
            , name='scen_name'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        scen_desc = arcpy.Parameter(
            displayName='Alternative Description'
            , name='scen_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        author = arcpy.Parameter(
            displayName='Author'
            , name='scenarioAuthor'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
        )

        date_from = arcpy.Parameter(
            displayName='Date From'
            , name='date_from'
            , datatype='GPDate'
            , parameterType='Required'
            , direction='Input'
        )

        date_thru = arcpy.Parameter(
            displayName='Date Through'
            , name='date_thru'
            , datatype='GPDate'
            , parameterType='Required'
            , direction='Input'
        )

        area_desc = arcpy.Parameter(
            displayName='Geographic Area Name or Description'
            , name='area_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        last_scen = arcpy.Parameter(
            displayName='Last Scen'
            , name='last_scen'
            , datatype='GPString'
            , parameterType='Derived'
            , direction='Output'
        )

        out_cell = arcpy.Parameter(
            displayName='Out Cell Layer'
            , name='out_cell'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        out_flux = arcpy.Parameter(
            displayName='Out Flux Layer'
            , name='out_flux'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        return [edit_alt, scen_name, scen_desc, author, date_from, date_thru, area_desc, last_scen, out_cell, out_flux]

    def updateParameters(self, parameters):
        """
        Updates parameters and provides input validation.
        :param parameters: parameter array
        :raise ValueError: 'Through Date must be after From Date.'
        """
        if parameters[4].altered and parameters[5].altered:
            if parameters[4].value > parameters[5].value:
                raise ValueError('Through Date must be after From Date.')

        if not parameters[0].altered:
            parameters[0].filter.type = 'ValueList'
            parameters[0].filter.list = get.get_loaded_scenarios('Micro')
            parameters[0].value = self.sb.get_current_scen_name

        if parameters[7].value is None or parameters[0].value != parameters[7].value:
            # Load existing values from database and display on form
            scen_id = get.get_scenario_id_by_name(parameters[0].value)
            vals = get.get_scenario_values(scen_id)
            fields = ('sdsFeatureName', 'sdsFeatureDescription', 'scenarioAuthor', 'dateIntervalStart',
                      'dateIntervalEnd', 'geoAreaDescription')
            for index, field in enumerate(fields):
                parameters[index + 1].value = vals[field]

        parameters[7].value = parameters[0].value

    def execute(self, parameters, messages):
        """
        Run the tool
        :param parameters: list of arcpy parameters
        :param messages: arcpy messages
        """
        BaseSBAS.execute(self, parameters, messages)

        layers = self.sb.edit_scenario(
            original_name=parameters[0].value,
            scen_name=parameters[1].value,
            scen_desc=parameters[2].value,
            author=parameters[3].value,
            date_from=parameters[4].value,
            date_thru=parameters[5].value,
            area_desc=parameters[6].value
        )

        for layer in layers:
            if 'fluxes' in layer.name.lower():
                arcpy.SetParameter(8, layer)
            else:
                arcpy.SetParameter(7, layer)


class LoadAlternative(BaseSBAS):
    """Will load an alternative and any associated cells/fluxes."""

    def __init__(self):
        """Tool initialization"""
        super(LoadAlternative, self).__init__()
        self.label = '01b - Load SBAS Alternative'
        self.description = 'Will load an alternative and any associated cells/fluxes.'
        self.canRunInBackground = False
        self.fields = {
            'Residual Volume': 'residualVolume'
            , 'Delta Volume': 'deltaVolume'
            , 'Placement Volume': 'placementVolume'
            , 'Removal Volume': 'removalVolume'
            , 'Confidence': 'valConfidence'
        }

    def getParameterInfo(self):
        """Parameter definitions"""
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        load_table_name = arcpy.Parameter(
            displayName='Alternative to load'
            , name='load_table_name'
            , datatype='GPString'
            , parameterType='Derived'
            , direction='Output'
        )

        field = arcpy.Parameter(
            displayName='Field under which to classify'
            , name='field'
            , datatype='GPString'
            , parameterType='Derived'
            , direction='Output'
            , enabled=False
        )
        field.value = 'Delta Volume'

        units = arcpy.Parameter(
            displayName='Units'
            , name='units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = 'ValueList'
        units.filter.list = UNITS

        out_layer = arcpy.Parameter(
            displayName='New Layer'
            , name='out_layer'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned = arcpy.Parameter(
            displayName='Unassigned Cells'
            , name='unassigned'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        scenario = arcpy.Parameter(
            displayName='Alternative to load'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_scenario_names()

        remove = arcpy.Parameter(
            displayName='Remove Other Alternatives from Map'
            , name='remove'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        remove.value = False

        assigned = arcpy.Parameter(
            displayName='Assigned Cells'
            , name='assigned'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned_fluxes = arcpy.Parameter(
            displayName='Unassigned Fluxes'
            , name='unassigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        assigned_fluxes = arcpy.Parameter(
            displayName='Assigned Fluxes'
            , name='assigned_fluxes'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        show_unassigned = arcpy.Parameter(
            displayName='Show Unassigned Features'
            , name='show_unassigned'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        show_unassigned.value = False

        show_units = arcpy.Parameter(
            displayName='Show Units on Labels'
            , name='show_units'
            , datatype='GPBoolean'
            , parameterType='Optional'
            , direction='Input'
        )
        show_units.value = True

        return [load_table_name, field, units, out_layer, scenario, remove,
                unassigned, assigned, unassigned_fluxes, assigned_fluxes, show_unassigned, show_units]

    def updateParameters(self, parameters):
        """
        Update the available scenario name list.
        :param parameters: arcpy toolbox parameter list
        """
        if not parameters[4].altered:
            parameters[4].filter.list = get.get_scenario_names()

    def execute(self, parameters, messages):
        """Executes the tool"""
        BaseSBAS.execute(self, parameters, messages)

        reclassify_field = parameters[1].value
        units = parameters[2].value
        scenario = parameters[4].value
        remove = parameters[5].value
        show_unassigned = parameters[10].value
        show_units = parameters[11].value

        configuration.set_current_units(units)

        # Remove other scenarios, if the user chose to.
        if remove:
            self.sb.remove_other_scenarios(scenario)

        result = list(statics.load_scenario(scenario, show_unassigned))

        # save resulting feature layers as output
        for i in range(4):
            if result[i] is not None:
                arcpy.SetParameter(i + 6, result[i])

        ch = ChangeSymbology()
        p = ch.getParameterInfo()
        p[0].value = scenario
        p[1].value = reclassify_field
        p[2].value = units
        p[3].value = parameters[7].value
        p[4].value = parameters[6].value
        p[5].value = parameters[8].value
        p[6].value = show_unassigned
        p[8].value = show_units

        ch.execute(p, messages)


class SetUpEnvironment(BaseSBAS):
    """Tool to load tables and basemap for SBAS."""

    def __init__(self):
        """Tool initialization"""
        super(SetUpEnvironment, self).__init__()
        self.label = '00  - Set Up SBAS Environment'
        self.description = 'Will load tables and basemap for an SBAS data frame.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""

        basemap = arcpy.Parameter(
            displayName='Basemap to load'
            , name='basemap'
            , datatype='GPString'
            , parameterType='Optional'
            , direction='Input'
            , enabled=True
        )
        basemap.filter.type = 'ValueList'
        basemap.filter.list = ['Shoreline', 'Counties', 'None']
        basemap.value = 'None'

        workspace = arcpy.Parameter(
            displayName='Default Workspace'
            , name='workspace'
            , datatype='DEWorkspace'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        workspace.filter.list = ['Local Database']
        workspace.value = '{}\\SBAS_Data.gdb'.format(sbas_dir)

        out_layer = arcpy.Parameter(
            displayName='Output Layer'
            , name='out_layer'
            , datatype='GPLayer'
            , parameterType='Derived'
            , direction='Output'
            , enabled=True
        )

        meta_city = arcpy.Parameter(
            displayName='Metadata: City'
            , name='meta_city'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        meta_state = arcpy.Parameter(
            displayName='Metadata: State'
            , name='meta_state'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        meta_org = arcpy.Parameter(
            displayName='Metadata: Organization'
            , name='meta_org'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        return [basemap, workspace, out_layer, meta_city, meta_state, meta_org]

    def updateParameters(self, parameters):
        """Updates parameter information"""
        for i, p in enumerate(parameters):
            if i > 2 and p.altered:
                return

        try:
            states = json.loads(open('{}\\templates\\states.json'.format(sbas_dir)).read())
        except (IOError, OSError):
            raise Exception('Failed to load list of U.S. states from file.')

        # ensure we have our config file
        cfg_file = '{}\\sbas.cfg'.format(sbas_dir)
        cfg = configuration.get_config(cfg_file)

        meta_city, meta_state, meta_org = parameters[3], parameters[4], parameters[5]
        meta_city.value = cfg.get('metadata', 'city')
        meta_state.filter.type = 'ValueList'
        meta_state.filter.list = states
        meta_state.value = cfg.get('metadata', 'state')
        meta_org.value = cfg.get('metadata', 'organization')

    def execute(self, parameters, messages):
        """Executes the tool"""
        self.sb.log('Executing SetUpEnvironment()')
        BaseSBAS.execute(self, parameters, messages)

        # set items that need to be written to the cfg file
        workspace, meta_city, meta_state, meta_organization = \
            parameters[1].value, parameters[3].value, parameters[4].value, parameters[5].value

        # ensure we have our config file
        cfg_file = '{}\\sbas.cfg'.format(sbas_dir)
        cfg = configuration.get_config(cfg_file)

        cfg.read(cfg_file)
        cfg.set('sbas', 'workspace', workspace)
        cfg.set('metadata', 'city', meta_city)
        cfg.set('metadata', 'state', meta_state)
        cfg.set('metadata', 'organization', meta_organization)
        configuration.write_config(cfg_file, cfg)

        basemap = unicode(parameters[0].value)
        workspace = unicode(parameters[1].value)

        arcpy.env.workspace = workspace
        self.workspace = arcpy.env.workspace
        self.sb.log('New workspace: {}'.format(arcpy.env.workspace))

        lyr = self.sb.set_up_environment(basemap, workspace, dataframe='SBAS')

        if lyr is not None:
            arcpy.SetParameter(2, lyr)


class ImportGenCade(BaseSBAS):
    """Tool to convert GenCade data into SBAS features."""

    def __init__(self):
        """Tool initialization"""
        super(ImportGenCade, self).__init__()
        self.label = '02b - Convert GenCade data to SBAS Features'
        self.description = 'Will convert GenCade data to SBAS features'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Parameter definitions"""
        debug = False
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Add Features to Micro Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_loaded_scenarios('micro')

        in_shapefile = arcpy.Parameter(
            displayName='GenCade Shapefile'
            , name='in_shapefile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        )
        in_shapefile.filter.list = ['shp']

        in_gen = arcpy.Parameter(
            displayName='GenCade GEN File'
            , name='in_gen'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        )
        in_gen.filter.list = ['gen']

        in_prt = arcpy.Parameter(
            displayName='GenCade PRT File'
            , name='in_prt'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        )
        in_prt.filter.list = ['prt']

        assigned_cells = arcpy.Parameter(
            displayName='Assigned Cells'
            , name='assigned_cells'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        assigned_fluxes = arcpy.Parameter(
            displayName='Assigned Fluxes'
            , name='assigned_fluxes'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned_cells = arcpy.Parameter(
            displayName='Unassigned Cells'
            , name='unassigned_cells'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        unassigned_fluxes = arcpy.Parameter(
            displayName='Unassigned Fluxes'
            , name='unassigned_fluxes'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        if debug:
            base_dir = r'C:\Users\k5opjeas\Documents\ArcGIS\Example_GenCade_SBAS'
            in_shapefile.value = base_dir + r'\updated_MB_shape.shp'
            in_gen.value = base_dir + r'\v1r6_98-04_jan15.gen'
            in_prt.value = base_dir + r'\v1r6_98-04_jan15.prt'

        return [scenario, in_shapefile, in_gen, in_prt,
                assigned_cells, assigned_fluxes, unassigned_cells,
                unassigned_fluxes]

    def updateParameters(self, parameters):
        """
        Get updated list of loaded scenarios.
        :param parameters: arcpy toolbox parameter list
        """
        if not parameters[0].altered:
            parameters[0].filter.list = get.get_loaded_scenarios('micro')
            val = self.sb.get_current_scen_name

            if val:
                parameters[0].value = val

    def execute(self, parameters, messages):
        """Executes the tool"""
        self.sb.log('Executing ImportGenCade()')
        BaseSBAS.execute(self, parameters, messages)

        scenario = parameters[0].value
        shapefile = unicode(parameters[1].value)
        genfile = unicode(parameters[2].value)
        prtfile = unicode(parameters[3].value)

        out = imports.import_gencade(genfile, prtfile, scenario, shapefile)
        for i, val in enumerate(out):
            if val is not None:
                arcpy.SetParameter(i + 4, val)


class AddFeaturesToMicroAlternative(BaseSBAS):
    """Tool to add features to an alternative"""

    def __init__(self):
        """Tool initialization"""
        super(AddFeaturesToMicroAlternative, self).__init__()
        self.label = '03a - Add Selected Features to Micro Alternative'
        self.description = 'Will ensure selected features are assigned to the given alternative.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_loaded_scenarios(budget_type='micro')

        return [scenario]

    def updateParameters(self, parameters):
        """
        Ensure tool updates with the latest list of loaded scenarios.
        :param parameters: list of arcpy parameters.
        """
        if not parameters[0].altered:
            parameters[0].filter.list = get.get_loaded_scenarios(budget_type='micro')

        if len(parameters[0].filter.list) == 0:
            raise Exception('Must load an alternative first.')

    def execute(self, parameters, messages):
        """Executes the tool"""
        self.sb.log('Executing AddFeaturesToMicroAlternative()')
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = parameters[0].value
        scen_id = get.get_scenario_id_by_name(scenario)
        flux_class = '{} Fluxes'.format(scenario)

        selected = self.sb.add_features_to_alternative(scenario)

        # Ensure any cells intersecting the fluxes are updated
        statics.update_intersecting_residuals(flux_class, scenario, selected['SedimentManagementCell'])

        load.reload_sbas_layers(scenario, 'Unassigned Cells', scen_id=scen_id)
        load.reload_sbas_layers('{} Fluxes'.format(scenario), 'Unassigned Fluxes', scen_id=scen_id)

        arcpy.RefreshActiveView()


class AddFeaturesToMacroAlternative(BaseSBAS):
    """Tool to add features to an alternative"""

    def __init__(self):
        """Tool initialization"""
        super(AddFeaturesToMacroAlternative, self).__init__()
        self.label = '03b - Add Selected Features to Macro Alternative'
        self.description = 'Will ensure selected features are assigned to the given alternative.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        macro = arcpy.Parameter(
            displayName='Macro Alternative'
            , name='macro'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        macro.filter.type = 'ValueList'
        macro.filter.list = get.get_loaded_scenarios(budget_type='macro')

        micro = arcpy.Parameter(
            displayName='Base Micro Alternative'
            , name='micro'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        micro.filter.type = 'ValueList'
        micro.filter.list = get.get_loaded_scenarios(budget_type='micro')

        out = arcpy.Parameter(
            displayName='Out Layer'
            , name='out'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        return [macro, micro, out]

    def updateParameters(self, parameters):
        """
        Ensure tool updates with the latest list of loaded scenarios.
        :param parameters: list of arcpy parameters.
        """
        if not parameters[0].altered:
            parameters[0].filter.list = get.get_loaded_scenarios(budget_type='macro')

        if not parameters[1].altered:
            parameters[1].filter.list = get.get_loaded_scenarios(budget_type='micro')

        if len(parameters[0].filter.list) == 0 or len(parameters[1].filter.list) == 0:
            raise Exception('Must load an alternative first.')

    def execute(self, parameters, messages):
        """Executes the tool"""
        self.sb.log('Executing AddFeaturesToMacroAlternative()')
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        macro = parameters[0].value
        micro = parameters[1].value

        macro_scen_id = get.get_scenario_id_by_name(macro)
        micro_scen_id = get.get_scenario_id_by_name(micro)

        self.sb.log('\tMacro Scen ID: {}'.format(macro_scen_id))
        self.sb.log('\tMicro Scen ID: {}'.format(micro_scen_id))

        selected = self.sb.get_selected_feature_ids()
        cell_ids = selected['SedimentManagementCell']
        move_features_to_macro(cell_ids)

        # Create the new scenario record
        for macro_cell_id in cell_ids:
            constituent_cells = get.get_constituent_cell_ids(macro_cell_id)
            micro_cell_scen_ids = get.get_cell_scenario_ids(constituent_cells, micro_scen_id)
            statics.insert_macro_scenario_records(
                macro_cell_id, macro_scen_id, micro_scen_id, micro_cell_scen_ids
            )

        # Ensure any cells intersecting the fluxes are updated
        load.reload_sbas_layers(micro, 'Unassigned Cells', scen_id=micro_scen_id, macro_id=macro_scen_id)
        arcpy.RefreshActiveView()

        macro_layer = arcpy.mapping.Layer(macro)
        return macro_layer


class UploadToEnterprise(BaseSBAS):
    """Tool to upload a scenario to the enterprise GDB"""

    def __init__(self):
        """Tool initialization"""
        super(UploadToEnterprise, self).__init__()
        self.label = '06b  - Upload Alternative to Enterprise'
        self.description = 'Will upload alternative\'s features and data to the enterprise geodatabase.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_loaded_scenarios()

        return [scenario]

    def updateParameters(self, parameters):
        """
        Ensure tool updates with the latest list of loaded scenarios.
        :param parameters: list of arcpy parameters.
        """
        if not parameters[0].altered:
            parameters[0].filter.list = get.get_loaded_scenarios()

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: list
        :param messages: unicode or str
        """
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario_name = parameters[0].value
        scenario_id = get.get_scenario_id_by_name(scenario_name)

        # get bounding box
        mxd = arcpy.mapping.MapDocument(self.sb.map_document)
        cell_lyr = arcpy.mapping.ListLayers(mxd, scenario_name)[0]
        flux_lyr = arcpy.mapping.ListLayers(mxd, '{0} Fluxes'.format(scenario_name))[0]
        bounds = get.get_bounding_coords(cell_lyr, flux_lyr)

        self.sb.log('Updating metadata record with bounding box')
        metadata.build_sbas_metadata(scenario_id, coords=bounds)

        # Prod
        destination = r'{0}\sbas_prod_user.sde'.format(os.path.dirname(__file__))
        destination_schema = 'ERD_GEO_SBAS'

        # Dev
        # destination = r'{0}\BlueBird_SBAS.sde'.format(os.path.dirname(__file__))
        # destination_schema = 'SBAS'

        self.sb.upload_to_enterprise(scenario_id, destination, destination_schema)


class ReverseFluxlineDirection(BaseSBAS):
    """Tool to reverse the direction of selected fluxlines"""

    def __init__(self):
        """Tool initialization"""
        super(ReverseFluxlineDirection, self).__init__()
        self.label = '05b - Reverse Fluxline Direction'
        self.description = 'Will reverse the direction of selected fluxlines.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_loaded_scenarios()

        out_layer = arcpy.Parameter(
            displayName='Flux Layer'
            , name='out_layer'
            , datatype='GPFeatureLayer'
            , parameterType='Derived'
            , direction='Output'
        )

        return [scenario, out_layer]

    def updateParameters(self, parameters):
        """
        Ensure tool updates with the latest list of loaded scenarios.
        :param parameters: list of arcpy parameters.
        :raises ValueError: Must select at least one fluxline in the alternative.
        """
        if parameters[0].value in (None, '(None)'):
            parameters[0].filter.list = get.get_loaded_scenarios()
        else:
            # ensure we have at least one fluxline selected
            flux_class = '{0} Fluxes'.format(parameters[0].value)
            desc_feature = arcpy.Describe(flux_class)
            obj_ids = desc_feature.FIDset.split('; ')
            if len(obj_ids) == 0 or obj_ids[0] == u'':
                raise ValueError('Must select at least one fluxline in the alternative.')

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: list
        :param messages: unicode or str
        """
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = parameters[0].value
        flux_class = '{0} Fluxes'.format(scenario)

        mxd = arcpy.mapping.MapDocument('CURRENT')
        df = arcpy.mapping.ListDataFrames(mxd, 'SBAS')[0]
        out_layer = arcpy.mapping.ListLayers(mxd, flux_class, df)
        arcpy.FlipLine_edit(flux_class)
        statics.update_intersecting_residuals(flux_class, scenario)
        arcpy.SetParameter(1, out_layer)


class ImportSBAS(BaseSBAS):
    """Tool to import legacy SBAS alternatives."""

    def __init__(self):
        """Tool initialization"""
        super(ImportSBAS, self).__init__()
        self.label = '01c - Import Legacy SBAS Alternative'
        self.description = 'Will import a legacy SBAS alternative.'
        self.canRunInBackground = False
        self.workspace = arcpy.env.workspace

    def getParameterInfo(self):
        """Create parameter definitions"""

        shapefiles = arcpy.Parameter(
            displayName='Legacy SBAS Shapefiles'
            , name='shapefiles'
            , datatype='DEFile'
            , multiValue=True
            , parameterType='Required'
            , direction='Input'
        )
        shapefiles.filter.list = ['shp']

        flux_shapefile = arcpy.Parameter(
            displayName='Fluxline Shapefile'
            , name='flux_shapefile'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
        )
        flux_shapefile.filter.list = ['shp']

        author = arcpy.Parameter(
            displayName='Original Author'
            , name='scenarioAuthor'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        # Parameter is required for metadata purposes
        scen_desc = arcpy.Parameter(
            displayName='Alternative Description'
            , name='scen_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        units = arcpy.Parameter(
            displayName='Original Units'
            , name='units'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )
        units.filter.type = 'ValueList'
        units.filter.list = UNITS

        area_desc = arcpy.Parameter(
            displayName='Geographic Area Name or Description'
            , name='area_desc'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
        )

        date_from = arcpy.Parameter(
            displayName='Date From (if not provided in legacy data)'
            , name='date_from'
            , datatype='GPDate'
            , parameterType='Optional'
            , direction='Input'
        )

        date_thru = arcpy.Parameter(
            displayName='Date Through (if not provided in legacy data)'
            , name='date_thru'
            , datatype='GPDate'
            , parameterType='Optional'
            , direction='Input'
        )

        return [shapefiles, author, scen_desc, units, area_desc, date_from, date_thru]

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        if parameters[5].value is not None and parameters[6].value is not None:
            if parameters[5].value > parameters[6].value:
                raise ValueError('Through Date must be after From Date.')

    def execute(self, parameters, messages):
        """Executes the tool"""

        shapefiles = [val.value for val in parameters[0].values]

        author, scen_desc, units, area_desc = \
            parameters[1].value, parameters[2].value, parameters[3].value, parameters[4].value

        scen_name, scen_id, date_from, date_thru = None, None, None, None
        self.sb.log('Reading metadata values from legacy cell shapefile')

        # find the cell shapefile, as it will have the start/end dates
        for shapefile in shapefiles:
            self.sb.log('Reading fields from {0}'.format(os.path.basename(shapefile)))
            list_fields = [field.name for field in arcpy.ListFields(shapefile)]

            if 'END_DATE' in list_fields:
                fields = ('ALT_ID', 'ALT_NAME', 'START_DATE', 'END_DATE')
                for row in da.get_rows(shapefile, fields):
                    # strip curly braces from scen_id
                    scen_id = uuid_to_string(row[0])
                    scen_name = row[1]
                    date_from = row[2]
                    date_thru = row[3]
                    del row
                    break
            elif 'Alt_ID' in list_fields:
                fields = ('Alt_ID', 'Alt_Name')
                for row in da.get_rows(shapefile, fields):
                    # strip curly braces from scen_id
                    scen_id = uuid_to_string(row[0])
                    scen_name = row[1]
                    del row
                    break

        if date_from is not None and isinstance(date_from, (datetime.datetime, datetime.date)):
            if date_from.year < 1800:
                date_from = parameters[5].value
        else:
            date_from = parameters[5].value
                
        if date_thru is not None and isinstance(date_thru, (datetime.datetime, datetime.date)):
            if date_thru.year < 1800:
                date_thru = parameters[6].value
        else:
            date_thru = parameters[6].value

        if date_from is None or date_thru is None:
            raise ValueError('From/Through date not available in legacy alternative, and no value provided.')

        for field in (scen_name, scen_id, date_from, date_thru):
            self.sb.log('Found:\t{0}\t{1}'.format(field, type(field)))

        self.sb.log('Executing CreateScenario() for {0} {1}'.format(scen_id, scen_name))

        cs = CreateScenario()
        cs_params = cs.getParameterInfo()
        cs_params[3].value = 'Micro'
        cs_params[4].value = scen_name
        cs_params[5].value = scen_desc
        cs_params[6].value = author
        cs_params[7].value = area_desc
        cs_params[8].value = date_from
        cs_params[9].value = date_thru
        cs_params[10].value = scen_id
        cs.execute(cs_params, messages)

        for shapefile in shapefiles:
            self.sb.log('Importing {0}'.format(os.path.basename(shapefile)))
            imports.import_sbas_shapefile(shapefile, units)


class ExportToZip(BaseSBAS):
    """Tool to export a scenario to a ZIP GDB"""

    def __init__(self):
        """Tool initialization"""
        super(ExportToZip, self).__init__()
        self.label = '06a - Export Alternative to ZIP'
        self.description = 'Will export a file geodatabase of the alternative, then ZIP it.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_scenario_names()

        out_dir = arcpy.Parameter(
            displayName='Output Directory'
            , name='out_dir'
            , datatype='DEFolder'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )

        out_file = arcpy.Parameter(
            displayName='Output Filename'
            , name='out_file'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )

        return [scenario, out_dir, out_file]

    def updateParameters(self, parameters):
        """
        Ensure tool updates with the latest list of loaded scenarios.
        :param parameters: list of arcpy parameters.
        """
        if not parameters[0].altered:
            parameters[0].filter.list = get.get_scenario_names()

        if parameters[2].altered:
            if not architecture.valid_windows_file(parameters[2].value):
                raise ValueError('Invalid filename.')

        if parameters[0].altered and not parameters[2].altered:
            parameters[2].value = 'SBAS-{0}.zip'.format(parameters[0].value.replace(' ', '_'))

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: list
        :param messages: unicode or str
        """
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scen_name = parameters[0].value
        out_file = '{0}\\{1}'.format(parameters[1].value, parameters[2].value)
        scen_id = get.get_scenario_id_by_name(scen_name)
        self.sb.log('Exporting {0} to {1}...'.format(scen_name, out_file))
        self.sb.export_alternative_zip(scen_id, out_file)


class ImportFromZip(BaseSBAS):
    """Tool to import scenarios from a ZIP GDB"""

    def __init__(self):
        """Tool initialization"""
        super(ImportFromZip, self).__init__()
        self.label = '01d  - Import Modern SBAS Alternatives'
        self.description = 'Used to import non-legacy SBAS alternatives from a ZIP file.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        zip_file = arcpy.Parameter(
            displayName='ZIP File'
            , name='zip_file'
            , datatype='DEFile'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        zip_file.filter.list = ['zip']

        return [zip_file]

    def updateParameters(self, parameters):
        """
        :param parameters: list of arcpy parameters.
        """
        if parameters[0].altered:
            path = str(parameters[0].value)
            if not os.path.isfile(path):
                raise OSError('File not found.')

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: list
        :param messages: unicode or str
        """
        BaseSBAS.execute(self, parameters, messages)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        zip_file = str(parameters[0].value)
        self.sb.log(zip_file)
        self.sb.import_alternative_zip(zip_file)


class DeleteAlternative(BaseSBAS):
    """Tool to import scenarios from a ZIP GDB"""

    def __init__(self):
        """Tool initialization"""
        super(DeleteAlternative, self).__init__()
        self.label = '07  - Delete SBAS Alternative'
        self.description = 'Removes the SBAS Alternative from the local workspace.'
        self.canRunInBackground = False

    def getParameterInfo(self):
        """
        Parameter definitions
        :rtype : list
        :return: list of parameters for tool
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenario = arcpy.Parameter(
            displayName='Alternative'
            , name='scenario'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )
        scenario.filter.type = 'ValueList'
        scenario.filter.list = get.get_scenario_names('Micro')

        warning = arcpy.Parameter(
            displayName='The alternative will be gone FOREVER. If you understand, type "DELETE" in the box below.'
            , name='warning'
            , datatype='GPString'
            , parameterType='Required'
            , direction='Input'
            , enabled=True
        )

        return [scenario, warning]

    def updateParameters(self, parameters):
        """
        :param parameters: list of arcpy parameters.
        """
        pass

    def execute(self, parameters, messages):
        """Executes the tool
        :param parameters: list
        :param messages: unicode or str
        """
        BaseSBAS.execute(self, parameters, messages)

        warning = parameters[1].value
        if warning != 'DELETE':
            raise ValueError('You must confirm deletion by typing "DELETE" on the form.')

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scen_name = parameters[0].value
        scen_id = get.get_scenario_id_by_name(scen_name)
        self.sb.delete_alternative(scen_id)
