#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
"""Module to contain architectural functions for use within the SBAS package."""

import arcpy
import ctypes
import shutil
import errno
import zipfile
import os
import re
import random
import collections

__author__ = 'k5opjeas'


def row_dictionary(row, fields):
    """
    Given a row and an ordered list of fields, return a row dicitonary.
    :param row: row values
    :param fields: field names to use as keys
    :return: dictionary of fields and row values
    """
    result = dict()
    for i, field in enumerate(fields):
        result[field] = row[i]
    return result


def row_namedtuple(row, fields, name):
    """
    Make a namedtuple from a row and an ordered list of fields.
    :param row: row values
    :param fields: field names
    :param name: name to give namedtuple type
    :return: namedtuple
    """
    names = ' '.join(fields)
    t = collections.namedtuple(name, names)
    return t(*row)


def build_where_clause(table, field, value):
    """
    Constructs a SQL WHERE clause to select rows having the specified value
    within a given field and table.
    :param table: name of table to query
    :param field: name of field to query
    :param value: value to match
    :return: a SQL WHERE clause for the table, field, and value
    :rtype : unicode
    """
    assert isinstance(table, (unicode, str))
    assert isinstance(field, (unicode, str))

    # Add DBMS-specific field delimiters
    delimited = arcpy.AddFieldDelimiters(table, field)

    # Determine field type
    field_type = str(arcpy.ListFields(table, field)[0].type)

    # Add single-quotes for string field values
    if field_type == 'String':
        value = "'{0}'".format(value)

    # Format WHERE clause
    return "{0} = {1}".format(delimited, value)


def build_in_clause(table, field, values, in_clause=True):
    """
    Build a valid IN clause for a table, field, and values.
    :param table: table to query
    :param field: field to query
    :param values: list or tuple of values for which to build an IN clause.\
    :param in_clause: whether to return an IN clause. If false, return a NOT IN clause.
    :return: IN clause matching the given criteria, with the values as strings.
    """
    assert isinstance(table, (unicode, str))
    assert isinstance(field, (unicode, str))
    assert isinstance(values, (tuple, list))
    assert isinstance(in_clause, bool)

    # Add DBMS-specific field delimiters
    delimited = arcpy.AddFieldDelimiters(table, field)

    if len(values) > 0:
        # We are going to assume all values are the same type.
        # If they aren't, the query wouldn't work anyway. :)
        if isinstance(values[0], (unicode, str)):
            where = "'" + "','".join(values) + "'"
        else:
            where = ','.join(str(v) for v in values)
    else:
        where = "''"

    if in_clause:
        return u'{0} in ({1})'.format(delimited, where)
    else:
        return u'{0} not in ({1})'.format(delimited, where)


def to_cubic_yards(volume):
    """Given a volume in cubic meters, returns value in cubic yards.
    :param volume: volume in cubic meters
    :returns volume in cubic yards
    :rtype : float
    """
    assert isinstance(volume, (float, int))
    return volume / 0.764554858


def to_cubic_meters(volume):
    """
    Given a volume in cubic yards, returns value in cubic meters.
    :param volume: volume in cubic yards
    :return: volume in cubic meters
    :rtype : float
    """
    assert isinstance(volume, (float, int))
    return volume * 0.764554858


def get_full_name_of_current_user():
    """
    Gets the full name of the current user in Windows. Non-portable.
    :rtype : unicode
    :return: full name of the current user
    """
    getusernameex = ctypes.windll.secur32.GetUserNameExW
    namedisplay = 3

    size = ctypes.pointer(ctypes.c_ulong(0))
    getusernameex(namedisplay, None, size)

    namebuffer = ctypes.create_unicode_buffer(size.contents.value)
    getusernameex(namedisplay, namebuffer, size)
    return namebuffer.value


def calculate_residual_volume(source_q, sink_q, delta, placement, removal):
    """Return the residual volume calculated from its component values.
    :param source_q: sum of all source flux Q values
    :param sink_q: sum of all sink flux Q values
    :param delta: delta volume
    :param placement: placement volume
    :param removal: removal volume
    :return: calculated residual volume
    :rtype : float
    """
    # sum(source flux q vals) - sum(sink q values) - deltaVolume + placementVolume - removalVolume
    for param in (source_q, sink_q, delta, placement, removal):
        assert isinstance(param, (float, int))
    return source_q - sink_q - delta + placement - removal


def uuid_to_string(in_uuid):
    """
    Convert a UUID4 (GUID) to the same string format used in SBAS.
    :param in_uuid: input UUID
    :return: string formatted UUID
    :rtype : str
    """
    return str(in_uuid)[1:-1].lower()


def silently_remove_dir(path):
    """Attempt to remove the directory. If the directory is not found, silently ignore the exception.
    Using this pattern will prevent race conditions wherein the directory may be created between the time
    we check for its existence and the time when we need to overwrite the directory.
    See http://stackoverflow.com/a/10840586
    :param path: str path to remove
    """
    try:
        shutil.rmtree(path)
    except OSError as e:
        # ENOENT = no such file or directory. If this is the error code returned,
        # silently ignore the error.
        if e.errno != errno.ENOENT:
            raise


def silently_remove_file(path):
    """Attempt to remove the file. If the file is not found, silently ignore the exception.
    Using this pattern will prevent race conditions wherein the file may be created between the time
    we check for its existence and the time when we need to overwrite the file.
    See http://stackoverflow.com/a/10840586
    :param path: str path to remove
    """
    try:
        os.remove(path)
    except OSError as e:
        # ENOENT = no such file or directory. If this is the error code returned,
        # silently ignore the error.
        if e.errno != errno.ENOENT:
            raise


def zipdir(path, zip_path):
    """
    Zips a directory `path` to ZIP file `zip_path`
    :param path: path of directory to archive
    :param zip_path: path to output ZIP file
    """
    relroot = os.path.abspath(os.path.join(path, os.pardir))
    with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zf:
        for root, dirs, files in os.walk(path):
            zf.write(root, os.path.relpath(root, relroot))
            for f in files:
                filename = os.path.join(root, f)
                if os.path.isfile(filename):
                    arcname = os.path.join(os.path.relpath(root, relroot), f)
                    zf.write(filename, arcname)


def valid_windows_file(filename):
    """
    Returns whether the filename appears to be valid in Windows.
    :param filename: filename to validate
    """
    r = re.compile('^[\w\-. ]+$')
    return r.match(filename) is not None


def random_alpha_string(length=8):
    """
    Return a random alpha string of the given length.
    :param length: length of string to return
    :return:
    """
    return ''.join(random.choice('abcdefghijklmnopqrstuvwxyz') for i in range(length))
