#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
#
"""Contains static functions for getting values"""

import arcpy
import architecture
import configuration
import da
import math

__author__ = 'k5opjeas'


def get_selected_littoral_id(scen_name):
    """
    Return the selected littoral ID, and whether the littoral is new to the alternative.
    :param scen_name: scenario name to search
    :return: littoral ID, whether the littoral is new to the alternative
    :rtype : tuple
    """
    # Relies on .Describe(), so this likely can't be unit tested
    desc = arcpy.Describe(scen_name)
    obj_ids = desc.FIDset.split('; ')
    if len(obj_ids) == 1 and obj_ids[0] != '':
        lit_id = da.get_single_value_from_table(scen_name, 'littoralCellIDPK')
        if lit_id is not None:
            return lit_id, False

    # Now the same for the unassigned cells FC
    un = 'Unassigned Cells'
    desc = arcpy.Describe(un)
    obj_ids = desc.FIDset.split('; ')
    if len(obj_ids) == 1 and obj_ids[0] != '':
        lit_id = da.get_single_value_from_table(un, 'littoralCellIDPK')
        return lit_id, True

    return None, False


def get_loaded_scenarios(budget_type=''):
    """
    Return a list of scenarios that have been loaded.
    :param budget_type: Type of budget to search. 'Micro', 'Macro', or anything else for all.
    :rtype : list
    :return: list of loaded scenarios
    """
    cur_scen_list = get_scenario_names(budget_type)
    mxd = arcpy.mapping.MapDocument('CURRENT')
    layers = arcpy.mapping.ListLayers(mxd)
    scen_list = []
    for lyr in layers:
        if lyr.name in cur_scen_list:
            scen_list.append(lyr.name)
    return scen_list


def get_scenario_names(budget_type=''):
    """
    Return a list of scenario names in the current workspace.
    :param budget_type: Type of budget to search. 'Micro', 'Macro', or anything else for all.
    :rtype : list
    :return: list of scenario names
    """
    arcpy.env.workspace = ws = configuration.get_workspace()

    scenarios = '{0}\\SedimentManagementScenario'.format(ws)
    budget_type = budget_type.lower().strip()

    if budget_type == 'macro':
        where = architecture.build_where_clause('SedimentManagementScenario', 'budgetType', 'Macro')
    elif budget_type == 'micro':
        where = architecture.build_where_clause('SedimentManagementScenario', 'budgetType', 'Micro')
    else:
        where = None

    return da.get_column_as_list(scenarios, 'sdsFeatureName', where)


def get_scenario_id_by_name(scen_name):
    """Get the scenarioIDPK value for a given scenario name.
    :param scen_name: scenario name to search
    :returns scenario ID
    :rtype : unicode
    """
    assert isinstance(scen_name, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(table, 'sdsFeatureName', scen_name)
    return da.get_single_value_from_table(table, 'scenarioIDPK', where)


def get_scenario_name_by_id(scen_id):
    """
    Get the sdsFeatureName for a scenario from its id.
    :rtype : unicode
    :param scen_id: scenario ID to search
    :return: scenario name
    """
    assert isinstance(scen_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    known_col = 'scenarioIDPK'
    desired_col = 'sdsFeatureName'
    where = architecture.build_where_clause(table, known_col, scen_id)
    return da.get_single_value_from_table(table, desired_col, where)


def get_cell_ids_in_scenario(scenario_id):
    """
    Given a scenario ID, return all littoral cell IDs that have matching values in the CellScenarios table.
    :rtype : list
    :param scenario_id: scenario ID to search
    :return: list of littoral cell IDs in the scenario
    """
    assert isinstance(scenario_id, (unicode, str))
    where = '"scenarioIDFK" = \'{0}\''.format(scenario_id)
    table = arcpy.env.workspace + '\\SedimentManagementCellScenario'
    return da.get_column_as_list(table, 'littoralCellIDFK', where)


def get_flux_ids_in_scenario(scenario_id):
    """Given a scenario ID, return all littoral cell IDs that have values
    in the CellScenarios table. List will be of strings.
    :param scenario_id: scenario ID
    :rtype : list
    """
    assert isinstance(scenario_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementFluxScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioIDFK', scenario_id)
    return da.get_column_as_list(table, 'sedimentTransportDirectionIDFK', where)


def get_cell_feature_name(cell_id):
    """
    Return the sdsFeatureName of the cell with the given ID.
    :rtype : unicode
    :param cell_id: littoral cell ID to search
    :return: cell feature name
    """
    assert isinstance(cell_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    field = 'sdsFeatureName'
    table = '{0}\\SedimentManagementCell'.format(ws)
    where = architecture.build_where_clause(table, 'littoralCellIDPK', cell_id)
    return da.get_single_value_from_table(table, field, where)


def get_flux_feature_name(dir_id):
    """
    Return the sdsFeatureName of the fluxline with the given ID.
    :rtype : unicode
    :param dir_id: flux ID to search
    :return: flux feature name
    """
    assert isinstance(dir_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    field = 'sdsFeatureName'
    table = '{0}\\SedimentTransportDirection'.format(ws)
    where = architecture.build_where_clause(table, 'sedimentTransportDirectionIDPK', dir_id)
    return da.get_single_value_from_table(table, field, where)


def get_unassigned_cells_in(scenario_id):
    """
    Generate an IN clause for unassigned cells.
    :rtype : unicode
    :param scenario_id: scenario to search
    :return: IN clause for unassigned cells
    """
    assert isinstance(scenario_id, (unicode, str))
    cell_ids = get_cell_ids_in_scenario(scenario_id)
    if len(cell_ids) > 0:
        where = "'" + "','".join(cell_ids) + "'"
    else:
        where = "''"
    return '"littoralCellIDPK" not in ({0})'.format(where)


def get_unassigned_fluxes_in(scenario_id):
    """
    Generate an IN clause for unassigned fluxes.
    :rtype : unicode
    :param scenario_id: scenario to search
    :return: IN clause for cells in the scenario
    """
    assert isinstance(scenario_id, (unicode, str))
    flux_ids = get_flux_ids_in_scenario(scenario_id)
    if len(flux_ids) > 0:
        where = "'" + "','".join(flux_ids) + "'"
    else:
        where = "''"
    return '"sedimentTransportDirectionIDPK" not in ({0})'.format(where)


def get_assigned_cells_in(scenario_id, key_field='littoralCellIDPK', cell_ids=None):
    """
    Return a partial IN clause for cells in the scenario
    :rtype : unicode
    :param scenario_id: scenario ID to search
    :param key_field: key field to use for the IN clause
    :param cell_ids: list of cell IDs to include in IN clause.
                     If not provided, defaults to all cell IDs in the scenario.
    :return: partial IN clause for assigned cells
    """
    assert isinstance(scenario_id, (unicode, str))
    assert isinstance(key_field, (unicode, str))
    assert cell_ids is None or isinstance(cell_ids, list)
    if cell_ids is None:
        cell_ids = get_cell_ids_in_scenario(scenario_id)

    scen_name = get_scenario_name_by_id(scenario_id)
    return architecture.build_in_clause(scen_name, key_field, cell_ids, in_clause=True)


def get_assigned_fluxes_in(scenario_id, key_field='sedimentTransportDirectionIDPK', flux_ids=None):
    """
    Return a partial IN clause for fluxes in the scenario
    :rtype : unicode
    :param scenario_id: scenario ID to search
    :param key_field: key field to use for the IN clause
    :param flux_ids: list of flux IDs to include in IN clause.
                     If not provided, defaults to all flux IDs in the scenario.
    :return: partial IN clause for assigned fluxes
    """
    assert isinstance(scenario_id, (unicode, str))
    assert isinstance(key_field, (unicode, str))
    if flux_ids is None:
        flux_ids = get_flux_ids_in_scenario(scenario_id)

    scen_name = get_scenario_name_by_id(scenario_id)
    return architecture.build_in_clause('{} Fluxes'.format(scen_name), key_field, flux_ids, in_clause=True)


def get_macro_volumes(macro_cell_id, micro_scen_id):
    """
    For a given macro cell, get the sum of all volumes of any
    constituent micro cells in the micro alternative.
    :param macro_cell_id: the macro cell to search
    :param micro_scen_id: the micro scenario to search
    :return: dictionary of the sums of all constituent volumes
    :rtype : dict
    """
    arcpy.env.workspace = ws = configuration.get_workspace()

    fields = ('deltaVolume', 'placementVolume', 'removalVolume', 'residualVolume')
    table = '{0}\\SedimentManagementCellScenario'.format(ws)

    retval = dict()
    for k in fields:
        retval[k] = 0.0

    micro_cell_ids = get_constituent_cell_ids(macro_cell_id, unassigned=False)

    where = architecture.build_in_clause(table, 'littoralCellIDFK', micro_cell_ids, True)
    where = '{0} and {1}'.format(where, architecture.build_where_clause(table, 'scenarioIDFK', micro_scen_id))

    for row in da.get_rows(table, fields, where):
        for i, val in enumerate(row):
            retval[fields[i]] += val

    return retval


def get_constituent_cell_ids(feature_id, unassigned=False):
    """
    Gets a list of the IDs of all micro cells within a macro cell.
    :rtype : list
    :param feature_id: macro cell ID to search
    :param unassigned: whether to search cells that have not been assigned to Macro
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    constituent = []

    if unassigned:
        table = 'SedimentManagementCell'
        column = 'littoralCellIDPK'
    else:
        table = 'SedimentManagementMacroCell'
        column = 'macroCellIDPK'

    table = '{0}\\{1}'.format(ws, table)
    where = architecture.build_where_clause(table, column, feature_id)
    shape = da.get_single_value_from_table(table, 'SHAPE@', where)

    for scenario in get_loaded_scenarios(budget_type='micro'):
        for row in da.get_rows(scenario, ('SHAPE@', 'littoralCellIDPK')):
            if shape.contains(row[0]):
                constituent.append(row[1])

    return constituent


def get_source_and_sink_cell(line):
    """
    Given a polyline, determine the source and sink cells.
    :param line: arcpy polyline object
    :return: source cell, sink cell
    :rtype : tuple
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    fields = ('SHAPE@', 'littoralCellIDPK')
    source_cell, sink_cell = 0, 0
    table = '{0}\\SedimentManagementCell'.format(ws)

    for row in da.get_rows(table, fields):
        if row[0]:
            if row[0].contains(line.firstPoint):
                source_cell = row[1]
            elif row[0].contains(line.lastPoint):
                sink_cell = row[1]

    return source_cell, sink_cell


def get_cell_scenario_ids(lit_ids, scen_id):
    """
    Return the a list of cell scenario IDs matching a set of cell IDs and a scenario ID.
    :param lit_ids: list of littoral cell IDs
    :param scen_id: scenario ID
    :return: list of matching cell scenario IDs
    :rtype : list
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = architecture.build_in_clause(table, 'littoralCellIDFK', lit_ids, in_clause=True)
    where = '{0} and {1}'.format(where, architecture.build_where_clause(table, 'scenarioIDFK', scen_id))
    return da.get_column_as_list(table, 'scenarioValIDPK', where)


def get_parent_macro_cell_scen_ids(micro_cell_scen_id):
    """
    Find any macro cell scenario IDs for which the micro cell is constituent.
    :param micro_cell_scen_id: micro cell scenario ID
    :return: list of matching macro cell scenario IDs
    :rtype : list
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentMacroLittoral'.format(ws)
    where = architecture.build_where_clause(table, 'microCellScenarioIDFK', micro_cell_scen_id)
    return da.get_column_as_list(table, 'macroCellScenarioIDFK', where)


def get_child_micro_scen_ids(macro_scen_id):
    """
    Return any micro scenario IDs that are constituent to the macro scenario.
    :param macro_scen_id: macro scenario to search
    :return: list of scenario IDs
    :rtype list:
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentMacroLittoral'.format(ws)
    where = architecture.build_where_clause(table, 'macroCellScenarioIDFK', macro_scen_id)
    micro_cell_scen_ids = da.get_column_as_list(table, 'microCellScenarioIDFK', where)
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = architecture.build_in_clause(table, 'scenarioValIDPK', micro_cell_scen_ids, True)
    child_scen_ids = da.get_column_as_list(table, 'scenarioIDFK', where)
    return list(set(child_scen_ids))


def get_source_and_sink_fluxes(cell_id, scenario_id):
    """Determine the source and sink flux data for a given cell in a scenario.
    :rtype : dict
    :param cell_id: littoral cell ID
    :param scenario_id: scenario ID
    :return: dictionary of source and sink fluxes for the cell
    """
    assert isinstance(cell_id, (unicode, str))
    assert isinstance(scenario_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()

    # Get matching cell
    tbl = '{0}\\SedimentManagementCell'.format(ws)
    field = 'littoralCellIDPK'
    where = architecture.build_where_clause(tbl, field, cell_id)
    cell = da.get_single_value_from_table(tbl, 'SHAPE@', where)

    if cell is None:
        return None

    # Find source and sink fluxes
    source, sink = [], []
    fields = ('SHAPE@', 'sedimentTransportDirectionIDPK', 'sdsFeatureName')
    where = get_assigned_fluxes_in(scenario_id)

    tbl = '{0}\\SedimentTransportDirection'.format(ws)
    for row in da.get_rows(tbl, fields, where):
        if cell.contains(row[0].lastPoint):
            srcq = get_q_val(row[1], scenario_id)
            source.append((row[1], row[2], srcq))
        elif cell.contains(row[0].firstPoint):
            snkq = get_q_val(row[1], scenario_id)
            sink.append((row[1], row[2], snkq))

    return {'source': source, 'sink': sink}


def get_sum_of_q_vals(lit_id, scen_id):
    """Return the sum of all source and sink q values associated with a littoral and scenario.
    :param lit_id: littoral cell ID
    :param scen_id: scenario ID
    :return: sum of source q values, sum of sink q values
    :rtype : tuple
    """
    assert isinstance(lit_id, (unicode, str))
    assert isinstance(scen_id, (unicode, str))
    attached_fluxes = get_source_and_sink_fluxes(lit_id, scen_id)
    srcq, snkq = 0.0, 0.0

    if attached_fluxes is not None:
        for source in attached_fluxes['source']:
            srcq += source[2]
        for sink in attached_fluxes['sink']:
            snkq += sink[2]

    return snkq, srcq


def get_scenario_values(scen_id):
    """
    Get existing values from the scenario table.
    :param scen_id: scenario ID to query
    :return: dictionary of field names and their values
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioIDPK', scen_id)
    fields = ('sdsFeatureName', 'sdsFeatureDescription', 'scenarioAuthor', 'dateIntervalStart', 'dateIntervalEnd',
              'dateScenarioCreated', 'sdsMetadataID', 'geoAreaDescription')
    retval = None
    for row in da.get_rows(table, fields, where):
        retval = architecture.row_dictionary(row, fields)
        del row
        break
    return retval


def get_bounding_coords(cell_layer, flux_layer):
    """
    Return a dictionary of the bounding coordinates of the cells, fluxes
    :param cell_layer: cell Layer object
    :param flux_layer: flux Layer object
    :return: bounding coordinates
    :rtype dict:
    """
    coords = dict()

    ecell = cell_layer.getExtent()
    eflux = flux_layer.getExtent()

    # find the farthest west, east, north, south of the two feature classes
    if ecell.XMin < eflux.XMin or math.isnan(eflux.XMin):
        coords['West'] = ecell.XMin
    else:
        coords['West'] = eflux.XMin

    if ecell.XMax > eflux.XMax or math.isnan(eflux.XMax):
        coords['East'] = ecell.XMax
    else:
        coords['East'] = eflux.XMax

    if ecell.YMin < eflux.YMin or math.isnan(eflux.YMin):
        coords['South'] = ecell.YMin
    else:
        coords['South'] = eflux.YMin

    if ecell.YMax > eflux.YMax or math.isnan(eflux.YMax):
        coords['North'] = ecell.YMax
    else:
        coords['North'] = eflux.YMax

    return coords


def get_q_val(flux_id, scen_id):
    """Find the q-value matching the given flux ID and scenario ID.
    :rtype : float
    :param flux_id: flux ID to search for
    :param scen_id: scenario ID to search for
    :return: q value for the flux in the scenario
    """
    assert isinstance(flux_id, (unicode, str))
    assert isinstance(scen_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\sedimentManagementFluxScenario'.format(ws)
    where1 = architecture.build_where_clause(table, 'sedimentTransportDirectionIDFK', flux_id)
    where2 = architecture.build_where_clause(table, 'scenarioIDFK', scen_id)
    where = '{0} and {1}'.format(where1, where2)
    q_val = da.get_single_value_from_table(table, 'qValue', where) or 0.0
    return q_val


def get_budget_type(scen_id):
    """
    Get the budget type of the scenario with the given ID.
    :param scen_id: scenario ID
    :return: type of budget, 'Micro' or 'Macro'
    :rtype : unicode
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioIDPK', scen_id)
    return da.get_single_value_from_table(table, 'budgetType', where)


def get_cell_id_from_cell_scen_id(cell_scen_id):
    """
    Given a cell scenario ID, get the littoralCellID of the cell
    :param cell_scen_id: cell scenario ID
    :return: macro or micro cell ID
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioValIDPK', cell_scen_id)
    return da.get_single_value_from_table(table, 'littoralCellIDFK', where)


def get_scen_id_from_cell_scen_id(cell_scen_id):
    """
    Return the scenario ID matching a cell scenario.
    :param cell_scen_id: unicode or str
    :return: scenario ID
    :rtype unicode:
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioValIDPK', cell_scen_id)
    return da.get_single_value_from_table(table, 'scenarioIDFK', where)


def get_selected_feature_id(feature_class, unassigned, id_field):
    """Ensure we only have one selected flux feature.
    :rtype : unicode
    :param feature_class: name of feature class to search
    :param unassigned: name of unassigned feature class to search
    :param id_field: id field to search
    :return: selected feature ID value
    :raises ValueError: One and only one feature must be selected.
    """
    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    selected_ex = ValueError('More than one feature selected.')

    count = 0
    id_field_val = ''
    desc_feature = arcpy.Describe(feature_class)

    # First check the alternative FC for matches, if and only if one record is selected
    obj_ids = desc_feature.FIDset.split('; ')
    if len(obj_ids) == 1 and obj_ids[0] != '':
        for row in da.get_rows(feature_class, [id_field]):
            id_field_val = row[0]
            count += 1

    if count > 1:
        raise selected_ex

    if id_field_val != '':
        return id_field_val

    # Now the same for the unassigned FC, if present
    if feature_class.endswith('Fluxes'):
        desc = 'Unassigned Fluxes'
    else:
        desc = 'Unassigned Cells'

    mxd = arcpy.mapping.MapDocument('CURRENT')
    if len(arcpy.mapping.ListLayers(mxd, desc)) > 0:
        desc_unassigned = arcpy.Describe(desc)
        count = 0
        obj_ids = desc_unassigned.FIDset.split('; ')
        if len(obj_ids) == 1 and obj_ids[0] != '':
            for row in da.get_rows(unassigned, [id_field]):
                id_field_val = row[0]
                count += 1

    if count > 1:
        raise selected_ex
    elif id_field_val == '':
        raise ValueError('No feature selected')

    return id_field_val


def get_scenario_by_name(scenario_name, lit_id=None, dir_id=None):
    """
    Returns a scenario matching the provided feature name.
    :rtype : dict
    :param scenario_name: scenario name to search
    :param lit_id: optional littoral ID to search
    :param dir_id: optional transport direction ID to search
    :return: scenario dictionary object
    """
    arcpy.env.workspace = ws = configuration.get_workspace()

    result = {
        'scenario': None
        , 'flux': []
        , 'cell': []
    }

    # Scenario needs to be loaded by ID
    arcpy.AddMessage(arcpy.env.workspace)
    arcpy.AddMessage(scenario_name)
    scen_id = get_scenario_id_by_name(scenario_name)

    # scenario
    scenarios = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(scenarios, 'scenarioIDPK', scen_id)
    result['scenario'] = da.row_dictionary_from_table(scenarios, where)
    arcpy.AddMessage(result['scenario'])

    # cells - can have several results, so we will manually search for all children
    cells = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = architecture.build_where_clause(cells, 'scenarioIDFK', scen_id)

    if lit_id:
        where = '{0} and {1}'.format(architecture.build_where_clause(
            cells, 'littoralCellIDFK', lit_id), where)

    fields = [field.name for field in arcpy.ListFields(cells)]
    for row in da.get_rows(cells, fields, where):
        result['cell'].append(architecture.row_dictionary(row, fields))

    # fluxes - same story as before. We need to manually get all children and keep them in a list.
    fluxes = '{0}\\SedimentManagementFluxScenario'.format(ws)
    where = architecture.build_where_clause(fluxes, 'scenarioIDFK', scen_id)

    if dir_id:
        where = '{0} and {1}'.format(architecture.build_where_clause(
            fluxes, 'sedimentTransportDirectionIDFK', dir_id), where)

    fields = [field.name for field in arcpy.ListFields(fluxes)]
    for row in da.get_rows(fluxes, fields, where):
        result['flux'].append(architecture.row_dictionary(row, fields))

    return result


if __name__ == '__main__':
    s = 'Toledo-Sandusky_03Recent'
    scen = get_scenario_by_name(s)
    print(scen['scenario'])
