#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
#
"""Contains static functions for loading data"""

import arcpy
import get
import architecture
import configuration
import da

__author__ = 'k5opjeas'


def load_sbas_layer(layer_name, where, in_table):
    """
    Given a layer name and a where clause, load cells and add them to the MXD.
    :param layer_name: layer name to load
    :param where: where clause of matchin features
    :param in_table: table to search for layer name
    :return: resulting arc layer object
    """
    for p in (layer_name, where, in_table):
        assert isinstance(p, (unicode, str))

    # Add layer to mxd
    # Force arcpy to overwrite any existing layer
    in_layer = '{0}\\{1}'.format(arcpy.env.workspace, in_table)
    arcpy.env.overwriteOutput = True
    arcpy.MakeFeatureLayer_management(in_layer, layer_name, where)
    arcpy.env.overwriteOutput = False
    lyr = arcpy.mapping.Layer(layer_name)
    return lyr    


def load_unassigned_cells(scenario_id):
    """
    Display all cells that do not have values in the scenario
    :param scenario_id: scenario ID to search for unassigned cells
    :return: arc layer object of unassigned cells
    """
    assert isinstance(scenario_id, (unicode, str))
    where = get.get_unassigned_cells_in(scenario_id)
    return load_sbas_layer('Unassigned Cells', where, 'SedimentManagementCell')


def load_unassigned_fluxes(scenario_id):
    """
    Display all fluxlines that do not have values in the scenario
    :param scenario_id: scenario ID to search
    :return: arc layer object of unassigned fluxes
    """
    assert isinstance(scenario_id, (unicode, str))
    where = get.get_unassigned_fluxes_in(scenario_id)
    return load_sbas_layer('Unassigned Fluxes', where, 'SedimentTransportDirection')


def load_assigned_cells(scenario_id):
    """
    Display all cells that have values in the scenario.
    :param scenario_id: scenario ID to search for assigned cells
    :return: arc layer object
    """
    assert isinstance(scenario_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioIDPK', scenario_id)
    budget_type = da.get_single_value_from_table(table, 'budgetType', where)

    if budget_type.lower().strip() == 'micro':
        table = 'SedimentManagementCell'
        key_field = 'littoralCellIDPK'
    else:
        table = 'SedimentManagementMacroCell'
        key_field = 'macroCellIDPK'

    where = get.get_assigned_cells_in(scenario_id, key_field)
    scenario_name = get.get_scenario_name_by_id(scenario_id)
    return load_sbas_layer(scenario_name, where, table)


def load_assigned_fluxes(scenario_id):
    """
    Display all fluxlines associated with the scenario.
    :param scenario_id: scenario ID to search
    :return: arc layer object of assigned fluxes
    """
    assert isinstance(scenario_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = architecture.build_where_clause(table, 'scenarioIDPK', scenario_id)
    budget_type = da.get_single_value_from_table(table, 'budgetType', where)

    if budget_type.lower().strip() == 'micro':
        table = 'SedimentTransportDirection'
        key_field = 'sedimentTransportDirectionIDPK'
    else:
        table = 'SedimentMacroFlux'
        key_field = 'macroTransportDirectionIDPK'

    where = get.get_assigned_fluxes_in(scenario_id, key_field=key_field)
    scenario_name = get.get_scenario_name_by_id(scenario_id)
    lyr = load_sbas_layer('{} Fluxes'.format(scenario_name), where, table)
    lyr.transparency = 50
    lyr.showLabels = True
    return lyr


def reload_sbas_layers(assigned_name, unassigned_name, scen=None, scen_id=None, macro_id=None):
    """
    Update the assigned and unassigned layer definitionQueries.
    :param assigned_name: name of the assigned feature layer on the document
    :param unassigned_name: name of the unassigned feature layer on the document
    :param scen: scenario object
    :param scen_id: scenario ID
    :param macro_id: macro scenario ID, if any
    """
    assert isinstance(assigned_name, (unicode, str))
    assert isinstance(unassigned_name, (unicode, str))
    assert isinstance(scen, dict) or isinstance(scen_id, (unicode, str))

    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    if not scen_id:
        scen_id = scen['scenario']['scenarioIDPK']

    if unassigned_name.endswith('Fluxes'):
        assigned_where = get.get_assigned_fluxes_in(scen_id)
        unassigned_where = get.get_unassigned_fluxes_in(scen_id)
    else:
        assigned_where = get.get_assigned_cells_in(scen_id)
        unassigned_where = get.get_unassigned_cells_in(scen_id)

    if macro_id is None:
        # re-set the assigned definitionquery
        assigned_lyr = arcpy.mapping.Layer(assigned_name)
        assigned_lyr.definitionQuery = assigned_where
    else:
        # update the macro layer definitionquery
        macro_where = get.get_assigned_cells_in(macro_id, 'macroCellIDPK')
        macro_name = get.get_scenario_name_by_id(macro_id)
        macro_lyr = arcpy.mapping.Layer(macro_name)
        macro_lyr.definitionQuery = macro_where

    # and the unassigned definitionquery (if present on map document)
    mxd = arcpy.mapping.MapDocument('CURRENT')
    if len(arcpy.mapping.ListLayers(mxd, 'Unassigned*')) > 0:
        unassigned_lyr = arcpy.mapping.Layer(unassigned_name)
        unassigned_lyr.definitionQuery = unassigned_where
