# vim: set fileencoding=utf-8 :
"""To contain functions/classes for collecting and publishing metadata."""
# Requires the requests HTTP library (included)

import get
import requests
import configuration
from classes import *
from architecture import *
import da
import xml.etree.ElementTree as ETree

__author__ = 'k5opjeas'


def create_metadata(scenario_id, template, **kwargs):
    """
    Creates an new metadata record with the values provided in kwargs, and inserts a record in the metadata
    table containing the metadata and scenario ID. XML string will be compacted before writing.
    :param scenario_id: scenario ID to insert into metadata table
    :type scenario_id: unicode or str
    :param template: template XML file to use
    :type template: unicode or str
    :param kwargs: keyword arguments to populate the XML record
    :type kwargs: dict
    :return: updated XML tree
    :rtype xml.ETree.ElementTree.ElementTree:
    """
    ETree.register_namespace('gmi', 'http://www.isotc211.org/2005/gmi')
    ETree.register_namespace('gco', 'http://www.isotc211.org/2005/gco')
    ETree.register_namespace('gmd', 'http://www.isotc211.org/2005/gmd')
    ETree.register_namespace('xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    ETree.register_namespace('gml', 'http://www.opengis.net/gml')
    ETree.register_namespace('gts', 'http://www.isotc211.org/2005/gts')
    ETree.register_namespace('srv', 'http://www.isotc211.org/2005/srv')
    ETree.register_namespace('xlink', 'http://www.w3.org/1999/xlink')
    meta = ETree.parse(template).getroot()
    meta_str = update_iso_metadata(meta, **kwargs)
    meta_str = compact_metadata_string(meta_str)
    insert_metadata(scenario_id, meta_str)
    return meta


def update_iso_metadata(meta, **kwargs):
    """
    Use string replacement to update ISO metadata.
    :param meta: metadata ETree object
    :param kwargs: Keyword arguments in the form:
        'text to replace': 'value to replace it with'
    :return: upated metadata unicode string
    :rtype unicode:
    """
    meta_str = ETree.tostring(meta, encoding='utf8', method='xml')
    for k, v in kwargs.iteritems():
        if v is None:
            v = u''
        meta_str = meta_str.replace(k, unicode(v))
    return meta_str


def insert_metadata(scenario_id, metadata):
    """
    Insert a new metadata record in the workspace's metadata table, clearing old records if present.
    :param scenario_id: scenario ID
    :type scenario_id: unicode or str
    :param metadata: metadata string
    :type metadata: unicode or str
    """
    table = '{0}\\Metadata'.format(arcpy.env.workspace)
    fields = ('scenarioIDFK', 'metadata')
    where = build_where_clause(table, fields[0], scenario_id)
    row = (scenario_id, metadata)
    da.clear_data(table, where)
    da.insert_row(table, fields, row)


def build_iso_kwargs(scen_id, coords=None):
    """
    Build keyword arguments for ISO metadata
    :param scen_id: scenario ID
    :param coords: bounding coordinate dictionary
    :return: keyword arguments
    :rtype dict:
    """
    scen_vals = get.get_scenario_values(scen_id)

    # Using isoformat() rather than strftime() so we can support dates
    # prior to 1900.
    create_date = scen_vals['dateScenarioCreated'].isoformat().split('T')[0]
    pub_date = datetime.datetime.utcnow().isoformat().split('T')[0]
    begin_date = scen_vals['dateIntervalStart'].isoformat().split('T')[0]
    end_date = scen_vals['dateIntervalEnd'].isoformat().split('T')[0]

    kwargs = {
        'scenario_id': scen_id,
        'template': r'{0}\Templates\SBAS-METADATA-TEMPLATE_MM.XML'.format(sbas_dir),
        '#DATE_METADATA_PUBLISHED#': pub_date,
        '#SCENARIO-SDSMETADATAID#': scen_vals['sdsMetadataID'],
        '#SCENARIO-SCENARIOAUTHOR#': scen_vals['scenarioAuthor'],
        '#SCENARIO-DATESCENARIOCREATED#': create_date,
        '#SCENARIO-SDSFEATURENAME#': scen_vals['sdsFeatureName'],
        '#SCENARIO-SDSFEATUREDESCRIPTION#': scen_vals['sdsFeatureDescription'],
        '#SCENARIO-DATEINTERVALEND#': end_date,
        '#SCENARIO-DATEINTERVALSTART#': begin_date,
        '#SCENARIO-GEOAREADESCRIPTION#': scen_vals['geoAreaDescription'],
        # ensure any records with the old URL are updated with the current URL
        '/SAM/SBAS/MapServer': '/ERDC/Sediment_Budget_Alternatives/MapServer',
        '/ERDC/SBAS/MapServer': '/ERDC/Sediment_Budget_Alternatives/MapServer'
    }
    if coords is not None:
        kwargs['#LATITUDE-NORTH#'] = coords['North']
        kwargs['#LATITUDE-SOUTH#'] = coords['South']
        kwargs['#LONGITUDE-WEST#'] = coords['West']
        kwargs['#LONGITUDE-EAST#'] = coords['East']

    return kwargs


def build_sbas_metadata(scen_id, coords=None):
    """
    Builds an SBAS metadata record from a template.
    :param scen_id: scenario ID
    :param coords: bounding box decimal degree coordinates of the scenario.
    :type coords: dict
    """
    kwargs = build_iso_kwargs(scen_id, coords)
    create_metadata(**kwargs)


def compact_metadata_string(meta_string):
    """
    Recursively remove whitespace from formatted XML strings.
    Slightly naive and probably slow algorithm, but it gets the job done.
    :param meta_string: metadata string to compact
    :type meta_string: unicode or str
    :return: compacted metadata string
    :rtype unicode:
    """
    meta_string = meta_string.replace('\n', '').replace('\r', '').replace('\t', '')
    meta_string = meta_string.replace('> ', '>').replace(' <', '<')
    if '> ' in meta_string or ' <' in meta_string:
        meta_string = compact_metadata_string(meta_string)
    return unicode(meta_string)


def post_to_metadata_manager(meta_string, user, password):
    """
    Write the metadata record to the metadata manager
    :param meta_string: metadata XML as string
    :param user: user ID for metadata manager
    :param password: password for metadata manager
    :return: metadata ID, document ID
    """
    # look for an existing metadata ID, if any
    meta = ETree.fromstring(meta_string)
    gmd = '{http://www.isotc211.org/2005/gmd}'
    gco = '{http://www.isotc211.org/2005/gco}'

    meta_id = meta.find('.//{0}fileIdentifier/{1}CharacterString'.format(gmd, gco))

    if meta_id.text in (None, '#SCENARIO-SDSMETADATAID#'):
        # create a metadata ID to use as the file identifier
        meta_id = '{{{0}}}'.format(uuid.uuid4())
        meta.find('.//{0}fileIdentifier/{1}CharacterString'.format(gmd, gco)).text = meta_id
        meta_string = compact_metadata_string(ETree.tostring(meta))
    else:
        meta_id = meta_id.text

    mgr_url = configuration.get_metadata_url()
    post_url = '{0}/manage/document'.format(mgr_url)
    last_posted_url = '{0}/find/document?f=xjson&searchText='.format(mgr_url)

    r = requests.post(post_url, data=meta_string, auth=(user, password))
    if r.status_code not in (200, 201):
        raise Exception('Error {0}: {1}'.format(r.status_code, r.reason))

    document_id = ''
    received = requests.get('{0}{1}'.format(last_posted_url, meta_id)).json()
    if len(received['features']):
        document_id = received['features'][0]['attributes']['catalog.docuuid']

    return meta_id, document_id


def update_metadata_id(scen_id, meta_id, doc_id):
    """
    Adds the sdsMetadataID to all tables/feature classes.
    Also stores the metadata manager document ID.
    :param scen_id: scenario ID to update
    :param meta_id: sdsMetadataID
    :param doc_id: metadata manager document ID
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    tables = ('Metadata', 'SedimentManagementCell', 'SedimentManagementScenario', 'SedimentTransportDirection')

    for table in tables:
        table = '{0}\\{1}'.format(ws, table)

        if table.endswith('SedimentManagementScenario'):
            where = build_where_clause(table, 'scenarioIDPK', scen_id)
        else:
            where = build_where_clause(table, 'scenarioIDFK', scen_id)

        vals = {'sdsMetadataID': meta_id}
        if table == 'Metadata':
            vals['documentID'] = doc_id

        try:
            da.update_rows(table, vals, where)
        except:
            print table, vals
            raise
