def GetESRIModule(module):
    """Returns the named ESRI module from comtypes.gen,
    creating the wrapper for it if necessary"""
    import sys
    mod = 'comtypes.gen.%s' % module
    try:
        __import__(mod, globals(), locals(), [module], -1)
    except ImportError:
        WrapModule(module)
        __import__(mod, globals(), locals(), [module], -1)
    return sys.modules[mod]

def GetLibPath():
    """Returns the location of the directory containing the ArcGIS object
    library (.olb) files"""
    import _winreg, os.path
    with _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,r"SOFTWARE\ESRI") as esrikey:
        # Lets first look for ArcGIS 10.0+ installations, e.g. HKLM\SOFTWARE\ESRI\Desktop10.0\InstallDir
        try:
            i = 0
            while True:
                subkeyname = _winreg.EnumKey(esrikey, i)
                if subkeyname.startswith("Desktop") or subkeyname.startswith("Engine"):
                    # Take the first one we find
                    with _winreg.OpenKey(esrikey, subkeyname) as subkey:
                        return os.path.join(_winreg.QueryValueEx(subkey, "InstallDir")[0], "com")
                i += 1
        except WindowsError:
            pass
        # Try the pre-10.0 key, e.g. HKLM\SOFTWARE\ESRI\ArcGIS\InstallDir
        try:
            with _winreg.OpenKey(esrikey, "ArcGIS") as subkey:
                return os.path.join(_winreg.QueryValueEx(subkey, "InstallDir")[0], "com")
        except WindowsError:
            pass
    raise WindowsError("ArcGIS InstallDir key not found!")

def WrapModules():
    """Force wrapping of all ArcObjects libraries (OLBs)"""
    import os.path
    import comtypes.client
    from glob import iglob
    comdir = GetLibPath()
    for olb in iglob(os.path.join(comdir, "*.olb")):
        comtypes.client.GetModule(olb)

def WrapModule(module):
    """Wrap a single ArcObjects library (OLB)"""
    import os
    import comtypes.client
    comdir = GetLibPath()
    olb = os.path.join(comdir, module + ".olb")
    comtypes.client.GetModule(olb)

def GetCurrentApp():
    """Returns the Application if the script is being run from
    within the application boundary of an ArcGIS application"""
    esriFramework = GetESRIModule("esriFramework")
    return NewObj(esriFramework.AppRef, esriFramework.IApplication)

def GetApp(app="ArcMap"):
    """Returns the Application if the script is being run from
    outside the application boundary of an ArcGIS application.
    "app" must be 'ArcMap' (default) or 'ArcCatalog'"""
    if not app in ("ArcMap", "ArcCatalog"):
        raise ValueError("app must be 'ArcMap' or 'ArcCatalog'")
    esriFramework = GetESRIModule("esriFramework")
    pAppROT = NewObj(esriFramework.AppROT, esriFramework.IAppROT)
    if pAppROT:
        pApp = None
        appcount = pAppROT.Count
        if appcount > 0:
            for i in range(appcount):
                pApp = pAppROT.Item(i)
                if pApp.Name == app:
                    break
        return pApp
    else:
        raise RuntimeError("AppROT object could not be created!")

def NewObj(MyClass, MyInterface):
    """Creates a new comtypes POINTER object where
    MyClass is the class to be instantiated,
    MyInterface is the interface to be assigned"""
    from comtypes.client import CreateObject
    try:
        ptr = CreateObject(MyClass, interface=MyInterface)
        return ptr
    except:
        return None

def CType(obj, interface):
    """Casts obj to interface and returns comtypes POINTER or None"""
    try:
        newobj = obj.QueryInterface(interface)
        return newobj
    except:
        return None

def CLSID(MyClass):
    """Return CLSID of MyClass as string"""
    return str(MyClass._reg_clsid_)

if __name__ == "__main__":
    import arcpy # Required to initialize ArcObjects (for ArcGIS 10)
    ##import arcgisscripting # Required to initialize ArcObjects (for either ArcGIS 10 or 9)
    pApp = GetCurrentApp() # try to get IApplication from AppRef first, in case we are in-process
    if not pApp:
        print "AppRef not available (script is not running from within an ArcGIS application), attempting AppROT..."
        pApp = GetApp() # defaults to look for ArcMap
    if pApp:
        print "%s HWND: %x" % (pApp.Name, pApp.hWnd)
    else:
        print "No ArcGIS application found!"