#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
"""Module to contain SBAS functions/methods that do not rely on an SBAS instance."""

from sbas import *
import arcpy
import get
import da

__author__ = 'k5opjeas'

sbas_dir = os.path.dirname(os.path.realpath(__file__))


def cell_contains_flux(cell_id, flux_id):
    """
    Return whether the shape with the given cell_id contains the flux with the given flux_id.
    :param cell_id: string cell ID key
    :param flux_id: string flux ID key
    :return: whether the cell contains the flux
    :rtype : bool
    """
    assert isinstance(cell_id, (unicode, str))
    assert isinstance(flux_id, (unicode, str))
    arcpy.env.workspace = ws = configuration.get_workspace()

    table = '{0}\\SedimentManagementCell'.format(ws)
    where = build_where_clause(table, 'littoralCellIDPK', cell_id)
    cell = da.get_single_value_from_table(table, 'SHAPE@', where)

    table = '{0}\\SedimentTransportDirection'.format(ws)
    where = build_where_clause(table, 'sedimentTransportDirectionIDPK', flux_id)
    flux = da.get_single_value_from_table(table, 'SHAPE@', where)

    if cell is None or flux is None:
        return False

    return cell.contains(flux.firstPoint) or cell.contains(flux.lastPoint)


def move_features_to_macro(feature_ids):
    """
    Move the selected features to the MacroCell table.
    :param feature_ids: list of feature IDs to convert into macro cells
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    fields = [
        'SHAPE@', 'sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK',
        'projectID', 'sdsID', 'sdsMetadataID', 'deltaVolume', 'placementVolume', 'removalVolume',
        'valConfidence', 'scenarioIDFK', 'residualVolume'
    ]

    macro_fields = fields + ['macroCellIDPK']
    micro_fields = fields + ['littoralCellIDPK']

    cell = '{0}\\SedimentManagementCell'.format(ws)
    macro = '{0}\\SedimentManagementMacroCell'.format(ws)
    where = build_in_clause(cell, 'littoralCellIDPK', feature_ids, True)

    for row in da.get_rows(cell, micro_fields, where):
        da.insert_row(macro, macro_fields, row)

    da.clear_data(cell, where)


def new_macro_cell_scenario(macro_scen_id, macro_cell_id, cell_scen_ids):
    """
    Insert a new macro cell scenario record.
    :param macro_scen_id: macro scenario ID
    :param macro_cell_id: the macro cell to add to the macro scenario
    :param cell_scen_ids: constituent micro cell scenario IDs
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    fields = (
        'deltaVolume', 'placementVolume', 'removalVolume', 'residualVolume',
        'scenarioValIDPK', 'scenarioIDFK', 'littoralCellIDFK', 'valConfidence'
    )

    cell_scen = '{0}\\SedimentManagementCellScenario'.format(ws)
    macro = '{0}\\SedimentMacroLittoral'.format(ws)
    macro_cell_scen_id = uuid.uuid4()
    row = (0.0, 0.0, 0.0, 0.0, macro_cell_scen_id, macro_scen_id, macro_cell_id, 'No Data')
    da.insert_row(cell_scen, fields, row)

    fields = ('relIDPK', 'macroCellScenarioIDFK', 'microCellScenarioIDFK')
    for micro_cell_scen_id in cell_scen_ids:
        row = (uuid.uuid4(), macro_cell_scen_id, micro_cell_scen_id)
        da.insert_row(macro, fields, row)


def update_cell_scenario(lit_id, scen_id, vals):
    """
    Update the appropriate cell scenario record with the given values.
    :param lit_id: littoral cell or macro cell ID
    :param scen_id: scenario ID
    :param vals: dictionary of volume values
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    where = build_where_clause(table, 'scenarioIDFK', scen_id)
    where = '{0} and {1}'.format(build_where_clause(table, 'littoralCellIDFK', lit_id), where)
    if 'sdsFeatureName' in vals.keys():
        del vals['sdsFeatureName']
    da.update_row(table, vals, where)


def update_cell(cell_scen):
    """
    Update a cell record in the SedimentManagementCell table.
    :param cell_scen: cell scenario dictionary
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    fields = ('scenarioIDFK', 'deltaVolume', 'placementVolume', 'removalVolume',
              'valConfidence', 'residualVolume')

    # Add volumes to our row dictionary
    vals = {i: cell_scen[i] for i in cell_scen if i in fields}
    for f in fields:
        if f not in vals.keys():
            vals[f] = 0.0

    table = '{0}\\SedimentManagementCell'.format(ws)
    where = build_where_clause(table, 'littoralCellIDPK', cell_scen['littoralCellIDFK'])
    da.update_row(table, vals, where)


def set_scenario_values(scen_id, scen_name, scen_desc, author, date_from, date_thru, area_desc):
    """
    Update scenario matching scen_id with the values provided.
    :param scen_id: scenario ID to search
    :param scen_name: updated scenario name
    :param scen_desc: updated description
    :param author: updated author
    :param date_from: updated from date
    :param date_thru: updated through date
    :param area_desc: updated geographic area description
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    table = '{0}\\SedimentManagementScenario'.format(ws)
    where = build_where_clause(table, 'scenarioIDPK', scen_id)
    vals = {
        'sdsFeatureName': scen_name,
        'sdsFeatureDescription': scen_desc,
        'scenarioAuthor': author,
        'dateIntervalStart': date_from,
        'dateIntervalEnd': date_thru,
        'geoAreaDescription': area_desc
    }
    da.update_row(table, vals, where)


def add_volumes_to_feature_classes(scen):
    """
    Add volume data to feature classes so they can be classified by their volumes.
    :param scen: scenario dictionary object
    """
    arcpy.env.workspace = ws = configuration.get_workspace()

    scenario_id = scen['scenario']['scenarioIDPK']

    # first cells
    table = '{0}\\SedimentManagementCell'.format(ws)
    for cell in scen['cell']:
        where = build_where_clause(table, 'littoralCellIDPK', cell['littoralCellIDFK'])
        vals = {
            'scenarioIDFK': scenario_id
            , 'residualVolume': cell['residualVolume']
            , 'deltaVolume': cell['deltaVolume']
            , 'placementVolume': cell['placementVolume']
            , 'removalVolume': cell['removalVolume']
            , 'valConfidence': cell['valConfidence']
        }
        da.update_row(table, vals, where)

    # then fluxes
    table = '{0}\\SedimentTransportDirection'.format(ws)
    for flux in scen['flux']:
        where = build_where_clause(table, 'sedimentTransportDirectionIDPK', flux['sedimentTransportDirectionIDFK'])
        vals = {'qValue': flux['qValue'], 'scenarioIDFK': scenario_id}
        da.update_row(table, vals, where)


def recalculate_cell_residual_volume(scen_id, scen_cell, new=False, quick=False):
    """Update the residual volume value based on its calculation.
    :param quick: whether to skip updating if fluxes are zero
    :param scen_id: scenario ID
    :param scen_cell: cell dictionary from a scenario
    :param new: indicates whether this record already exists in the scenario
    """
    assert isinstance(scen_id, (unicode, str))
    assert isinstance(scen_cell, dict)

    lit_id = scen_cell['littoralCellIDFK']
    snkq, srcq = get.get_sum_of_q_vals(lit_id, scen_id)

    # Do not update database if quick flag is on
    if quick and snkq == 0 and srcq == 0:
        return

    old_vol = scen_cell['residualVolume']

    # Calculate residualVolume as:
    # sum(source flux q vals) - sum(sink q values) - deltaVolume + placementVolume - removalVolume
    res_vol = calculate_residual_volume(
        srcq, snkq, scen_cell['deltaVolume'], scen_cell['placementVolume'],
        scen_cell['removalVolume'])

    scen_cell['residualVolume'] = res_vol

    # create values dictionary for update_cell_scenario()
    vals = dict()
    for k in ('deltaVolume', 'placementVolume', 'removalVolume', 'residualVolume'):
        vals[k] = scen_cell[k]

    if new:
        # insert appropriate record in the cell scenario table
        insert_cell_scen(scen_cell)

    elif res_vol != old_vol:
        # update appropriate record in SedimentManagementCell
        update_cell(scen_cell)

        # update appropriate record in SedimentManagementCellScenario
        update_cell_scenario(lit_id, scen_id, vals)


def update_intersecting_residuals(flux_class, scen_name, gids=None):
    """
    Ensure any cells intersecting selected fluxes are updated
    :param flux_class: name of the flux feature class on the map
    :param scen_name: name of the scenario to update
    :param gids: list or tuple of graphic IDs to query.
    :return:
    """
    where = None
    if gids is not None:
        # self.log('Adding cells to selection')
        where = build_in_clause(scen_name, 'littoralCellIDPK', gids, True)
        arcpy.SelectLayerByAttribute_management(scen_name, 'ADD_TO_SELECTION', where)
    else:
        arcpy.SelectLayerByLocation_management(scen_name, 'INTERSECT', flux_class)

    scen_id = get.get_scenario_id_by_name(scen_name)
    scen = None
    for row in arcpy.SearchCursor(scen_name, where_clause=where):
        cell_id = row.getValue('littoralCellIDPK')
        scen = get.get_scenario_by_name(scen_name, lit_id=cell_id)
        recalculate_cell_residual_volume(scen_id, scen['cell'][0])

    return scen


def recalculate_intersect(intersect, new_cell_dictionary, scen_id):
    """
    Iterate through intersect records and recalculate their residual volumes.
    :param intersect: name of intersect feature class
    :param new_cell_dictionary: values from the cell dictionary
    :param scen_id: scenario ID to update
    """
    for row in da.get_rows(intersect, ['littoralCellIDPK']):
        cell = new_cell_dictionary[row[0]]
        recalculate_cell_residual_volume(scen_id, cell, new=True, quick=False)
        update_cell(cell)


def insert_cell_scen(cell_scen):
    """
    Insert a cell scenario record into SedimentManagementCellScenario
    :param cell_scen: cell scenario dictionary
    """
    arcpy.env.workspace = ws = configuration.get_workspace()
    out_file = '{0}\\SedimentManagementCellScenario'.format(ws)
    fields, row = [], []
    for key, val in cell_scen.iteritems():
        if key != 'units':
            fields.append(key)
            row.append(val)
    da.insert_row(out_file, fields, row)


def new_scenario(budget_type, scen_name, scen_desc, scenarioauthor,
                 date_from, date_thru, area_desc, scen_id=None):
    """
    Create and save a new SBAS scenario. Returns list of scenario table views.
    :param budget_type: 'Micro' or 'Macro'
    :param scen_name: scenario name
    :param scen_desc: scenario description
    :param scenarioauthor: scenario author
    :param date_from: date from
    :param date_thru: date through
    :param area_desc: geographic area description
    :param scen_id: scenario ID GUID string
    """
    # add only the scenario record. The cell and flux scenario records are added within another tool.
    ws = configuration.get_workspace()
    arcpy.env.workspace = ws
    table = '{0}\\SedimentManagementScenario'.format(ws)

    if scen_id is None:
        scen_id = str(uuid.uuid4())

    scen = Scenario(scen_id, scen_name, scen_desc, None, None,
                    None, None, scenarioauthor, datetime.datetime.date(datetime.datetime.today()),
                    budget_type, date_from, date_thru, area_desc, None, None)

    # get a list of fields and their row values
    fields, row = [], []
    for key, val in scen.iteritems():
        fields.append(key)
        row.append(val)
    da.insert_row(table, fields, row)

    metadata.build_sbas_metadata(scen_id)


def new_littoral_scenario(scenario_id, lit_id, confidence='No Data'):
    """
    Creates and saves a new littoral cell scenario record with 0 values.
    :rtype : dict
    :param scenario_id: scenario ID
    :param lit_id: littoral cell ID
    :param confidence: confidence level
    :return: new cell scenario dictionary object
    """
    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    cell_scen_id = str(uuid.uuid4())

    cell_scen = CellScenario(
        scenarioIDFK=scenario_id
        , littoralCellIDFK=lit_id
        , scenarioValIDPK=cell_scen_id
        , valConfidence=confidence
    )

    # insert appropriate record in the cell scenario table
    insert_cell_scen(cell_scen)

    # update appropriate record into the SedimentManagementCell table
    update_cell(cell_scen)
    return cell_scen


def new_flux_scenario(scenario_id, dir_id, confidence='No Data',
                      source=None, sink=None):
    """
    Creates and saves a new flux scenario record with 0 values.
    :rtype : dict
    :param scenario_id: scenario ID
    :param dir_id: transport direction ID
    :param confidence: confidence value
    :param source: source cell ID
    :param sink: sink cell ID
    :return: flux scenario dictionary object
    :raise: exceptions from .da.InsertCursor
    """

    flux_scen_id = str(uuid.uuid4())

    flux_scen = FluxScenario(
        scenarioIDFK=scenario_id
        , qValue=0.0
        , sourceCellIDFK=source
        , sinkCellIDFK=sink
        , sedimentTransportDirectionIDFK=dir_id
        , valConfidence=confidence
        , fluxValuesIDPK=flux_scen_id
    )

    # database operations
    out_file = r'{0}\SedimentManagementFluxScenario'.format(arcpy.env.workspace)

    fields, row = [], []
    for key, val in flux_scen.iteritems():
        fields.append(key)
        row.append(val)
    da.insert_row(out_file, fields, row)

    return flux_scen


def load_scenario(scen_name, show_unassigned=False):
    """
    Load the scenario onto the current map document.
    :rtype : tuple
    :param scen_name: name of scenario to load
    :param show_unassigned: whether to display unassigned cells and fluxes
    :return: unassigned, assigned, unassigned fluxes, assigned fluxes
    """
    assert isinstance(scen_name, (unicode, str))

    arcpy.env.workspace = configuration.get_workspace()

    arcpy.AddMessage(arcpy.env.workspace)

    scen = get.get_scenario_by_name(scen_name)
    scenario_id = scen['scenario']['scenarioIDPK']
    arcpy.AddMessage(scenario_id)

    # Display assigned and unassigned cells
    unassigned, unassigned_fluxes = None, None
    if show_unassigned:
        unassigned = load.load_unassigned_cells(scenario_id)
    assigned = load.load_assigned_cells(scenario_id)

    unassigned_fluxes, assigned_fluxes = None, None
    if show_unassigned:
        unassigned_fluxes = load.load_unassigned_fluxes(scenario_id)
    assigned_fluxes = load.load_assigned_fluxes(scenario_id)

    result = (unassigned, assigned, unassigned_fluxes, assigned_fluxes)

    # Add fields required for displaying correct symbology
    fields = {
        'scenarioIDFK': 'STRING'
        , 'residualVolume': 'DOUBLE'
        , 'deltaVolume': 'DOUBLE'
        , 'placementVolume': 'DOUBLE'
        , 'removalVolume': 'DOUBLE'
        , 'valConfidence': 'STRING'
    }

    for table in ('SedimentManagementCell', 'SedimentManagementMacroCell'):
        existing_fields = [col.name for col in arcpy.ListFields(table)]
        for field, datatype in fields.iteritems():
            if field not in existing_fields:
                arcpy.AddField_management(table, field, datatype)

    # Same for flux qValue
    for table in ('SedimentTransportDirection', 'SedimentMacroFlux'):
        existing_fields = [col.name for col in arcpy.ListFields(table)]

        fields = {
            'qValue': 'DOUBLE',
            'scenarioIDFK': 'STRING'
        }

        for field, datatype in fields.iteritems():
            if field not in existing_fields:
                arcpy.AddField_management(scen_name + ' Fluxes', field, datatype)

    add_volumes_to_feature_classes(scen)

    return result


def recalculate_scenario_residual_volumes(scenario_name, new=False):
    """Recalculate the residual volumes for all cells in a scenario.
    :param scenario_name: name of scenario to update
    :param new: indicates whether this record already exists in the scenario
    """
    assert isinstance(scenario_name, (unicode, str))
    scen = get.get_scenario_by_name(scenario_name)
    scen_id = scen['scenario']['scenarioIDPK']
    for cell in scen['cell']:
        recalculate_cell_residual_volume(scen_id, cell, new)


def get_new_cell_scenario(lit_id, scen):
    """
    Create new flux scenario, and return the scenario
    :param lit_id: littoral cell ID
    :param scen: scenario dictionary
    :rtype : scenario
    :return: updated scenario object
    """
    # Create cell scenario with default values
    scen_name = scen['scenario']['sdsFeatureName']
    scen_id = scen['scenario']['scenarioIDPK']
    new_littoral_scenario(scen_id, lit_id, 'No Data')
    scen = get.get_scenario_by_name(scen_name, lit_id=lit_id)
    return scen


def update_macro_cell_scenario(scen_id, macro_cell_id, volumes):
    """
    Update volume data for a macro cell scenario.
    :param scen_id: scenario ID to update
    :param macro_cell_id: macro cell id
    :param volumes: volume dictionary
    """
    # Update the cell scenario values
    update_cell_scenario(macro_cell_id, scen_id, volumes)

    # Update values in layer on mxd
    scenario_name = get.get_scenario_name_by_id(scen_id)
    mxd = arcpy.mapping.MapDocument('CURRENT')
    if len(arcpy.mapping.ListLayers(mxd, scenario_name)) == 0:
        return

    where = build_where_clause(scenario_name, 'macroCellIDPK', macro_cell_id)
    da.update_rows('SedimentManagementMacroCell', volumes, where)


def insert_macro_scenario_records(macro_cell_id, macro_scen_id, micro_scen_id,
                                  cell_scen_ids):
    """
    Perform insert on cell scenario table, then update the volumes to the correct values.
    :param cell_scen_ids: cell scenario ID list
    :param macro_cell_id: macro cell to update
    :param macro_scen_id: macro scenario to update
    :param micro_scen_id: constituent micro scenario
    """
    new_macro_cell_scenario(macro_scen_id, macro_cell_id, cell_scen_ids)
    volumes = get.get_macro_volumes(macro_cell_id, micro_scen_id)
    update_macro_cell_scenario(macro_scen_id, macro_cell_id, volumes)


def update_associated_macro_cells(micro_scen_val_id, vals):
    """
    Find any macro cells associated with a micro scenario value, and
    update its values based on the micro scenario values.
    :param micro_scen_val_id: micro scenario value ID
    :param vals: dictionary of micro values
    """
    macros = get.get_parent_macro_cell_scen_ids(micro_scen_val_id)
    for macro_cell_scen_id in macros:
        macro_cell_id = get.get_cell_id_from_cell_scen_id(macro_cell_scen_id)
        macro_scen_id = get.get_scen_id_from_cell_scen_id(macro_cell_scen_id)
        update_macro_cell_scenario(macro_scen_id, macro_cell_id, vals)


if __name__ == '__main__':
    s = 'Toledo-Sandusky_03Recent'
    load_scenario(s)
