#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
"""To contain class definitions for SBAS helper classes"""

# Requires the comtypes package: https://pypi.python.org/pypi/comtypes
# Requires slightly modified code snippets provided by Mark Cederholm
# for accessing ArcObjects via Python. (Included)

import uuid
import sys
import inspect
import logging
import traceback
import datetime
import os
import arcpy

# Import code snippets provided by Mark Cederholm for accessing
# ArcObjects via Python. http://www.pierssen.com/arcgis/misc.htm
# scenarioAuthor appears to have provided the snippets for unrestricted use:
#   "Free stuff for GIS Developers and Users!" > http://www.pierssen.com
import python_arcobjects.Snippets as Snippets
# Ensure all Arc desktop and standalone modules are wrapped into Python
Snippets.GetDesktopModules()
Snippets.GetStandaloneModules()
Snippets.GetModule('esriEditor.olb')

import comtypes.gen.esriArcMapUI as esriArcMapUI
import comtypes.gen.esriGeometry as esriGeometry
import comtypes.gen.esriCarto as esriCarto

__author__ = 'k5opjeas'

sbas_dir = os.path.dirname(os.path.realpath(__file__))


class GenCade(dict):
    """Parent class for data from GenCade files."""

    def __init__(self, gc_id, first_id, second_id):
        """Initialize the GenCade data."""
        super(GenCade, self).__init__()

        for name in ['gc_id', 'first_id', 'second_id']:
            self[name] = eval(name)


class GenCadeCell(GenCade):
    """Will store cell information from GenCade files."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, gc_id, first_id, second_id, deltaVolume=0.0, residualVolume=0.0,
                 placementVolume=0.0, removalVolume=0.0):
        """Initialize the GenCade cell.
        :param gc_id: id field for the GenCade cell object
        :param first_id: first GenCade cell ID represented in the object
        :param second_id: second (last) GenCade cell ID represented in the object
        :param deltaVolume: delta volume for the GenCade cell
        :param residualVolume: residual volume for the GenCade cell (not currently transmitted)
        :param placementVolume: placement volume for the GenCade cell (not currently transmitted)
        :param removalVolume: removal volume for the GenCade cell (not currently transmitted)
        """
        super(GenCadeCell, self).__init__(gc_id, first_id, second_id)
        self['deltaVolume'], self['residualVolume'], self['placementVolume'], self['removalVolume'] = \
            deltaVolume, residualVolume, placementVolume, removalVolume

        for name in self.iterkeys():
            self[name] = eval(name)


class GenCadeFlux(GenCade):
    """Will store fluxline information from GenCade files."""

    def __init__(self, gc_id, first_id, second_id, q_left=0.0, q_right=0.0):
        """
        Initializes a GenCade flux object.
        :param gc_id: id field for the GenCade flux object
        :param first_id: first GenCade flux ID represented in the object
        :param second_id: second (last) GenCade flux ID represented in the object
        :param q_left: q value for sediment moving left
        :param q_right: q value for sediment moving right
        """
        super(GenCadeFlux, self).__init__(id, first_id, second_id)
        self['q_left'] = q_left
        self['q_right'] = q_right

        for name in self.iterkeys():
            self[name] = eval(name)


class Littoral(dict):
    """Parent class of littoral cells and fluxes, as a dictionary."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, sdsFeatureName, sdsFeatureDescription=None, mediaIDFK=None, projectID=None,
                 sdsID=None, sdsMetadataID=None):
        """Initialize a littoral dictionary
        :param sdsFeatureName: string name of feature
        :param sdsFeatureDescription: optional description of feature
        :param mediaIDFK: string, not used
        :param projectID: string, not currently used
        :param sdsID: string, not used
        :param sdsMetadataID: string, not used
        """
        super(Littoral, self).__init__()

        for name in ['sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK', 'projectID',
                     'sdsID', 'sdsMetadataID']:
            self[name] = eval(name)


class LittoralCell(Littoral):
    """An instance of a littoral cell, as a dictionary."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, sdsFeatureName, sdsFeatureDescription=None, mediaIDFK=None, projectID=None,
                 sdsID=None, sdsMetadataID=None, littoralCellIDPK=None):
        """Initialize a littoral cell instance.
        :param sdsFeatureName: string name of feature
        :param sdsFeatureDescription: optional description of feature
        :param mediaIDFK: string, not used
        :param projectID: string, not currently used
        :param sdsID: string, not used
        :param sdsMetadataID: string, not used
        :param littoralCellIDPK: string primary key for feature in the geodatabase
        """
        super(LittoralCell, self).__init__(sdsFeatureName)
        self['littoralCellIDPK'] = None

        # succinctly set all dictionary values
        for name in self.iterkeys():
            self[name] = eval(name)


class Flux(Littoral):
    """An instance of a littoral flux, as a dictionary."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, sdsFeatureName, sdsFeatureDescription=None, mediaIDFK=None, projectID=None,
                 sdsID=None, sdsMetadataID=None, repTransportType=None, rePurpose=None,
                 sedimentTransportDirectionIDPK=None):
        """Initialize a littoral cell instance.
        :param sdsFeatureName: string name of feature
        :param sdsFeatureDescription: optional description of feature
        :param mediaIDFK: string, not used
        :param projectID: string, not currently used
        :param sdsID: string, not used
        :param sdsMetadataID: string, not used
        :param repTransportType: string, not currently used
        :param rePurpose: string, not currently used
        :param sedimentTransportDirectionIDPK: string primary key for feature in the geodatabase
        """
        super(Flux, self).__init__(sdsFeatureName, sdsFeatureDescription, mediaIDFK, projectID,
                                   sdsID, sdsMetadataID)

        for name in ['repTransportType', 'rePurpose', 'sedimentTransportDirectionIDPK']:
            self[name] = None

        # succinctly set all dictionary values
        for name in self.iterkeys():
            self[name] = eval(name)


class Scenario(Littoral):
    """A sediment budget scenario (or `alternative`) as a dictionary."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.
    __type_string__ = 'Scenario'

    def __init__(self, scenarioIDPK, sdsFeatureName, sdsFeatureDescription=None, mediaIDFK=None,
                 projectID=None, sdsID=None, sdsMetadataID=None, scenarioAuthor=None, dateScenarioCreated=None,
                 budgetType=None, dateIntervalStart=None, dateIntervalEnd=None, geoAreaDescription=None,
                 qUnits=None, uncertaintyMethod=None):
        """
        Initializes a scenario (or `alternative`)
        :param scenarioIDPK: string primary key for scenario in the geodatabase
        :param sdsFeatureName: string name of the scenario
        :param sdsFeatureDescription: string description of the scenario
        :param mediaIDFK: string, not used
        :param projectID: string, not currently used
        :param sdsID: string, not used
        :param sdsMetadataID: string
        :param scenarioAuthor: name of the author of the scenario
        :param dateScenarioCreated: datetime when the scenario was created
        :param budgetType: string, 'Micro' or 'Macro'
        :param dateIntervalStart: start datetime for the scenario
        :param dateIntervalEnd: end datetime for the scenario
        :param geoAreaDescription: string description of geographic area
        :param qUnits: string, not currently used
        :param uncertaintyMethod: string, not currently used
        """
        super(Scenario, self).__init__(sdsFeatureName, sdsFeatureDescription, mediaIDFK, projectID,
                                       sdsID, sdsMetadataID)

        for name in ['scenarioAuthor', 'dateScenarioCreated', 'budgetType', 'dateIntervalStart',
                     'dateIntervalEnd', 'geoAreaDescription', 'qUnits', 'scenarioIDPK']:
            self[name] = None

        # succinctly set all dictionary values
        for name in self.iterkeys():
            self[name] = eval(name)

        if self['dateScenarioCreated'] is None:
            self['dateScenarioCreated'] = datetime.date.today()


class CellScenario(dict):
    """Holds cell information for a single cell scenario."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, scenarioIDFK=None, littoralCellIDFK=None, deltaVolume=0, placementVolume=0,
                 removalVolume=0, residualVolume=0, valConfidence='No Data', scenarioValIDPK=None):

        """
        Initialize a cell scenario record.
        :param scenarioIDFK: string scenario ID foreign key
        :param littoralCellIDFK: string cell ID foreign key
        :param deltaVolume: float change in volume for the cell in the scenario
        :param placementVolume: float placement volume for the cell in the scenario
        :param removalVolume: float removal volume for the cell in the scenario
        :param residualVolume: foat residual volume for the cell in the scenario
        :param valConfidence: degree of confidence ('High', 'Medium', 'Low', 'No Data')
        :param scenarioValIDPK: string primary key for the cell scenario value in the geodatabase
        """
        super(CellScenario, self).__init__()

        for name in ['scenarioIDFK', 'littoralCellIDFK', 'deltaVolume', 'placementVolume', 'removalVolume',
                     'residualVolume', 'valConfidence', 'scenarioValIDPK']:
            self[name] = eval(name)


class FluxScenario(dict):
    """Holds sediment flux information for a single flux scenario."""
    # Developer note: field names are retained in camelCase to match fields within the
    # geodatabase. This breaks PEP-8 conventions, but is done to make our lives easier.

    def __init__(self, scenarioIDFK=None, qValue=0.0, sourceCellIDFK=None, sinkCellIDFK=None,
                 sedimentTransportDirectionIDFK=None, valConfidence='No Data', fluxValuesIDPK=None):
        """
        Initializes a flux scenario object.
        :param scenarioIDFK: string foreign key for the scenario
        :param qValue: float q value for the flux line in the scenario
        :param sourceCellIDFK: string key for sediment source cell for the scenario
        :param sinkCellIDFK: string key for sediment sink cell for the scenario
        :param sedimentTransportDirectionIDFK: string key for the fluxline
        :param valConfidence: degree of confidence ('High', 'Medium', 'Low', 'No Data')
        :param fluxValuesIDPK: string primary key for the flux scenario value in the geodatabase
        """
        super(FluxScenario, self).__init__()

        for name in ['scenarioIDFK', 'qValue', 'sourceCellIDFK', 'sinkCellIDFK',
                     'sedimentTransportDirectionIDFK', 'valConfidence', 'fluxValuesIDPK']:
            self[name] = eval(name)


class Loggable(object):
    """Parent class for any Arc class for which we want logging."""

    def __init__(self):
        """Initialize logging for the class."""
        super(Loggable, self).__init__()
        base = os.path.basename(inspect.getfile(inspect.currentframe()))
        script_name = os.path.splitext(base)[0]

        if script_name == '<stdin>':
            script_name = 'SBAS'

        self._logger = logging.getLogger(__name__)
        self._initlogging(r'{0}\logs\{1}-{2}.log'.format(sbas_dir, script_name,
                                                         datetime.date.today()))
        sys.excepthook = self._myexcepthook

    def _initlogging(self, file_path):
        """Start logging to specified file_path
        :param file_path: path to log file to open for append.
        """
        assert isinstance(file_path, (unicode, str))

        d = os.path.dirname(file_path)
        if not os.path.isdir(d):
            os.makedirs(d)

        logging.basicConfig(
            filename=file_path
            , format='%(asctime)s - %(levelname)s - %(message)s'
            , level=logging.INFO
            , filemode='a'
        )

        # add a channel for standard out
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        self._logger.addHandler(ch)

    def log(self, message, log_type='info'):
        """Write to log file and Arc Message buffer
        :param message: message string to write to logfile
        :param log_type: indicator for type of logging, e.g. 'error'
        """
        assert isinstance(log_type, (unicode, str))

        if log_type == 'error':
            self._logger.error(message)
        elif log_type == 'info':
            self._logger.info(message)
        elif log_type == 'warn':
            self._logger.warn(message)

        arcpy.AddMessage(message)

    def log_params(self, parameters):
        """Write parameters, values, and datatypes to log
        :param parameters: list of parameters to log.
        """
        assert isinstance(parameters, (tuple, list))
        for p in parameters:
            self.log('\t{0}\t{1}\t{2}'.format(p.name, p.value, type(p.value)))

    def _myexcepthook(self, extype, exval, trace):
        """Exception hook to capture and log all exception stack traces
        :param extype: exception type
        :param exval: exception value
        :param trace: stack trace
        """
        self._logger.error('Logging an uncaught exception', exc_info=(extype, exval, trace))
        arcpy.AddMessage(trace)
        traceback.print_exception(extype, exval, trace)


class ArcObjects(Loggable):
    """Class to work with legacy ArcObjects for use in SBAS.
    Requires the comtypes package: https://pypi.python.org/pypi/comtypes
    Requires slightly modified code snippets provided by Mark Cederholm
    for accessing ArcObjects via Python. (Included)"""

    def __init__(self):
        """Initialize the ArcObjects instance with default parent class values."""
        super(ArcObjects, self).__init__()
        self.p_app, self.p_doc, self.p_mxdoc = None, None, None
        self.get_arcmap_doc()

    @staticmethod
    def enumerate_label_classes(layer):
        """Enable iteration over the label classes on a layer.
        :param layer: layer whose labels we are interested in
        """
        feat_layer = Snippets.CType(layer, esriCarto.IGeoFeatureLayer)

        lbl_classes = feat_layer.AnnotationProperties

        for i in range(lbl_classes.Count):
            lbl_class, lbl_class_id = Snippets.CType(
                lbl_classes, esriCarto.IAnnotateLayerPropertiesCollection2).QueryItem(i)
            yield lbl_class

    def rename_data_frame(self, name):
        """Will rename the data frame on the map document to the provided name.
        :param name: new name to give the data frame.
        """
        doc = self.get_arcmap_doc()
        doc.Name = name
        self.p_mxdoc.UpdateContents()

    def activate_data_frame(self, name):
        """
        Activates a data frame matching the given name.
        :param name: data frame to activate
        """
        maps = self.p_mxdoc.Maps
        df = self.p_mxdoc.ActiveView
        for i in range(maps.Count):
            if maps[i].Name == name:
                df = maps[i]
                break

        self.p_mxdoc.ActiveView = df

    def get_arcmap_doc(self):
        """Return ArcMap's currently open document, for situations where the
        document may not yet have been written to disk.
        :rtype : comtypes.POINTER(IMap)
        """
        self.p_app = Snippets.GetApp('ArcMap')
        doc = self.p_app.Document
        self.p_mxdoc = Snippets.CType(doc, esriArcMapUI.IMxDocument)
        return self.p_mxdoc.FocusMap

    @staticmethod
    def get_layer(name, map_document):
        """Returns a layer in an MXD matching the given name.
        :param name: name of layer
        :param map_document: map document object to search
        """
        for i in range(map_document.LayerCount):
            if map_document.Layer(i).Name == name:
                return Snippets.CType(map_document.Layer[i], esriCarto.IFeatureLayer)

        return None

    def get_layer_by_name(self, lyr_name):
        """
        Return an IFeatureLayer pointer to a layer matching the provided name.
        :param lyr_name: layer name to search
        :return: ESRI IFeatureLayer pointer to the matching layer
        """
        assert isinstance(lyr_name, (unicode, str))
        p_map = self.get_arcmap_doc()
        for i in range(p_map.LayerCount):
            if p_map.Layer(i).Name == lyr_name:
                return Snippets.CType(p_map.Layer[i], esriCarto.IFeatureLayer)

        return None

    def get_layer_names(self):
        """
        Get a list of all layer names in the document
        :rtype : list
        :return: list of all layer names in the document
        """
        layers = []
        p_map = self.get_arcmap_doc()
        for i in range(p_map.LayerCount):
            layers.append(p_map.Layer(i).Name)

        return layers

    def clear_selected_graphics(self):
        """Remove only the selected graphics from the map."""
        self.log('clear_selected_graphics()')
        p_map = self.get_arcmap_doc()
        p_sel_elements = self.get_selected_elements()
        p_graphic_cont = Snippets.CType(p_map, esriCarto.IGraphicsContainer)
        p_sel_elements.Reset()
        p_el = p_sel_elements.Next()

        while p_el:
            p_graphic_cont.DeleteElement(p_el)
            p_el = p_sel_elements.Next()

        self.p_mxdoc.ActiveView.Refresh()

    def clear_all_graphics(self):
        """Will remove all graphics from the map."""
        self.log('Clearing graphics')
        p_map = self.get_arcmap_doc()
        p_graphic_cont = Snippets.CType(p_map, esriCarto.IGraphicsContainer)
        p_graphic_cont.DeleteAllElements()
        self.p_mxdoc.ActiveView.Refresh()

    def get_selected_elements(self):
        """
        Return the selected graphical elements
        :return: ESRI IEnumElement pointer to the selected elements
        """
        p_graphic_cont = Snippets.CType(self.get_arcmap_doc(), esriCarto.IGraphicsContainer)
        p_graph_sel = Snippets.CType(p_graphic_cont, esriCarto.IGraphicsContainerSelect)
        return p_graph_sel.SelectedElements

    @staticmethod
    def create_sediment_feature(p_feat_layer, p_geom, feat_id, feature):
        """
        Adds a new sediment budget feature
        :param p_feat_layer: pointer to a feature layer
        :param p_geom: pointer to a geometry type
        :param feat_id: ID to store in the created feature
        :param feature: dictionary of feature's values
        :raises ValueError: if a geometry type other than POLYGON or POLYLINE is provided
        """
        assert isinstance(feat_id, (unicode, str))
        assert isinstance(feature, dict)

        # Get the shape of the graphic, and use it as the shape
        # for the feature.
        p_feat = p_feat_layer.FeatureClass.CreateFeature()
        p_feat.Shape = p_geom

        # bring in littoral_cell class values and write them to the feature
        for key, val in feature.iteritems():
            index = p_feat.Fields.FindField(key)
            if index < 0:
                # Our primary key is different because this is a macro feature.
                # Override the primary key.
                s_geom = ArcObjects.get_graphic_geom_as_unicode(p_geom)
                if s_geom == u'POLYGON':
                    key = u'macroCellIDPK'
                elif s_geom == u'POLYLINE':
                    key = u'macroTransportDirectionIDPK'
                else:
                    raise ValueError('Bad geometry {0} for feature'.format(s_geom))

                index = p_feat.Fields.FindField(key)

            if key.endswith('IDPK'):
                p_feat.Value[index] = feat_id
            else:
                p_feat.Value[index] = val

        p_feat.Store()

    def add_graphic_to_cur_lyr(self, p_el, g_id, scenario_name, feature_name=None):
        """Add a single graphic to selected feature layer, or default feature layer
        if no feature layer is selected.
        :param g_id: feature graphic ID
        :param scenario_name: str name of layer to add
        :param p_el: pointer to the graphic element
        :param feature_name: name to give the feature
        """
        self.log('add_graphic_to_cur_lyr("{0}", "{1}", "{2}", "{3}")'.format(
            p_el, g_id, scenario_name, feature_name
        ))

        bad_geom = Exception('No matching geometry found')

        p_geom = p_el.Geometry
        if not p_geom:
            raise bad_geom

        # Get the shape of the graphic, and use it as the shape
        # for the feature.
        s_geom = ArcObjects.get_graphic_geom_as_unicode(p_geom)
        if s_geom == u'POLYLINE':
            if scenario_name is None:
                lyr_name = 'Unassigned Fluxes'
            else:
                lyr_name = '{} Fluxes'.format(scenario_name)

            layer = self.get_layer_by_name(lyr_name)

        elif s_geom == u'POLYGON':
            if scenario_name is None:
                lyr_name = 'Unassigned Cells'
            else:
                lyr_name = scenario_name

            layer = self.get_layer_by_name(lyr_name)
        else:
            raise bad_geom

        if not p_geom.GeometryType == layer.FeatureClass.ShapeType:
            raise bad_geom

        self.log('\tFound matching geometry')
        self.log(feature_name is None)

        geom = p_el.Geometry.GeometryType

        if geom == esriGeometry.esriGeometryPolyline:
            if feature_name is None:
                feature_name = 'Flux{0}'.format(g_id[:4])
            feat = Flux(feature_name)
        elif geom == esriGeometry.esriGeometryPolygon:
            if feature_name is None:
                feature_name = 'Littoral{0}'.format(g_id[:4])
            feat = LittoralCell(feature_name)
        else:
            raise bad_geom

        ArcObjects.create_sediment_feature(layer, p_geom, g_id, feat)
        return s_geom

    @staticmethod
    def get_graphic_geom_as_unicode(p_geom):
        """
        Return the geometry of the graphic as a unicode
        :rtype : unicode
        :param p_geom: pointer to geometry
        :return: POINT, POLYLINE, POLYGON, or OTHER
        """
        geom_type = p_geom.GeometryType
        if geom_type == esriGeometry.esriGeometryPoint:
            return u'POINT'
        elif geom_type == esriGeometry.esriGeometryPolyline:
            return u'POLYLINE'
        elif geom_type == esriGeometry.esriGeometryPolygon:
            return u'POLYGON'
        return u'OTHER'

    def add_selected_graphics_to_tables(self, scenario_name, feature_name=None, do_all=True):
        """Add graphics to selected feature layer, or default feature layer
        if no feature layer is selected.
        :param scenario_name: str name of scenario to update
        :rtype : dict
        :param feature_name: string name of feature in which to insert graphics
        :param do_all: Whether to convert all graphics. Will convert only selected graphics if False.
        """
        self.log('add_selected_graphics_to_tables("{0}", "{1}", "{2}")'.format(scenario_name, feature_name, do_all))

        retval = dict()

        # Find the selected elements
        count = 0
        name = feature_name

        if do_all:
            p_map = self.get_arcmap_doc()
            pointer = Snippets.CType(p_map, esriCarto.IGraphicsContainer)
        else:
            pointer = self.get_selected_elements()

        pointer.Reset()
        p_el = pointer.Next()

        # Check for None pointer, indicating no graphics in container (or no selected
        # graphics)
        if not p_el:
            raise ValueError('No selected graphics on current layer.')

        while p_el:
            count += 1
            if count > 1 and feature_name is not None:
                # feature name will use the base name plus the count if
                # we have more than one feature to add
                name = '{0}_{1}'.format(feature_name, count)
            g_id = unicode(uuid.uuid4())
            g_type = self.add_graphic_to_cur_lyr(p_el, g_id, scenario_name, name)
            retval[g_id] = name, g_type
            p_el = pointer.Next()

        return retval
