#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
#
"""Stores functions related to the configuration file"""

import ConfigParser
import arcpy
import os

__author__ = 'k5opjeas'

sbas_dir = os.path.dirname(os.path.realpath(__file__))


def get_workspace(cfg_file='{}\\sbas.cfg'.format(sbas_dir)):
    """
    Read the last used workspace from a local text file, since Arc does not
    respect workspace changes done via tools.
    :param cfg_file: path to config file
    :type cfg_file: unicode or str
    :rtype : unicode
    :return: last used workspace, or current workspace if none.
    """
    cfg = get_config(cfg_file)
    return cfg.get('sbas', 'workspace')


def set_workspace(workspace, cfg_file='{}\\sbas.cfg'.format(sbas_dir)):
    """
    Save the last used workspace to a local text file, since Arc does not
    respect workspace changes done via tools.
    :param workspace: Path to workspace geodatabase
    :type workspace: unicode or str
    :param cfg_file: path to configuration file
    :type cfg_file: unicode or str
    :raise Exception: Could not find workspace
    :raise Exception: Failed to write workspace file
    """
    if not os.path.isdir(workspace):
        raise Exception('Could not find workspace {}'.format(workspace))

    cfg = get_config(cfg_file)
    cfg.set('sbas', 'workspace', workspace)
    write_config(cfg_file, cfg)


def get_current_units():
    """
    Return the units currently in use. Default to Cubic Yards if none.
    :rtype : unicode
    :return: Units currently in use, or default of u'Cubic Yards/Year'
    """
    ws = get_workspace()
    arcpy.env.workspace = ws

    units = 'Cubic Yards/Year'
    cfg_file = '{}\\sbas.cfg'.format(sbas_dir)

    try:
        cfg = get_config(cfg_file)
        return cfg.get('sbas', 'units')
    except (IOError, ConfigParser.Error):
        return units


def set_current_units(units):
    """
    Save the units to a local text file.
    :param units: unicode or str
    """
    cfg_file = '{}\\sbas.cfg'.format(sbas_dir)
    cfg = get_config(cfg_file)
    cfg.set('sbas', 'units', units)
    write_config(cfg_file, cfg)


def get_config(cfg_file=None):
    """
    Reads the configuration file and returns the resulting ConfigParser object
    :param cfg_file: path to configuration file
    :type cfg_file: unicode or str
    :return: configuration object
    :rtype ConfigParser.ConfigParser:
    """
    if cfg_file is None:
        cfg_file = '{}\\sbas.cfg'.format(sbas_dir)
    if not os.path.exists(cfg_file):
        import shutil
        shutil.copy('{}\\templates\\{}'.format(sbas_dir, os.path.basename(cfg_file)), sbas_dir)
    cfg = ConfigParser.ConfigParser()
    cfg.read(cfg_file)
    return cfg


def write_config(cfg_file, cfg_object):
    """
    Write the configuration object to file.
    :param cfg_file: path to write
    :type cfg_file: unicode or str
    :param cfg_object: ConfigParser object to write
    :type cfg_object: ConfigParser.ConfigParser
    """
    with open(cfg_file, 'w') as f:
        cfg_object.write(f)


def get_metadata_url():
    """
    Gets the REST URL for metadata publishing and searching.
    :return: Metadata REST URL
    """
    cfg_file = '{0}\\sbas.cfg'.format(sbas_dir)
    cfg = get_config(cfg_file)
    return cfg.get('metadata', 'rest_url')


def get_credentials():
    """
    Gets username and password from the sbas config file.
    :return: username, password
    :rtype tuple:
    """
    cfg_file = '{0}\\sbas.cfg'.format(sbas_dir)
    cfg = get_config(cfg_file)
    return cfg.get('metadata', 'user'), cfg.get('metadata', 'pass')
