#!/usr/bin/env python2
# vim: set fileencoding=utf-8 :
"""Contains logic for importing data from other file formats."""

import arcpy
import classes
import architecture
import statics
import configuration
import get
import da
import os
import uuid
import json
from arcpy.arcobjects import geometries

__author__ = 'k5opjeas'

lflux_marker = 'RATE TO LEFT'
rflux_marker = 'RATE TO RIGHT'
lflux_inlet_marker = 'INLET L TO R'
rflux_inlet_marker = 'INLET R TO L'


def import_gen_file(path):
    """
    Will read gencade data from a .GEN file
    :rtype : dict
    :param path: path to the .GEN file
    :return: gencade dictionary object
    :raise Exception:
    """
    assert isinstance(path, (unicode, str))

    gencade = dict()
    cells, current_cell, fluxes, current_flux, lines = [], [], [], [], []
    found_marker = False
    cid = 1
    fid = 1
    azimuth = 0.0

    try:
        lines = [line.strip() for line in open(path, 'r')]
    except IOError:
        fname = os.path.basename(path)
        raise IOError('Failure to read GenCade file: {0}'.format(fname))

    for i in range(len(lines)):

        if lines[i].upper().startswith('AZIMUTH:'):
            # Grab the azimuth, which will help us guess which order the
            # features appear in the shapefile.
            azimuth = float(lines[i].split(':')[1].strip())
            continue

        elif '** SBAS **' in lines[i]:
            # Indicate we have found the SBAS section.
            found_marker = True
            continue

        if found_marker and lines[i].startswith('ISBAS'):

            splt = lines[i].split(':')

            if len(splt) != 2:
                raise ValueError('Unrecognized SBAS line: `{0}`'.format(lines[i]))

            to_append = splt[1].strip()

            if splt[0][5] == 'V':
                current_cell.append(to_append)

                if len(current_cell) == 2:
                    gc_cell = classes.GenCadeCell(cid, current_cell[0], current_cell[1])
                    cells.append(gc_cell)
                    current_cell = []
                    cid += 1

            elif splt[0][5] == 'Q':
                current_flux.append(to_append)

                if len(current_flux) == 2:
                    gc_flux = classes.GenCadeFlux(fid, current_flux[0], current_flux[1])
                    fluxes.append(gc_flux)
                    current_flux = []
                    fid += 1

            else:
                raise ValueError('Unrecognized SBAS line: `{0}`'.format(lines[i]))

    if len(current_flux) != 0:
        raise ValueError('Mis-matched number of fluxes in GEN file.')

    if len(current_cell) != 0:
        raise ValueError('Mis-matched number of cells in GEN file.')

    gencade['cells'] = cells
    gencade['fluxes'] = fluxes
    gencade['azimuth'] = azimuth
    return gencade


def found_flux(line):
    """
    Return whether the flux marker is found in the line.
    :param line: line to inspect
    """
    return lflux_marker in line or lflux_inlet_marker in line or rflux_marker in line or rflux_inlet_marker in line


def import_prt_file(path, gc):
    """
    Import gencade cell volumes from a PRT file.
    :rtype : dict
    :param path: path to PRT file
    :param gc: gencade dictionary object
    :return: updated gencade dictionary object, units
    :raise Exception:
    """
    assert isinstance(path, (unicode, str))

    sbas_marker = 'SBAS OUTPUT'
    dvol_marker = 'CHANGE RATE'
    pvol_marker = 'BEACH FILL RATE'
    rvol_marker = 'REMOVAL RATE'
    byp_src_marker = 'BYPASS SOURCE RATE '
    byp_snk_marker = 'BYPASS SINK RATE'
    cy_marker = '(CY/YR):'
    cell_marker = 'SBAS CELL    '
    flux_marker = 'SBAS FLUX ARROW  '
    cid = 0
    fid = 0
    found_marker = False
    units = None

    try:
        lines = [line.strip() for line in open(path, 'r')]

    except IOError:
        fname = os.path.basename(path)
        raise IOError('Failure to read GenCade file: {0}'.format(fname))

    for line in lines:

        if sbas_marker in line:
            found_marker = True
            continue

        pvol = pvol_marker in line
        bypsrc = byp_src_marker in line
        rvol = rvol_marker in line
        bypsnk = byp_snk_marker in line

        if found_marker and 'WARNING' not in line:
            if cell_marker in line:
                # Cell ID should be in the same line as the cell marker
                cid = int(line.split(cell_marker)[1].strip())
            elif flux_marker in line:
                # Flux ID should be in the same line as the ce
                fid = int(line.split(flux_marker)[1].strip())

        if found_marker and (pvol or bypsrc or dvol_marker in line or rvol or bypsnk):

            # Set the key to search for in the dictionary when storing
            # the volume value
            volume = float(line.split(':')[1].strip())

            if pvol or bypsrc:
                volume_key = 'placementVolume'
            elif dvol_marker in line:
                volume_key = 'deltaVolume'
            elif rvol or bypsnk:
                volume_key = 'removalVolume'
            else:
                volume_key = ''

            # Units marker should be on the same line
            if units is None:
                if cy_marker in line:
                    units = 'Cubic Yards/Year'
                else:
                    units = 'Cubic Meters/Year'

            if 'Meters' in units:
                # store as cubic yards in the database
                volume = architecture.to_cubic_yards(volume)

            # removal volumes should be imported as absolute values
            if volume_key == 'removalVolume':
                volume = abs(volume)

            for j in range(len(gc['cells'])):
                cell = gc['cells'][j]

                if cell['gc_id'] == cid:
                    gc['cells'][j][volume_key] += volume
                    break

        elif found_marker and found_flux(line):

            if lflux_marker in line or lflux_inlet_marker in line:
                flux_key = 'q_left'
            else:
                flux_key = 'q_right'

            # Q values should be absolute values
            q_val = abs(float(line.split(':')[1].strip()))

            if 'Meters' in units:
                # store as cubic yards in the database
                q_val = architecture.to_cubic_yards(q_val)

            for j in range(len(gc['fluxes'])):
                flux = gc['fluxes'][j]

                if flux['gc_id'] == fid:
                    flux[flux_key] = q_val
                    gc['fluxes'][j] = flux
                    break

    return gc


def import_gen_shapefile(path, gc, scen_id):
    """
    Import the cells and fluxes from a shapefile and gencade object.
    :rtype : dict
    :param path: path to .SHP file
    :param gc: gencade dictionary object
    :param scen_id: scenario ID to save in the cells
    :return: dictionary of cell and flux ID keys
    :raise Exception:
    """
    assert isinstance(path, (unicode, str))
    assert isinstance(scen_id, (unicode, str))

    cid, fid = 0, 0
    result = {
        'cell_keys': []
        , 'flux_keys': []
    }

    # Grab JSON representations of all cells and fluxes
    js = get_shapefile_json(path)
    # Re-order the items in the json object if necessary
    js = order_shapefile(js, gc['azimuth'])

    # Now we have the JSON representations of the features in memory.
    # Try to guess whether each one is a cell or a flux, based on the
    # number of points in its first path. We have to do this because
    # GenCade exports shapefiles where all geometries are polylines.
    # Start with cells because we will need to determine source and
    # sink cell IDs when doing the fluxes, so the cells need to exist
    # first!
    for row in js:
        jstring = row[0]
        jdict = row[1]

        if len(jdict['paths'][0]) > 2:
            # We assume this is a polygon since it has more than 2 points
            cid += 1
            cell = [c for c in gc['cells'] if c['gc_id'] == cid]

            if len(cell) != 1:
                raise Exception('Unknown cell ID {0}'.format(cid))

            cell = cell[0]
            lit_id_pk = add_json_cell(jstring, cell, scen_id)
            result['cell_keys'].append(lit_id_pk)

    for row in js:
        jstring = row[0]
        jdict = row[1]

        if len(jdict['paths'][0]) == 2:
            # We know this is a polyline since it has only two points
            fid += 1
            # There are two fluxes per fluxValuesID
            # Silly hack to keep track of the real fluxValuesID
            real_fid = int(round(fid / 2.0))

            flux = [f for f in gc['fluxes'] if f['gc_id'] == real_fid]

            if len(flux) != 1:
                raise Exception('Unknown flux ID {0} count {1}'.format(
                    real_fid, len(flux)))

            flux = flux[0]
            trans_dir_id_pk = add_json_flux(jstring, flux, scen_id)
            result['flux_keys'].append(trans_dir_id_pk)

    return result


def add_json_cell(jstring, cell, scen_id):
    """Given a dictionary representation of a cell geometry, add the
    littoral cell it represents to the cell scenario (alternative).
    :rtype : unicode
    :param jstring: JSON string representation of a littoral cell
    :param cell: dictionary storing cell values
    :param scen_id: scenario ID string
    :return: littoral cell ID of the added cell
    """
    assert isinstance(jstring, (unicode, str))
    assert isinstance(cell, dict)
    assert isinstance(scen_id, (unicode, str))

    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    # The JSON string we were passed is for a polyline, even though the
    # feature in question is really a polygon. We need to convert it to
    # a polygon by hacking the JSON string we get from the shapefile.
    jstring = jstring.replace('"paths":', '"rings":')

    # Now we can work with the geometry as normal
    polygon = arcpy.AsShape(jstring, True)

    # Write to LittoralCell table
    table = '{0}\\SedimentManagementCell'.format(ws)

    lit_id_pk = str(uuid.uuid4())
    ft_name = 'Littoral {0}-{1}'.format(cell['first_id'], cell['second_id'])

    fields = ('SHAPE@', 'sdsFeatureName', 'littoralCellIDPK', 'deltaVolume',
              'scenarioIDFK', 'residualVolume', 'placementVolume')
    data = (polygon, ft_name, lit_id_pk, cell['deltaVolume'],
            scen_id, cell['residualVolume'], cell['placementVolume'])
    da.insert_row(table, fields, data)
    del polygon

    # Write to CellScenario table
    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    scen_val_id_pk = str(uuid.uuid4())
    fields = ('deltaVolume', 'scenarioValIDPK', 'scenarioIDFK', 'littoralCellIDFK', 'residualVolume',
              'placementVolume', 'removalVolume')
    data = (cell['deltaVolume'], scen_val_id_pk, scen_id, lit_id_pk, cell['residualVolume'],
            cell['placementVolume'], cell['removalVolume'])
    da.insert_row(table, fields, data)

    return lit_id_pk


def add_json_flux(jstring, flux, scen_id):
    """Add the fluxline the JSON string represents to the flux scenario (alternative).
    :rtype : unicode
    :param jstring: string JSON representation of a flux geometry
    :param flux: flux data dictionary
    :param scen_id: scenario ID string
    :return: transport direction ID of the added flux
    """
    assert isinstance(jstring, (unicode, str))
    assert isinstance(flux, dict)
    assert isinstance(scen_id, (unicode, str))

    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    polyline = arcpy.AsShape(jstring, True)

    if polyline.firstPoint.X < polyline.lastPoint.X:
        # left-to-right line
        q_val = flux['q_left']
        sdsfeaturename = 'Flux{0}-{1} R-L'.format(flux['first_id'], flux['second_id'])
    else:
        # right-to-left line
        q_val = flux['q_right']
        sdsfeaturename = 'Flux{0}-{1} L-R'.format(flux['first_id'], flux['second_id'])

    # Write to TransportDirection table
    table = '{0}\\SedimentTransportDirection'.format(ws)
    sed_trans_dir_id_pk = str(uuid.uuid4())

    fields = ('SHAPE@', 'sdsFeatureName', 'sedimentTransportDirectionIDPK', 'qValue')
    data = (polyline, sdsfeaturename, sed_trans_dir_id_pk, q_val)
    da.insert_row(table, fields, data)

    # Write to FluxScenario table
    table = '{0}\\SedimentManagementFluxScenario'.format(ws)
    source, sink = get.get_source_and_sink_cell(polyline)
    flux_vals_id_pk = str(uuid.uuid4())
    data = (scen_id, q_val, source, sink, sed_trans_dir_id_pk, flux_vals_id_pk)
    fields = ('scenarioIDFK', 'qValue', 'sourceCellIDFK', 'sinkCellIDFK',
              'sedimentTransportDirectionIDFK', 'fluxValuesIDPK')
    da.insert_row(table, fields, data)
    del polyline

    return sed_trans_dir_id_pk


def order_shapefile(js, azimuth):
    """Determine if the order of the features in the shapefile matches the
    horizontal direction of the azimuth, and re-order if necessary.
    :rtype : list
    :param js: list of data imported from JSON
    :param azimuth: azimuth in decimal degrees
    :return: updated JSON data list
    """
    assert isinstance(js, list)
    assert isinstance(azimuth, (float, int))

    direction = 0.0
    last_avg = None

    for feature in js:
        paths = feature[1]['paths'][0]
        if len(paths) < 3:
            # Path only has two points, meaning it's a polyline. Skip it,
            # as it's easier to determine direction of the shapes from
            # polygons.
            continue

        else:
            # Get the average X-coordinate of the point. If it is less than
            # the last average, push the 'overall' direction a bit negative;
            # otherwise, push it a bit positive.
            tot = 0
            for point in paths:
                tot += point[0]

            avg = float(tot) / len(paths)

            if last_avg:
                if avg < last_avg:
                    direction -= 0.1
                else:
                    direction += 0.1

            last_avg = avg

    # If our 'overall direction' doesn't match the lateral direction of the
    # azimuth, re-order the items in `js`.
    if (direction < 0 and azimuth < 180) \
            or (direction >= 0 and azimuth >= 180):
        js.reverse()

    return js


def import_sbas_shapefile(shape_file, units='Cubic Yards'):
    """
    Writes cell and flux records to the appropriate feature classes and business tables.
    :param shape_file: path to shapefile
    :param units: original units of measure
    """

    # Determine geometry type and, therefore, the fields to use.
    # SBAS shapefiles are segregated by feature type (cell or flux).
    shape = da.get_single_value_from_table(shape_file, 'SHAPE@')
    if isinstance(shape, geometries.Polyline):
        cell = False
    elif isinstance(shape, geometries.Polygon):
        cell = True
    else:
        raise ValueError('Invalid geometry for feature.')

    # try to handle SBAS and SBAS-A files, which have different field lists
    table_fields = [f.name for f in arcpy.ListFields(shape_file)]
    if cell:
        if 'CONFIDENCE' in table_fields:
            fields = ('SHAPE@', 'NAME', 'ALT_ID', 'ID', 'RESIDUAL', 'CONFIDENCE', 'DV', 'PLACEMENT', 'REMOVAL')
        else:
            fields = ('SHAPE@', 'Name', 'Alt_ID', 'ID', 'Residual')
    else:
        if 'CONFIDENCE' in table_fields:
            fields = ('SHAPE@', 'ALT_ID', 'QVALUE', 'ID', 'NAME', 'CONFIDENCE')
        else:
            fields = ('SHAPE@', 'Alt_ID', 'QValue', 'ID', 'Name')

    for row in da.get_rows(shape_file, fields):
        if cell:
            import_sbas_cell(list(row), units)
        else:
            import_sbas_flux(list(row), units)


def import_sbas_cell(row, units='Cubic Yards'):
    """
    Will write necessary valies to the cell and cell scenario tables.
    :param row: row pulled from the shapefile
    :type row: list
    :param units: original units of measure
    :type units: str or unicode
    """
    # ('SHAPE@', 'NAME', 'ALT_ID', 'ID', 'RESIDUAL'[, 'CONFIDENCE', 'DV', 'PLACEMENT', 'REMOVAL'])

    arcpy.env.workspace = ws = configuration.get_workspace()

    # convert guids to expected format
    row[2] = architecture.uuid_to_string(row[2])
    row[3] = architecture.uuid_to_string(row[3])

    table = '{0}\\SedimentManagementCell'.format(ws)
    fields = ('SHAPE@', 'sdsFeatureName', 'littoralCellIDPK')
    ins_row = (row[0], row[1], row[3])
    da.insert_row(table, fields, ins_row)
    del ins_row, fields

    table = '{0}\\SedimentManagementCellScenario'.format(ws)
    fields = ['scenarioIDFK', 'littoralCellIDFK', 'residualVolume']

    # convert units for writing to the database
    if units.startswith('Cubic Meters'):
        row[4] = architecture.to_cubic_yards(row[4])

    if len(row) > 5:
        fields += ['valConfidence', 'deltaVolume', 'placementVolume', 'removalVolume']
        # convert units for writing to the database
        if units.startswith('Cubic Meters'):
            for i in range(6, 8):
                row[i] = architecture.to_cubic_yards(row[i])

    fields.append('scenarioValIDPK')
    # insert everything but the shape and name
    ins_row = row[2:len(row)]
    ins_row.append(str(uuid.uuid4()))
    da.insert_row(table, fields, ins_row)
    del ins_row, fields


def import_sbas_flux(row, units='Cubic Yards'):
    """
    Will write necessary values to the flux and flux scenario tables.
    :param row: row pulled from the shapefile
    :type row: list
    :param units: original units of measure
    """
    # ('SHAPE@', 'ALT_ID', 'QVALUE', 'ID', 'NAME'[, 'CONFIDENCE'])
    arcpy.env.workspace = ws = configuration.get_workspace()

    # convert units for writing to the database
    if units.startswith('Cubic Meters'):
        row[2] = architecture.to_cubic_yards(row[2])

    # convert guids to expected format
    row[1] = architecture.uuid_to_string(row[1])
    row[3] = architecture.uuid_to_string(row[3])

    table = '{0}\\SedimentTransportDirection'.format(ws)
    fields = ('SHAPE@', 'sdsFeatureName', 'sedimentTransportDirectionIDPK')
    ins_row = (row[0], row[4], row[3])
    da.insert_row(table, fields, ins_row)
    del ins_row, fields

    # we must reproject the feature into WGS84 so .contains() will work correctly
    source, sink = get.get_source_and_sink_cell(row[0].projectAs(arcpy.SpatialReference(4326)))

    table = '{0}\\SedimentManagementFluxScenario'.format(ws)
    fields = ['scenarioIDFK', 'qValue', 'sedimentTransportDirectionIDFK']
    if len(row) > 5:
        fields.append('valConfidence')
        if row[5].strip() == '':
            row[5] = 'No Data'
        ins_row = [row[1], row[2], row[3], row[5]]
    else:
        ins_row = row[1:len(row) - 1]
    fields += ['sourceCellIDFK', 'sinkCellIDFK', 'fluxValuesIDPK']
    ins_row += source, sink, uuid.uuid4()
    da.insert_row(table, fields, ins_row)
    del ins_row, fields


def get_shapefile_json(path):
    """
    Gets the JSON representations of the shapes, and adds them to
    a list for use in shapefile conversion logic.
    :param path: path to .SHP file
    :return: list of JSON strings
    """
    js = []
    for row in da.get_rows(path, ['SHAPE@JSON']):
        js.append([row[0], json.loads(row[0])])
    return js


def import_gencade(gen_file, prt_file, scen, shape_file):
    """Wrapper for all logic related to importing gencade data.
    :rtype : tuple
    :param gen_file: path to .GEN file
    :param prt_file: path to .PRT file
    :param scen: name of scenario/alternative to update
    :param shape_file: path to .SHP file
    :return: assigned cells, assigned fluxes, unassigned cells, unassigned fluxes
    """
    for param in (gen_file, prt_file, scen, shape_file):
        assert isinstance(param, (unicode, str))

    ws = configuration.get_workspace()
    arcpy.env.workspace = ws

    scen_id = get.get_scenario_id_by_name(scen)
    gc = import_gen_file(gen_file)
    gc = import_prt_file(prt_file, gc)
    import_gen_shapefile(shape_file, gc, scen_id)

    # re-set the scenario's assigned fluxes definitionquery
    assigned_fluxes = arcpy.mapping.Layer('{0} Fluxes'.format(scen))
    where = get.get_assigned_fluxes_in(scen_id)
    assigned_fluxes.definitionQuery = where

    # and the unassigned definitionquery
    mxd = arcpy.mapping.MapDocument('CURRENT')
    unassigned_fluxes, unassigned_cells = None, None
    if len(arcpy.mapping.ListLayers(mxd, 'Unassigned Fluxes')) > 0:
        unassigned_fluxes = arcpy.mapping.Layer('Unassigned Fluxes')
        where = get.get_unassigned_fluxes_in(scen_id)
        unassigned_fluxes.definitionQuery = where

    # same for cells
    assigned_cells = arcpy.mapping.Layer(scen)
    where = get.get_assigned_cells_in(scen_id)
    assigned_cells.definitionQuery = where

    if len(arcpy.mapping.ListLayers(mxd, 'Unassigned Cells')) > 0:
        unassigned_cells = arcpy.mapping.Layer('Unassigned Cells')
        where = where.replace(' in (', ' not in (')
        unassigned_cells.definitionQuery = where

    statics.recalculate_scenario_residual_volumes(scen)

    return assigned_cells, assigned_fluxes, unassigned_cells, unassigned_fluxes


if __name__ == '__main__':
    shp = r"C:\Users\k5opjeas\Documents\ArcGIS\Legacy_SBAS_Microbudgets\Microbudgets\PI-Buffalo_02MidCent_Flux.shp"
    units = 'Cubic Meters/Year'
    import_sbas_shapefile(shp, units)