﻿#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
"""Module to contain the core SBAS functions. Helper classes and any functions
requiring ArcObjects are in the architecture module."""

from architecture import *
from classes import *
import statics
import configuration
import load
import get
import da
import metadata
import tempfile

__author__ = 'k5opjeas'

sbas_dir = os.path.dirname(os.path.realpath(__file__))


class SBAS(Loggable):
    """Class to contain methods for creating SBAS budgets.
    Requires the comtypes package: https://pypi.python.org/pypi/comtypes
    Requires slightly modified code snippets provided by Mark Cederholm
    for accessing ArcObjects via Python. (Included)"""

    def __init__(self, map_document='CURRENT'):
        """
        Initializes an SBAS object. Supply an mxd path when debugging outside
        ArcMap.
        """
        super(SBAS, self).__init__()
        self.map_document = map_document
        self.symbology_field = ''
        self.micro_tables = (
            'Metadata', 'SedimentManagementCellScenario',
            'SedimentManagementFluxScenario', 'SedimentManagementScenario',
            'SedimentManagementCell', 'SedimentTransportDirection')

    def get_search_val(self, lyr):
        """
        Returns the value to search for updating symbology.
        :rtype : unicode
        :param lyr: layer object to investigate
        :return: symbology value field
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        self.log('\t{0}'.format(lyr.symbologyType))

        if self.symbology_field != '':
            search_val = self.symbology_field
        elif lyr.symbologyType == 'GRADUATED_COLORS':
            search_val = lyr.symbology.valueField
        else:
            search_val = 'valConfidence'

        self.log('search_val: {0}'.format(search_val))
        return search_val

    def edit_scenario(self, original_name, scen_name, scen_desc, author,
                      date_from, date_thru, area_desc):
        """
        Update the scenario with new values, and rename any layers referencing
        the old name.
        :param original_name: name of scenario to edit
        :param scen_name: updated scenario name
        :param scen_desc: updated description
        :param author: updated author
        :param date_from: updated from date
        :param date_thru: updated through date
        :param area_desc: updated geographic area description
        :return: list of updated layer objects
        :rtype list:
        """
        self.log('Updating scenario table')
        scen_id = get.get_scenario_id_by_name(original_name)
        statics.set_scenario_values(scen_id, scen_name, scen_desc, author,
                                    date_from, date_thru, area_desc)

        self.log('Updating metadata record')
        mxd = arcpy.mapping.MapDocument(self.map_document)
        df = arcpy.mapping.ListDataFrames(mxd, 'SBAS')[0]
        cell_lyr = arcpy.mapping.ListLayers(mxd, original_name)[0]
        flux_lyr = arcpy.mapping.ListLayers(mxd, '{0} Fluxes'.format(
            original_name))[0]
        coords = get.get_bounding_coords(cell_lyr, flux_lyr)
        metadata.build_sbas_metadata(scen_id, coords)

        self.log('Editing layer names on MXD')
        layers = arcpy.mapping.ListLayers(mxd, '{0}*'.format(original_name), df)

        for layer in layers:
            layer.name = layer.name.replace(original_name, scen_name)

        return layers

    def apply_symbology(self, layer_name, units, value_field='residualVolume',
                        show_units=True):
        """
        Apply a new symbology and re-classify shapes on the layer.
        :param layer_name: layer name to re-classify
        :param units: units for classification
        :param value_field: value field to search
        :param show_units: whether to show the units in the label expression
        :return: modified arc layer object
        """
        for p in (layer_name, units, value_field):
            assert isinstance(p, (unicode, str))

        self.log('apply_symbology("{0}", "{1}", "{2}")'.format(
            layer_name, units, value_field))

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        template = sbas_dir + '\\Templates\\'
        if layer_name.startswith('Unassigned'):
            if layer_name.endswith('Fluxes'):
                template += 'Unassigned Fluxes2.lyr'
            else:
                template += 'Unassigned Cells.lyr'
        else:
            if layer_name.endswith('Fluxes'):
                template += 'FluxLines2.lyr'
            else:
                # override template for confidence symbology
                if value_field == 'valConfidence':
                    template += 'Cells_Confidence.lyr'
                else:
                    template += 'Cells2.lyr'

        arcpy.ApplySymbologyFromLayer_management(layer_name, template)
        lyr = arcpy.mapping.Layer(layer_name)

        # Apply symbology and gradiant values to layer
        if lyr.name.startswith('Unassigned'):
            # no additional symbology classification for unassigned cells.
            self.log('\t\tDone classifying unassigned cells.')
            return lyr

        self.log('\tSetting transparency')
        lyr.transparency = 50

        if lyr.name == 'Unassigned Fluxes':
            # Our work here is done
            return lyr

        if lyr.name.endswith('Fluxes'):
            # Reclassify assigned fluxes based on q-val
            if 'Yards' in units:
                lyr.labelClasses[0].expression = 'Round([qValue])'
            else:
                lyr.labelClasses[0].expression = 'Round(0.764554858*[qValue])'

            # And we're done
            return lyr

        if value_field == 'valConfidence':
            self.log('\tSetting confidence where expression')
            lyr.labelClasses[0].expression = '[{0}]'.format(value_field)
            lyr.showLabels = True
            lyr.symbology.showOtherValues = False

            self.log('\t\t`{0}`'.format(lyr.labelClasses[0].expression))

            # our work here is done
            return lyr

        # Remaining possible classifications are by volumes.
        self.log('\tSetting volume display expression')
        lyr.symbology.valueField = value_field
        lyr.symbology.classBreakValues = [
            -9999999999999999999, -1, 1, 9999999999999999999]
        lyr.symbology.classBreakLabels = [
            'Cell Loss', 'Cell Balance', 'Cell Gain']

        year_label = ''
        if 'Year' in units:
            year_label = '/YR'

        # Only show the units in the expression if the user chooses to
        if 'Yards' in units:
            expr = 'Round([{0}])'.format(value_field)
            if show_units:
                expr = '{0} & " YD3{1}"'.format(expr, year_label)
        else:
            expr = 'Round(0.764554858*[{0}])'.format(value_field)
            if show_units:
                expr = '{0} & " M3{1}"'.format(expr, year_label)

        lyr.labelClasses[0].expression = expr
        self.log('\t\t`{0}`'.format(lyr.labelClasses[0].expression))
        lyr.showLabels = True
        lyr.symbology.reclassify()
        self.symbology_field = value_field
        return lyr

    def remove_other_scenarios(self, current):
        """
        Will remove any other scenario layers from the map document.
        :param current: current scenario name
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws
        mxd = arcpy.mapping.MapDocument(self.map_document)
        df = arcpy.mapping.ListDataFrames(mxd, 'SBAS')[0]
        del mxd
        layers = arcpy.mapping.ListLayers(df)

        for layer in layers:
            if layer.isFeatureLayer \
                    and not layer.name.startswith(current) \
                    and not layer.name.startswith('Unassigned'):
                # See if the layer has an sdsFeatureName field, indicating it's
                # an SBAS layer and needs removal
                if len(arcpy.ListFields(layer.name, 'sdsFeatureName')) > 0:
                    self.log('Removing {0}'.format(layer.name))
                    arcpy.mapping.RemoveLayer(df, layer)

    @property
    def get_current_scen_name(self):
        """
        Return the name of the currently loaded scenario.
        :rtype : unicode
        :return: current scenario name
        """
        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        scenarios = get.get_scenario_names(budget_type='micro')
        mxd = arcpy.mapping.MapDocument(self.map_document)
        df = arcpy.mapping.ListDataFrames(mxd, 'SBAS')
        scen_name = None
        if len(df) > 0:
            df = df[0]
            layers = arcpy.mapping.ListLayers(df)

            for layer in layers:
                if layer.name in scenarios:
                    scen_name = layer.name
                    break

        return scen_name

    def get_selected_feature_ids(self, macro=False):
        """
        Search all SBAS classes on the map document and get the IDs of all littoral cells
        and fluxes that are selected.
        :param macro: bool
        :rtype : dict
        :return: selected cell IDs and flux IDs
        """
        # DEVNOTE: This function cannot be tested outside of ArcMap, as the .Describe()
        # function will not work correctly with layers whose names differ from the name
        # of the feature class to which they belong. So, when we have layers named
        # "Unassigned Fluxes", arcpy.Describe() bombs, pretending it cannot find the
        # layer in question despite the fact that it exists on the document. It behaves
        # correctly when called directly from ArcMap. ಠ_ಠ
        arcpy.env.workspace = ws = configuration.get_workspace()

        retval = {
            'SedimentManagementCell': []
            , 'SedimentTransportDirection': []
            , 'SedimentManagementMacroCell': []
        }

        search_columns = ('littoralCellIDPK', 'sedimentTransportDirectionIDPK')
        mxd = arcpy.mapping.MapDocument(self.map_document)

        for layer in arcpy.mapping.ListLayers(mxd):
            self.log('\tSearching {}'.format(layer.name))

            # Determine whether the layer has one of our search columns
            try:
                desc = arcpy.Describe(layer.name)
                search_column = None
                for field in desc.fields:
                    if field.name in search_columns:
                        search_column = field.name
                        break
            except AttributeError:
                # Feature layer does not have "fields". It's probably a basemap.
                self.log('Skipping probable basemap {}'.format(layer.name))
                continue
            except IOError:
                # Another basemap...?
                self.log('Skipping possible basemap {}'.format(layer.name))
                continue

            if search_column is None:
                # Layer is not an SBAS layer.
                continue

            # Get selected object IDs on the layer
            obj_ids = desc.FIDset.split('; ')

            for obj_id in obj_ids:
                if len(obj_id) == 0:
                    continue

                if search_column == search_columns[0]:
                    if macro:
                        table = 'SedimentManagementMacroCell'
                    else:
                        table = 'SedimentManagementCell'
                else:
                    table = 'SedimentTransportDirection'

                # Get the feature's IDPK value
                full_table = '{0}\\{1}'.format(ws, table)
                where = build_where_clause(full_table, 'OBJECTID', obj_id)
                feat_id = da.get_single_value_from_table(full_table, search_column, where)
                self.log('\t\tFound table {0} feature {1}'.format(full_table, feat_id))
                if feat_id not in retval[table]:
                    retval[table].append(feat_id)

        self.log(retval)
        return retval

    def update_flux_values(self, scen_name, flux_class, q_val, confidence='No Data', dir_id=None,
                           sdsfeaturename=None, new=False, sdsfeaturedesc=''):
        """
        Update the scenario values for the given fluxline.
        :param new: whether these are new fluxes
        :param sdsfeaturename: str name of feature
        :param scen_name: scenario name to update
        :param flux_class: name of the fluxline class layer on the map document
        :param q_val: q value
        :param confidence: string confidence value
        :param dir_id: transport direction ID
        :param sdsfeaturedesc: feature description
        :rtype : tuple
        :return: whether we added a new flux scenario entry, updated scenario object
        """
        if confidence is None:
            confidence = 'No Data'

        for p in (scen_name, flux_class, confidence):
            assert isinstance(p, (unicode, str))

        assert isinstance(q_val, (float, int))

        arcpy.env.workspace = ws = configuration.get_workspace()
        scen = None

        if dir_id is None:
            dir_id = get.get_selected_feature_id(
                flux_class, 'Unassigned Fluxes', 'sedimentTransportDirectionIDPK')

        if sdsfeaturename is None:
            table = '{0}\\SedimentTransportDirection'.format(ws)
            where = build_where_clause(table, 'sedimentTransportDirectionIDPK', dir_id)
            sdsfeaturename = da.get_single_value_from_table(table, 'sdsFeatureName', where)

        vals = {
            'qValue': q_val
            , 'valConfidence': confidence
        }

        if new:
            # Insert the new flux scenario record and update our scen object
            scen = self.get_new_flux_scenario(dir_id, scen_name)

        table = '{0}\\SedimentManagementFluxScenario'.format(ws)
        scen_id = get.get_scenario_id_by_name(scen_name)
        where = build_where_clause(table, 'scenarioIDFK', scen_id)
        where = '{0} and {1}'.format(build_where_clause(table, 'sedimentTransportDirectionIDFK', dir_id), where)
        da.update_row(table, vals, where)

        if scen is None:
            scen = get.get_scenario_by_name(scen_name, dir_id=dir_id)

        vals['sdsFeatureName'] = sdsfeaturename
        vals['sdsFeatureDescription'] = sdsfeaturedesc
        del vals['valConfidence']

        # adjust field names depending on whether it's micro or macro
        budget_type = get.get_budget_type(scen_id).lower()
        if budget_type == 'micro':
            flux_field = 'sedimentTransportDirectionIDPK'
        else:
            flux_field = 'macroTransportDirectionIDPK'

        # update volume value in layer on mxd
        for flux in scen['flux']:
            where = build_where_clause(
                flux_class, flux_field, flux['sedimentTransportDirectionIDFK'])
            da.update_row(flux_class, vals, where)

        # Ensure any cells intersecting the fluxes are updated
        if not new:
            scen = statics.update_intersecting_residuals(flux_class, scen_name)

        cells = arcpy.mapping.Layer(scen_name)

        return new, scen, cells

    def update_cell_values(self, scen_name, vals, sdsfeaturename=None, lit_id=None, sdsfeaturedesc=None):
        """
        Save the updated cell values to the appropriate tables
        :rtype : tuple
        :param scen_name: scenario name to update
        :param vals: dictionary of values to save
        :param sdsfeaturename: the name to save to the cell's feature name field
        :param lit_id: littoral ID
        :param sdsfeaturedesc: feature description
        :return: updated scenario, whether the cell is new to the scenario
        """
        arcpy.env.workspace = ws = configuration.get_workspace()

        new = False

        if lit_id is None:
            lit_id, new = get.get_selected_littoral_id(scen_name)

            if lit_id is None:
                self.log('No littoral to add.')
                return

        if sdsfeaturename is None:
            table = '{0}\\SedimentManagementCell'.format(ws)
            where = build_where_clause(table, 'littoralCellIDPK', lit_id)
            sdsfeaturename = da.get_single_value_from_table(table, 'sdsFeatureName', where)

        scen = get.get_scenario_by_name(scen_name, lit_id)
        scen_id = scen['scenario']['scenarioIDPK']

        if len(scen['cell']) == 0:
            # Insert the new flux scenario record and update our scen object
            new = True
            scen = statics.get_new_cell_scenario(lit_id, scen)

        # Get the sums of the source and sink flux q values
        snkq, srcq = get.get_sum_of_q_vals(lit_id, scen_id)

        self.log('\tQ values\tsrc: {0}  snk: {1}'.format(srcq, snkq))

        # Calculate residualVolume
        vals['residualVolume'] = calculate_residual_volume(
            srcq, snkq, vals['deltaVolume'], vals['placementVolume'], vals['removalVolume'])

        # Update values in micro scenario
        self.log('update_cell_scenario("{0}", "{1}", {2}'.format(lit_id, scen_id, vals))
        statics.update_cell_scenario(lit_id, scen_id, vals)

        # Update values in macro scenario
        micro_scen_val_id = scen['cell'][0]['scenarioValIDPK']
        statics.update_associated_macro_cells(micro_scen_val_id, vals)

        # update values in layer on mxd
        self.log('\tUpdating values in layer on mxd')
        vals['sdsFeatureName'] = sdsfeaturename

        for cell in scen['cell']:
            where = build_where_clause(scen_name, 'littoralCellIDPK', cell['littoralCellIDFK'])
            da.update_row(scen_name, vals, where)

        self.log('\tUpdating value in main table')
        vals = {
            'sdsFeatureName': sdsfeaturename,
            'sdsFeatureDescription': sdsfeaturedesc
        }

        for cell in scen['cell']:
            table = '{0}\\SedimentManagementCell'.format(ws)
            where = build_where_clause(table, 'littoralCellIDPK', cell['littoralCellIDFK'])
            da.update_row(table, vals, where)

        return scen, new

    def get_new_flux_scenario(self, dir_id, scen_name):
        """
        Determine source and sink cell IDs, create new flux scenario, and return
        the scenario
        :param dir_id: fluxline key string
        :param scen_name: scenario name
        :rtype : scenario
        :return: updated scenario object
        :raise: exception when failing to insert
        """
        # This is a new flux scenario. Determine source and sink cells IDs.
        arcpy.env.workspace = ws = configuration.get_workspace()

        scen_id = get.get_scenario_id_by_name(scen_name)
        budget_type = get.get_budget_type(scen_id).lower()

        if budget_type == 'micro':
            flux_table = '{0}\\SedimentTransportDirection'.format(ws)
            field = 'sedimentTransportDirectionIDPK'
        else:
            flux_table = '{0}\\SedimentMacroFlux'.format(ws)
            field = 'macroTransportDirectionIDPK'

        where = build_where_clause(flux_table, field, dir_id)

        try:
            # Get the selected flux feature's shape
            line = da.get_single_value_from_table(flux_table, 'SHAPE@', where)
        except:
            self.log(str(sys.exc_info()), 'error')
            raise

        # Get the source and sink cell IDs
        source_cell, sink_cell = get.get_source_and_sink_cell(line)

        self.log('\t\tSource: {0}\t\tSink: {1}'.format(source_cell, sink_cell))

        # Create flux scenario with default values
        statics.new_flux_scenario(scen_id, dir_id, 'No Data', source_cell, sink_cell)
        scen = get.get_scenario_by_name(scen_name, dir_id=dir_id)
        return scen

    def calc_new_cell_volumes(self, existing_fluxes, flux_class, new_cell_dictionary, scen_name):
        """Ensure the cell volumes are recalculated if they intersect existing fluxes
        :param existing_fluxes: list of existing flux ids
        :param flux_class: str, unicode name of flux class
        :param new_cell_dictionary: dict containing new cell scenarios
        :param scen_name: str, unicode name of scenario
        """
        scen_id = get.get_scenario_id_by_name(scen_name)

        self.log(new_cell_dictionary.values)

        self.log('Selecting existing fluxes')
        where = build_in_clause(flux_class, 'sedimentTransportDirectionIDPK', existing_fluxes, True)
        self.log(where)
        arcpy.SelectLayerByAttribute_management(flux_class, 'NEW_SELECTION', where)

        self.log('Selecting new cells')
        where = build_in_clause(scen_name, 'littoralCellIDPK', new_cell_dictionary.keys(), True)
        arcpy.SelectLayerByAttribute_management(scen_name, 'NEW_SELECTION', where)

        # give the intersect table a random name, as it will be removed shortly anyway
        intersect = random_alpha_string()
        self.log('Creating temporary intersect {}'.format(intersect))
        arcpy.Intersect_analysis([flux_class, scen_name], intersect, 'ALL')

        self.log('Recalculating in intersect')
        statics.recalculate_intersect(intersect, new_cell_dictionary, scen_id)

        self.log('Removing intersect {}'.format(intersect))
        intersect = '{0}\\{1}'.format(arcpy.env.workspace, intersect)
        arcpy.Delete_management(intersect)

    def set_up_environment(self, basemap, workspace, dataframe):
        """
        Will insert template tables into workspace, and add the selected basemap to the MXD.
        :param workspace: path to workspace to use
        :param dataframe: name of dataframe to use
        :param basemap: string indicating which basemap to load, if any.
        :return: output basemap layer
        """
        for p in (basemap, workspace, dataframe):
            assert isinstance(p, (unicode, str))

        configuration.set_workspace(workspace)
        mxd = arcpy.mapping.MapDocument(self.map_document)

        # Note: there is no way to programmatically create a dataframe
        # within arcpy. There's also no easy way to create a dataframe
        # within ArcObjects. Renaming an existing one, however, is trivial.
        # Ideally, I will eventually locate a means of creating a dataframe,
        # but the documentation is poor in this regard and comtypes objects
        # don't seem to match the documentation I'm reading.
        dataframes = arcpy.mapping.ListDataFrames(mxd, dataframe)

        if len(dataframes) == 0:
            ArcObjects().rename_data_frame(dataframe)
            dataframes = arcpy.mapping.ListDataFrames(mxd, dataframe)

        df = dataframes[0]

        # Ensure data frame is in our desired decimal degree projection
        if df.spatialReference.name != 'GCS_WGS_1984':
            df.spatialReference = arcpy.SpatialReference(4326)

        self.log('\t\tSpatialReference: {}'.format(df.spatialReference.name))

        # Get a list of tables and feature classes from the template geodatabase
        template_gdb = r'{}\Templates\SBAS.gdb'.format(sbas_dir)
        workspace = arcpy.env.workspace
        arcpy.env.workspace = template_gdb
        self.log('\tTemplate: {}'.format(template_gdb))
        tables = arcpy.ListTables()
        feature_classes = arcpy.ListFeatureClasses()

        # Remove any template basemaps the user did not choose.
        if basemap in ('Shoreline', 'None'):
            feature_classes.remove('BasemapCounties')

        if basemap in ('Counties', 'None'):
            feature_classes.remove('BasemapShoreline')

        arcpy.env.workspace = workspace
        out_layer = None

        # Copy template table to workspace if it does not exist
        mxd_table_list = arcpy.ListTables()
        for table in tables:
            if table not in mxd_table_list:
                self.log('\tCopying {} rows to workspace GDB'.format(table))
                in_table = r'{0}\{1}'.format(template_gdb, table)
                out_table = r'{0}\{1}'.format(arcpy.env.workspace, table)
                arcpy.CopyRows_management(in_table, out_table)

        mxd_fc_list = arcpy.ListFeatureClasses()
        for fc in feature_classes:
            if fc not in mxd_fc_list:
                self.log('\tCopying {} features to workspace GDB'.format(fc))
                in_table = r'{0}\{1}'.format(template_gdb, fc)
                out_table = r'{0}\{1}'.format(arcpy.env.workspace, fc)
                arcpy.CopyFeatures_management(in_table, out_table)

        # Add basemap to data frame if the user chose to
        for fc in feature_classes:
            if fc in ('BasemapShoreline', 'BasemapCounties'):
                out_layer = arcpy.mapping.Layer(fc)
                arcpy.mapping.AddLayer(df, out_layer)

        return out_layer

    def add_features_to_alternative(self, scenario_name, selected=None):
        """
        Will ensure all selected features are added to the specified alternative
        with default values for their fields.
        :param scenario_name: name of scenario to update
        :param selected: dictionary of selected cells and fluxes
        """
        assert isinstance(scenario_name, (unicode, str))
        assert selected is None or isinstance(selected, dict)

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        self.log('add_features_to_alternative("{}")'.format(scenario_name))
        units = configuration.get_current_units()
        flux_class = '{} Fluxes'.format(scenario_name)

        if selected is None:
            selected = self.get_selected_feature_ids()

        vals = {
            'deltaVolume': 0.0
            , 'placementVolume': 0.0
            , 'removalVolume': 0.0
            , 'valConfidence': 'No Data'
        }

        for feat_id in selected['SedimentManagementCell']:
            self.log('\tAdding cell {}'.format(feat_id))
            self.update_cell_values(scenario_name, vals, units, lit_id=feat_id)

        for feat_id in selected['SedimentTransportDirection']:
            self.log('\tAdding flux {}'.format(feat_id))
            self.update_flux_values(scenario_name, flux_class, 0.0, 'No Data', feat_id, new=True)

        # re-load the alternative
        scenario_id = get.get_scenario_id_by_name(scenario_name)
        load.reload_sbas_layers(scenario_name, 'Unassigned Cells', scen_id=scenario_id)
        load.reload_sbas_layers('{} Fluxes'.format(scenario_name), 'Unassigned Fluxes',
                                scen_id=scenario_id)

        self.log('\tDone adding values to scenario.')
        return selected

    def graphic_conversion_worker(self, scen_name, feat_name, convert_all, clear, micro_budget=None):
        """
        Main driver for logic to convert graphics.
        :param scen_name: name of scenario to which we are adding graphics
        :param feat_name: base name of features to create from graphics
        :param convert_all: whether to convert all graphics
        :param clear: whether to clear all graphics, selected graphics, or none
        :param micro_budget: if macro, the name of the constituent micro budget
        :return: None
        """
        self.log('graphic_conversion_worker("{0}","{1}","{2}","{3}","{4}")'.format(
            scen_name, feat_name, convert_all, clear, micro_budget
        ))
        self.log('\tscen_name: {}'.format(type(scen_name)))

        ws = configuration.get_workspace()
        arcpy.env.workspace = ws

        ao = ArcObjects()
        ids = ao.add_selected_graphics_to_tables(scen_name, feat_name, convert_all)

        if clear == 'Clear All Graphics':
            ao.clear_all_graphics()
        elif clear == 'Clear Only Selected Graphics':
            ao.clear_selected_graphics()

        del ao
        self.log('Graphics cleared')

        scen = None
        new_flux = False
        micro_scen_id = None
        macro_scen_id = None
        existing_fluxes = None
        new_cells = None

        if scen_name is None:
            return scen, new_flux, existing_fluxes, new_cells, macro_scen_id

        # Get existing flux ids
        scen_id = get.get_scenario_id_by_name(scen_name)
        existing_fluxes = get.get_flux_ids_in_scenario(scen_id)

        flux_class = '{} Fluxes'.format(scen_name)
        q_val, confidence = 0.0, 'No Data'

        if micro_budget is not None:
            # Get the constituent micro scenario.
            # Move the graphics to the macro cell table.
            micro_scen_id = get.get_scenario_id_by_name(micro_budget)
            statics.move_features_to_macro(ids.keys())
            macro_scen_id = scen_id

        # Create scenario values for the features
        new_cells = dict()
        for g_id, data in ids.iteritems():
            feat_name, g_type = data[0], data[1]

            if g_type == 'POLYLINE':
                new_flux = True
                try:
                    new, scen, cells = self.update_flux_values(
                        scen_name, flux_class, q_val, confidence, g_id, new=True
                    )
                except Exception:
                    import traceback
                    self.log(traceback.format_exc())
                    raise
            else:

                if micro_budget is None:
                    # Create a new cell scenario record and keep it for later use
                    # when updating residual volumes
                    cell = statics.new_littoral_scenario(scen_id, g_id, confidence)
                    new_cells[g_id] = cell
                else:
                    self.log('>>>>>ADD MICROS TO MACRO')
                    self.log('\tGID: {0}\tMicro Scen ID: {1}'.format(g_id, micro_scen_id))
                    micro_ids = get.get_constituent_cell_ids(g_id)
                    cell_scen_ids = get.get_cell_scenario_ids(micro_ids, micro_scen_id)
                    statics.insert_macro_scenario_records(g_id, scen_id, micro_scen_id, cell_scen_ids)

        if scen is None:
            scen = get.get_scenario_by_name(scen_name)

        return scen, new_flux, existing_fluxes, new_cells, macro_scen_id

    def upload_to_enterprise(self, scenario_id, destination, destination_schema):
        """
        Upload records for scenario to the enterprise.
        :param scenario_id: unicode or str
        :param destination: unicode or str
        :param destination_schema: unicode or str
        """
        self.log('Uploading metadata')
        arcpy.env.workspace = ws = configuration.get_workspace()

        table = '{0}\\Metadata'.format(ws)
        where = build_where_clause(table, 'scenarioIDFK', scenario_id)
        meta_str = da.get_single_value_from_table(table, 'metadata', where)
        user, password = configuration.get_credentials()
        meta_id, doc_id = metadata.post_to_metadata_manager(meta_str, user, password)
        metadata.update_metadata_id(scenario_id, meta_id, doc_id)

        target = r'{0}\{1}'.format(destination, destination_schema)
        self.upload_scenario(scenario_id, target)
        self.upload_cells(scenario_id, target)
        self.upload_fluxes(scenario_id, target)

        if get.get_budget_type(scenario_id) == 'Macro':

            self.upload_macro(scenario_id, target)

            # Upload any consituent micro scenarios if the scenario is a macro scenario
            for micro_scen_id in get.get_child_micro_scen_ids(scenario_id):
                self.log('Uploading child scenario {}'.format(micro_scen_id))
                self.upload_to_enterprise(micro_scen_id, destination, destination_schema)

    def upload_macro(self, scenario_id, target):
        """
        Uploads extra data associated with macro scenarios.
        :param scenario_id: unicode or str
        :param target: unicode or str
        """
        arcpy.env.workspace = ws = configuration.get_workspace()
        table = 'SedimentManagementMacroCell'
        target_table = '{0}.{1}'.format(target, table)

        try:
            # macro cells
            where = get.get_assigned_cells_in(scenario_id, 'MACROCELLIDPK')
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('MACROCELLIDPK', 'macroCellIDPK')
            fields = ('SHAPE@', 'macroCellIDPK', 'sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK', 'projectID',
                      'sdsID', 'sdsMetadataID')
            ins_count = da.copy_data(table, target_table, where, fields)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

            # macro littoral - need to get cell scenario IDs first
            table = '{0}\\SedimentManagementCellScenario'.format(ws)
            where = build_where_clause(table, 'scenarioIDFK', scenario_id)
            macro_scen_val_ids = da.get_column_as_list(table, 'scenarioValIDPK', where)
            table = 'SedimentMacroLittoral'
            target_table = '{0}.{1}'.format(target, table)
            where = build_in_clause(target_table, 'MACROCELLSCENARIOIDFK', macro_scen_val_ids)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('MACROCELLSCENARIOIDFK', 'macroCellScenarioIDFK')
            ins_count = da.copy_data('{0}\\{1}'.format(ws, table), target_table, where)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

            # macro fluxes
            table = 'SedimentMacroFlux'
            target_table = '{0}.{1}'.format(target, table)
            where = get.get_assigned_fluxes_in(scenario_id, 'MACROTRANSPORTDIRECTIONIDPK')
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('MACROTRANSPORTDIRECTIONIDPK', 'macroTransportDirectionIDPK')
            fields = ('SHAPE@', 'sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK', 'projectID',
                      'sdsID', 'sdsMetadataID', 'repTransportType', 'rePurpose', 'macroTransportDirectionIDPK')
            ins_count = da.copy_data(table, target_table, where, fields)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.log('Line {0}:\t{1}'.format(exc_tb.tb_lineno, e.message), 'ERROR')
            self.log(arcpy.GetMessages(), 'ERROR')
            raise

    def upload_scenario(self, scenario_id, target):
        """
        Uploads the scenario to the target table.
        :param scenario_id: unicode or str
        :param target: unicode or str
        """
        arcpy.env.workspace = ws = configuration.get_workspace()
        table = 'SedimentManagementScenario'
        target_table = '{0}.{1}'.format(target, table)

        try:
            where = build_where_clause(target_table, 'SCENARIOIDPK', scenario_id)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('SCENARIOIDPK', 'scenarioIDPK')
            ins_count = da.copy_data('{0}\\{1}'.format(ws, table), target_table, where)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.log('Line {0}:\t{1}'.format(exc_tb.tb_lineno, e.message), 'ERROR')
            self.log(arcpy.GetMessages(), 'ERROR')
            raise

    def upload_cells(self, scenario_id, target):
        """
        Upload cells under the scenario to the target data sets.
        :param scenario_id: unicode or str
        :param target: unicode or str
        """
        where = get.get_assigned_cells_in(scenario_id).replace('littoralCellIDPK', 'LITTORALCELLIDPK')

        try:
            # clear and insert cells
            table = 'SedimentManagementCell'
            target_table = '{0}.{1}'.format(target, table)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('LITTORALCELLIDPK', 'littoralCellIDPK')
            fields = ('SHAPE@', 'sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK', 'projectID', 'sdsID',
                      'sdsMetadataID', 'littoralCellIDPK', 'scenarioIDFK')
            ins_count = da.copy_data(table, target_table, where, fields)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

            # clear and insert cell scenario values
            table = 'SedimentManagementCellScenario'
            target_table = '{0}.{1}'.format(target, table)
            where = build_where_clause(target_table, 'SCENARIOIDFK', scenario_id)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('SCENARIOIDFK', 'scenarioIDFK')
            ins_count = da.copy_data(table, target_table, where, '*')
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.log('Line {0}:\t{1}'.format(exc_tb.tb_lineno, e.message), 'ERROR')
            self.log(arcpy.GetMessages(), 'ERROR')
            raise

    def upload_fluxes(self, scenario_id, target):
        """
        Upload fluxes under the scenario to the target data sets.
        :param scenario_id: unicode or str
        :param target: unicode or str
        """
        where = get.get_assigned_fluxes_in(scenario_id).replace(
            'sedimentTransportDirectionIDPK', 'SEDIMENTTRANSPORTDIRECTIONIDPK')

        try:
            # clear and insert flux lines
            table = 'SedimentTransportDirection'
            target_table = '{0}.{1}'.format(target, table)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('SEDIMENTTRANSPORTDIRECTIONIDPK', 'sedimentTransportDirectionIDPK')
            fields = ('SHAPE@', 'sdsFeatureName', 'sdsFeatureDescription', 'mediaIDFK', 'projectID', 'sdsID',
                      'sdsMetadataID', 'repTransportType', 'rePurpose', 'sedimentTransportDirectionIDPK')
            ins_count = da.copy_data(table, target_table, where, fields)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

            # clear and insert flux scenario values
            table = 'SedimentManagementFluxScenario'
            target_table = '{0}.{1}'.format(target, table)
            where = build_where_clause(target_table, 'SCENARIOIDFK', scenario_id)
            del_count = da.clear_data(target_table, where)
            if del_count > 0:
                self.log('{0}: deleted {1} rows'.format(table, del_count))

            where = where.replace('SCENARIOIDFK', 'scenarioIDFK')
            ins_count = da.copy_data(table, target_table, where)
            self.log('{0}: inserted {1} rows'.format(table, ins_count))

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.log('Line {0}:\t{1}'.format(exc_tb.tb_lineno, e.message), 'ERROR')
            self.log(arcpy.GetMessages(), 'ERROR')
            raise

    def export_alternative_zip(self, scen_id, out_file):
        """
        Saves a geodatabase of the matching scenario as a ZIP file.
        :param scen_id: scenario ID to search
        :param out_file: path of output ZIP file
        """
        arcpy.env.workspace = ws = configuration.get_workspace()

        # Make sure our temp workspace is clean
        scen_name = get.get_scenario_name_by_id(scen_id)
        tmp = tempfile.gettempdir()
        fgdb_name = '{0}.gdb'.format(scen_name)
        fgdb = '{0}\\{1}'.format(tmp, fgdb_name)
        silently_remove_dir(fgdb)
        arcpy.CreateFileGDB_management(tmp, fgdb_name)

        # Put all the scenario data into the temp FGDB
        for table in self.micro_tables:
            source = '{0}.{1}'.format(ws, table)
            target = '{0}\{1}'.format(fgdb, table)

            if table == 'SedimentManagementScenario':
                field = 'scenarioIDPK'
            else:
                field = 'scenarioIDFK'

            where = build_where_clause(source, field, scen_id)

            if table in ('SedimentMacroFlux', 'SedimentManagementCell',
                         'SedimentMacroCell', 'SedimentTransportDirection'):
                self.log('Copying features from {0}'.format(table))
                fc_name = random_alpha_string(12)
                arcpy.MakeFeatureLayer_management(source, fc_name, where)
                assert arcpy.Exists(fc_name)
                arcpy.CopyFeatures_management(fc_name, target)
            else:
                self.log('Copying {0}'.format(table))
                arcpy.CreateTable_management(fgdb, table, source)
                da.copy_data(source, target, where)

        # Make ZIP from the temp GDB
        self.log('Making ZIP file...')
        silently_remove_file(out_file)
        zipdir(fgdb, out_file)

        # Clean up
        self.log('Cleaning up...')
        silently_remove_dir(fgdb)

    def import_alternative_zip(self, archive):
        """
        Import any SBAS data from a zipped geodatabase
        :param archive: path to ZIP file
        """
        arcpy.env.workspace = ws = configuration.get_workspace()

        existing_alternatives = get.get_scenario_names('Micro')

        out_dir = tempfile.mkdtemp()
        with zipfile.ZipFile(archive, 'r') as f:
            f.extractall(out_dir)

        # Pull scenario data from the extracted GDB
        for root, _, files in os.walk(out_dir):
            if root.lower().endswith('.gdb'):
                for table in self.micro_tables:
                    self.log('Importing {0}'.format(table))
                    source = '{0}\{1}'.format(root, table)
                    target = '{0}\{1}'.format(ws, table)
                    if table in ('SedimentMacroFlux', 'SedimentManagementCell',
                                 'SedimentMacroCell', 'SedimentTransportDirection'):
                        arcpy.Append_management(source, target, 'TEST')
                    else:
                        da.copy_data(source, target)

        new_alternatives = get.get_scenario_names('Micro')
        for alt in new_alternatives:
            if alt not in existing_alternatives:
                self.log('New Alternative: {0}'.format(alt))

    def delete_alternative(self, scenario_id):
        """
        Removes data matching the scenario ID from all tables/FCs.
        :param scenario_id: scenario ID to delete
        """
        arcpy.env.workspace = ws = configuration.get_workspace()
        for table in self.micro_tables:
            self.log('Deleting scenario from {0}'.format(table))
            target = '{0}.{1}'.format(ws, table)

            if table == 'SedimentManagementScenario':
                field = 'scenarioIDPK'
            else:
                field = 'scenarioIDFK'

            where = build_where_clause(target, field, scenario_id)
            da.clear_data(target, where)

#
# if __name__ == '__main__':
#     scen_id = 'd34c4456-b45c-4bb5-9269-18ee449cce26'
#     sb = SBAS()
#     sb.delete_alternative(scen_id)
